/* s_serv.h: A header for the server functions.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#ifndef INCLUDED_serv_h
#define INCLUDED_serv_h

#include "config.h"

/* The number of seconds between calls to try_connections(). Fiddle with
 * this ONLY if you KNOW what you're doing!
 */
#define TRY_CONNECTIONS_TIME     60

/* collect ziplinks compression ratios/etc every minute */
#define ZIPSTATS_TIME            60

/* number of seconds to wait after server starts up, before
 * starting try_connections()
 * TOO SOON and you can nick collide like crazy.
 */
#define STARTUP_CONNECTIONS_TIME 60

struct Client;
struct ConfItem;
struct Channel;
/* Capabilities */
struct Capability {
  const char *name; /* name of capability */
  unsigned int cap; /* mask value */
};

/* server stuff */
#define CAP_CAP     0x00000001 /* supports CAPABs */
#define CAP_QS      0x00000002 /* supports quit storm removal */
/* channel stuff */
#define CAP_EX      0x00000008 /* Can do +e */
#define CAP_CHW     0x00000010 /* Can do wall */
#define CAP_IE      0x00000040 /* Can do +I */
/* link stuff */
#define CAP_KLN     0x00000200 /* can do remote K-Lines */
#define CAP_GLN     0x00000400 /* can do G-Lines */
#define CAP_ZIP     0x00008000 /* can compress links */
#define CAP_ENC     0x00010000 /* can encrypt links */
/* channel stuff */
#define CAP_KNOCK   0x00020000 /* can do KNOCK */
/* link stuff */
#define CAP_TB      0x00040000 /* can burst topics */
#define CAP_UNKLN   0x00100000 /* can do remote K-Line removal */
#define CAP_CLUSTER 0x00200000 /* supports clustering */
#define CAP_ENCAP   0x00400000 /* supports encapsulation */

#define CAP_MASK    (CAP_QS | CAP_EX | CAP_CHW | CAP_IE | CAP_KLN | CAP_GLN | \
                     CAP_CLUSTER | CAP_ENCAP | CAP_ZIP | CAP_ENC | CAP_KNOCK | \
                     CAP_UNKLN)

#ifdef HAVE_LIBZ
#define CAP_ZIP_SUPPORTED CAP_ZIP
#else
#define CAP_ZIP_SUPPORTED 0
#endif

#ifdef HAVE_LIBCRYPTO
struct EncCapability {
  const char *name; /* name of capability (cipher name) */
  unsigned int cap; /* mask value */
  int keylen; /* keylength (bytes) */
  int cipherid; /* ID number of cipher type */
};

/* Cipher numbers - ONLY APPEND to this list, NEVER replace! */
#define CIPHER_BF     1
#define CIPHER_CAST   2
#define CIPHER_DES    3
#define CIPHER_3DES   4
#define CIPHER_IDEA   5
#define CIPHER_RC5_8  6
#define CIPHER_RC5_12 7
#define CIPHER_RC5_16 8

/* Cipher Capabilities */
#define CAP_ENC_BF_128     0x00000001
#define CAP_ENC_BF_256     0x00000002
#define CAP_ENC_CAST_128   0x00000004
#define CAP_ENC_DES_56     0x00000008
#define CAP_ENC_3DES_168   0x00000010
#define CAP_ENC_IDEA_128   0x00000020
#define CAP_ENC_RC5_8_128  0x00000040
#define CAP_ENC_RC5_12_128 0x00000080
#define CAP_ENC_RC5_16_128 0x00000100

#define CAP_ENC_ALL        0xFFFFFFFF

#ifdef HAVE_EVP_BF_CFB
#define USE_CIPHER_BF 1
/* Check for bug handling variable length blowfish keys */
#if OPENSSL_VERSION_NUMBER >= 0x00000000L
#define USE_CIPHER_BF_V 1
#else
#define USE_CIPHER_BF_V 0
#endif
#else
#define USE_CIPHER_BF_V 0
#define USE_CIPHER_BF   0
#endif
/* Cast */
#ifdef HAVE_EVP_CAST5_CFB
#define USE_CIPHER_CAST 1
#else
#define USE_CIPHER_CAST 0
#endif
/* DES */
#ifdef HAVE_EVP_DES_CFB
#define USE_CIPHER_DES 1
#else
#define USE_CIPHER_DES 0
#endif
/* 3DES */
#ifdef HAVE_EVP_DES_EDE3_CFB
#define USE_CIPHER_3DES 1
#else
#define USE_CIPHER_3DES 0
#endif
/* IDEA */
#ifdef HAVE_EVP_IDEA_CFB
#define USE_CIPHER_IDEA 1
#else
#define USE_CIPHER_IDEA 0
#endif
/* RC5 */
#ifdef HAVE_EVP_RC5_32_12_16_CFB
#define USE_CIPHER_RC5 1
#else
#define USE_CIPHER_RC5 0
#endif

/* Only enable ciphers supported by available version of OpenSSL */
#define CAP_ENC_MASK (((USE_CIPHER_BF   * CAP_ENC_ALL) & CAP_ENC_BF_128)     | \
                      ((USE_CIPHER_BF_V * CAP_ENC_ALL) & CAP_ENC_BF_256)     | \
                      ((USE_CIPHER_CAST * CAP_ENC_ALL) & CAP_ENC_CAST_128)   | \
                      ((USE_CIPHER_DES  * CAP_ENC_ALL) & CAP_ENC_DES_56)     | \
                      ((USE_CIPHER_3DES * CAP_ENC_ALL) & CAP_ENC_3DES_168)   | \
                      ((USE_CIPHER_IDEA * CAP_ENC_ALL) & CAP_ENC_IDEA_128)   | \
                      ((USE_CIPHER_RC5  * CAP_ENC_ALL) & CAP_ENC_RC5_8_128)  | \
                      ((USE_CIPHER_RC5  * CAP_ENC_ALL) & CAP_ENC_RC5_12_128) | \
                      ((USE_CIPHER_RC5  * CAP_ENC_ALL) & CAP_ENC_RC5_16_128))

#define IsCapableEnc(x, cap)  ((x)->localClient->enc_caps &   (cap))
#define SetCapableEnc(x, cap) ((x)->localClient->enc_caps |=  (cap))
#define ClearCapEnc(x, cap)   ((x)->localClient->enc_caps &= ~(cap))

#endif /* HAVE_LIBCRYPTO */

#define DoesCAP(x)      ((x)->caps)

#define CHECK_SERVER_CRYPTLINK    1
#define CHECK_SERVER_NOCRYPTLINK  0

/* Capability macros */
#define IsCapable(x, cap)  (((x)->localClient->caps & (cap)) == (cap))
#define NotCapable(x, cap) (((x)->localClient->caps & (cap)) == 0)
#define SetCapable(x, cap) ((x)->localClient->caps |= (cap))
#define ClearCap(x, cap)   ((x)->localClient->caps &= ~(cap))

#define SLINKCMD_SET_ZIP_OUT_LEVEL     1
#define SLINKCMD_START_ZIP_OUT         2
#define SLINKCMD_START_ZIP_IN          3
#define SLINKCMD_SET_CRYPT_IN_CIPHER   4
#define SLINKCMD_SET_CRYPT_IN_KEY      5
#define SLINKCMD_START_CRYPT_IN        6
#define SLINKCMD_SET_CRYPT_OUT_CIPHER  7
#define SLINKCMD_SET_CRYPT_OUT_KEY     8
#define SLINKCMD_START_CRYPT_OUT       9
#define SLINKCMD_INJECT_RECVQ          10
#define SLINKCMD_INJECT_SENDQ          11
#define SLINKCMD_INIT                  12
#define SLINKCMD_ZIPSTATS              13

#ifndef HAVE_SOCKETPAIR
#define LAST_SLINK_FD   7
#else
#define LAST_SLINK_FD   5
#endif

#define SLINKRPL_FLAG_DATA 0x0001 /* reply has data following */
#define SLINKRPL_ERROR     1
#define SLINKRPL_ZIPSTATS  2

#define MAX_SLINKRPL       2

typedef void SlinkRplHnd(unsigned int replyid, unsigned int datalen,
  unsigned char *data, struct Client * client_p);
struct SlinkRplDef {
  unsigned int replyid;
  SlinkRplHnd *handler;
  unsigned int flags;
};

extern struct SlinkRplDef slinkrpltab[];
/* Globals */
extern struct Capability captab[];
#ifdef HAVE_LIBCRYPTO
extern struct EncCapability CipherTable[];
#endif

extern int MaxClientCount; /* peak clients */
extern int MaxConnectionCount; /* peak number connections */

extern int refresh_user_links;
/* return values for hunt_server */
#define HUNTED_NOSUCH   (-1) /* hunted server not found */
#define HUNTED_ISME     0    /* hunted server is this server */
#define HUNTED_PASS     1    /* message passed onward successfully */

extern int check_server(const char *name, struct Client * server,
  int cryptlink);
extern int hunt_server(struct Client * client_pt, struct Client * source_pt,
  const char *command, int server, int parc, char **parv);
extern const char *my_name_for_link(const char *name, struct ConfItem * conf);
extern void send_capabilities(struct Client *, struct ConfItem * conf, int,
  int);
extern void write_links_file(void *);
extern int server_estab(struct Client * client_p);
extern void set_autoconn(struct Client *, char *, int);
extern const char *show_capabilities(struct Client * client);
extern void try_connections(void *unused);
extern void start_collect_zipstats(void);
extern void collect_zipstats(void *unused);

extern void add_server_to_list(struct Client *);
extern void remove_server_from_list(struct Client *);

extern void initServerMask(void);
extern void sendnick_TS(struct Client *, struct Client *);
extern int serv_connect(struct ConfItem *, struct Client *);
extern unsigned long nextFreeMask(void);
extern void cryptlink_init(struct Client * client_p, struct ConfItem * aconf,
  int fd);
extern void cryptlink_regen_key(void *);

struct EncCapability *check_cipher(struct Client * client_p,
  struct ConfItem * aconf);
#endif /* INCLUDED_s_serv_h */
