/* m_kill.c: Kills a user.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "hash.h"  /* for find_client() */
#include "ircd.h"
#include "numeric.h"
#include "sprintf_irc.h"
#include "s_log.h"
#include "s_serv.h"
#include "s_conf.h"
#include "send.h"
#include "whowas.h"
#include "irc_string.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"

static char buf[BUFSIZE];
static void ms_kill(struct Client *, struct Client *, int, char **);
static void mo_kill(struct Client *, struct Client *, int, char **);
static void relay_kill(struct Client *, struct Client *, struct Client *,
   const char *, const char *);

struct Message kill_msgtab = {
  "KILL", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, ms_kill, m_ignore, mo_kill}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&kill_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&kill_msgtab);
}
const char *_version = "1.82.2.8";
#endif

/* mo_kill
 *  parv[0] = sender prefix
 *  parv[1] = kill victim
 *  parv[2] = kill path
 */
static void mo_kill(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  struct Client *target_p;
  const char *inpath = client_p->name;
  char *user;
  static char noreason[] = "No Reason";
  const char *reason;
  user = parv[1];
  reason = parv[2]; /* defined or NULL (parc >= 2) */
  if (!IsOperLocalKill(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to perform local kills.", me.name, parv[0]);
    return;
  }
  if (!BadPtr(reason)) {
    if (strlen(reason) > (size_t) KILLLEN)
      ((char *)reason)[KILLLEN] = '\0';
  }
  else
    reason = noreason;
  if ((target_p = find_client(user)) == NULL) {
    /* chase the user if they changed nicknames */
    if ((target_p = get_history(user, (long)KILLCHASETIMELIMIT)) == NULL) {
      sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, parv[0], user);
      return;
    }
    sendto_one(source_p, ":%s NOTICE %s :User %s changed nickname to %s; "
      "re-adjusting target to comphensate.", me.name, parv[0], user,
      target_p->name);
  }
  if (IsServer(target_p) || IsMe(target_p)) {
    sendto_one(source_p, form_str(ERR_CANTKILLSERVER), me.name, parv[0]);
    return;
  }
  if (!MyConnect(target_p) && (!IsOperGlobalKill(source_p))) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to perform global kills.", me.name, parv[0], target_p->name);
    return;
  }
  if (MyConnect(target_p))
    sendto_one(target_p, ":%s!%s@%s KILL %s :%s", parv[0], source_p->username,
      source_p->host, target_p->name, reason);
  sendto_realops_flags(UMODE_ALL, L_ALL, "Received KILL message for %s from "
    "%s, via path: %s (%s)", target_p->name, parv[0], me.name, reason);
  ilog(L_INFO, "KILL %s by %s via %s (%s)", target_p->name, parv[0],
    me.name, reason);

  /* pass KILL to other servers, unless it was a suicide kill */
  if (!MyConnect(target_p)) {
    relay_kill(client_p, source_p, target_p, inpath, reason);
    /* Set FLAGS_KILLED, which prevents exit_one_client from sending QUIT */
    target_p->flags |= FLAGS_KILLED;
  }
  ircsprintf(buf, "Killed (%s (%s))", parv[0], reason);
  exit_client(client_p, target_p, source_p, buf);
}

/* ms_kill
 *  parv[0] = sender prefix
 *  parv[1] = kill victim
 *  parv[2] = kill path and reason
 */
static void ms_kill(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  struct Client *target_p;
  char *user;
  static char noreason[] = "No Reason";
  char *reason;
  char *path;
  int chasing = 0;
  *buf = '\0';
  if (*parv[1] == '\0') {
    sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
      "KILL");
    return;
  }
  user = parv[1];
  if (BadPtr(parv[2])) {
    reason = noreason;
    /* hyb6 takes the nick of the killer from the path *sigh* -fl */
    path = source_p->name;
  }
  else {
    reason = strchr(parv[2], ' ');
    if (reason) {
      *reason = '\0';
      reason++;
    }
    else
      reason = noreason;
    path = parv[2];
  }
  if ((target_p = find_client(user)) == NULL) {
    /* If the user has recently changed nick, but only if its not an uid,
     * automatically rewrite the KILL for this new nickname. --this keeps
     * servers in synch when nick change and kill collide
     */
    if ((*user == '.') || (!(target_p = get_history(user,
        (long)KILLCHASETIMELIMIT)))) {
      sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, parv[0], user);
      return;
    }
    sendto_one(source_p, ":%s NOTICE %s :User %s changed nickname to %s; "
      "re-adjusting target to comphensate.", me.name, parv[0], user,
      target_p->name);
    chasing = 1;
  }
  if (IsServer(target_p) || IsMe(target_p)) {
    sendto_one(source_p, form_str(ERR_CANTKILLSERVER), me.name, parv[0]);
    return;
  }
  if (MyConnect(target_p)) {
    if (IsServer(source_p)) {
      /* dont send clients kills from a hidden server */
      if (ConfigServerHide.hide_servers && !IsOper(target_p))
        sendto_one(target_p, ":%s KILL %s :%s", me.name, target_p->name,
          reason);
      else
        sendto_one(target_p, ":%s KILL %s :%s", parv[0], target_p->name,
          reason);
    }
    else
      sendto_one(target_p, ":%s!%s@%s KILL %s :%s", parv[0], source_p->username,
        source_p->host, target_p->name, reason);
  }
  if (IsOper(source_p)) /* send it normally */
    sendto_realops_flags(UMODE_ALL, L_ALL, "Received KILL message for %s from "
      "%s, via path: %s (%s)", target_p->name, parv[0], source_p->user->server,
      reason);
  else
    sendto_realops_flags(UMODE_SKILL, L_ALL, "Received KILL message for %s "
      "from %s (%s)", target_p->name, parv[0], reason);

  ilog(L_INFO, "KILL %s by %s via %s %s", target_p->name, parv[0],
    parv[0], reason);
  relay_kill(client_p, source_p, target_p, path, reason);
  /* FLAGS_KILLED prevents a quit being sent out */
  target_p->flags |= FLAGS_KILLED;
  /* reason comes supplied with its own ()'s */
  if (ConfigServerHide.hide_servers && IsServer(source_p))
    ircsprintf(buf, "Killed (%s (%s))", me.name, reason);
  else
    ircsprintf(buf, "Killed (%s (%s))", parv[0], reason);
  exit_client(client_p, target_p, source_p, buf);
}
static void relay_kill(struct Client * one, struct Client * source_p,
  struct Client * target_p, const char *inpath, const char *reason)
{
  dlink_node *ptr;
  struct Client *client_p;
  DLINK_FOREACH(ptr, serv_list.head) {
    client_p = (struct Client *) ptr->data;
    if (!client_p || client_p == one)
      continue;
    if (MyOper(source_p))
      sendto_one(client_p, ":%s KILL %s :%s %s", source_p->name, target_p->name,
        source_p->name, reason);
    else
      sendto_one(client_p, ":%s KILL %s :%s %s", source_p->name, target_p->name,
        inpath, reason);
  }
}
