/* m_nick.c: Sets a users nick.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "hash.h"
#include "fdlist.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "s_conf.h"
#include "s_stats.h"
#include "s_user.h"
#include "hash.h"
#include "whowas.h"
#include "s_serv.h"
#include "send.h"
#include "channel.h"
#include "s_log.h"
#include "resv.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "common.h"
#include "packet.h"

static void mr_nick(struct Client *, struct Client *, int, char **);
static void m_nick(struct Client *, struct Client *, int, char **);
static void ms_nick(struct Client *, struct Client *, int, char **);
static int nick_from_server(struct Client *, struct Client *, int, char **,
  time_t, char *);
static int check_clean_nick(struct Client *, struct Client *, char *, char *,
  char *);
static int check_clean_user(struct Client *, char *, char *, char *);
static int check_clean_host(struct Client *, char *, char *, char *);
static int clean_nick_name(char *);
static int clean_user_name(char *);
static int clean_host_name(char *);
static int perform_nick_collides(struct Client *, struct Client *,
  struct Client *, int, char **, time_t, char *);

struct Message nick_msgtab = {
  "NICK", 0, 0, 1, 0, MFLG_SLOW, 0,
  {mr_nick, m_nick, ms_nick, m_ignore, m_nick}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&nick_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&nick_msgtab);
}
const char *_version = "1.122.2.4";
#endif

/* mr_nick
 *  parv[0] = sender prefix
 *  parv[1] = nickname
 */
static void mr_nick(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  struct Client *target_p;
  char nick[NICKLEN];
  char *s;
  if (parc < 2 || BadPtr(parv[1])) {
    sendto_one(source_p, form_str(ERR_NONICKNAMEGIVEN), me.name,
      BadPtr(parv[0]) ? "*" : parv[0]);
    return;
  }
  /* Terminate the nick at the first ~ */
  if ((s = strchr(parv[1], '~')))
    *s = '\0';
  /* and if the first ~ was the first letter.. */
  if (BadPtr(parv[1])) {
    sendto_one(source_p, form_str(ERR_ERRONEUSNICKNAME), me.name,
      BadPtr(parv[0]) ? "*" : parv[0], parv[1]);
    return;
  }
  /* copy the nick and terminate it */
  strlcpy(nick, parv[1], sizeof(nick));
  /* check the nickname is ok */
  if (!clean_nick_name(nick) || find_nick_resv(nick)) {
    sendto_one(source_p, form_str(ERR_ERRONEUSNICKNAME), me.name,
      BadPtr(parv[0]) ? "*" : parv[0], parv[1]);
    return;
  }
  if ((target_p = find_client(nick)) == NULL) {
    set_initial_nick(client_p, source_p, nick);
    return;
  }
  else if (source_p == target_p) {
    strcpy(source_p->name, nick);
    return;
  }
  else
    sendto_one(source_p, form_str(ERR_NICKNAMEINUSE), me.name, "*", nick);
}

/* m_nick
 *  parv[0] = sender prefix
 *  parv[1] = nickname
 */
static void m_nick(struct Client * client_p, struct Client * source_p, int parc,
  char *parv[])
{
  char nick[NICKLEN];
  struct Client *target_p;
  if (parc < 2 || BadPtr(parv[1])) {
    sendto_one(source_p, form_str(ERR_NONICKNAMEGIVEN), me.name, parv[0]);
    return;
  }
  /* mark end of grace period, to prevent nickflooding */
  if (!IsFloodDone(source_p))
    flood_endgrace(source_p);
  /* truncate nick to NICKLEN */
  strlcpy(nick, parv[1], sizeof(nick));
  /* check the nickname is ok */
  if (!clean_nick_name(nick) || find_nick_resv(nick)) {
    sendto_one(source_p, form_str(ERR_ERRONEUSNICKNAME), me.name, parv[0],
      nick);
    return;
  }
  if ((target_p = find_client(nick))) {
    /* If(target_p == source_p) the client is changing nicks between equivalent
     * nicknames ie: [nick] -> {nick}
     */
    if (target_p == source_p) {
      /* check the nick isnt exactly the same */
      if (strcmp(target_p->name, nick))
        change_local_nick(client_p, source_p, nick);
      return;
    }
    /* if the client that has the nick isnt registered yet (nick but no user)
     * then drop the unregged client
     */
    if (IsUnknown(target_p)) {
      /* the old code had an if(MyConnect(target_p)) here.. but I cant see how
       * that can happen, m_nick() is local only --fl_
       */
      exit_client(NULL, target_p, &me, "Overridden");
      change_local_nick(client_p, source_p, nick);
      return;
    }
    else {
      sendto_one(source_p, form_str(ERR_NICKNAMEINUSE), me.name, parv[0], nick);
      return;
    }
  }
  else {
    change_local_nick(client_p, source_p, nick);
    return;
  }
}

/* ms_nick
 * server -> server nick change
 *  parv[0] = sender prefix
 *  parv[1] = nickname
 *  parv[2] = TS when nick change
 * server introducing new nick
 *  parv[0] = sender prefix
 *  parv[1] = nickname
 *  parv[2] = hop count
 *  parv[3] = TS
 *  parv[4] = umode
 *  parv[5] = username
 *  parv[6] = hostname
 *  parv[7] = server
 *  parv[8] = ircname
 */
static void ms_nick(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  struct Client *target_p;
  char nick[NICKLEN];
  time_t newts = 0;
  if (parc < 2 || BadPtr(parv[1])) {
    sendto_one(source_p, form_str(ERR_NONICKNAMEGIVEN), me.name, parv[0]);
    return;
  }
  /* parc == 3 on nickchange, parc == 9 on new nick */
  if ((IsClient(source_p) && (parc != 3)) || (IsServer(source_p) && (parc !=
    9))) {
    char tbuf[BUFSIZE] = {0};
    int j;
    for (j = 0; j < parc; j++) {
      strcat(tbuf, parv[j]);
      strcat(tbuf, " ");
    }
    sendto_realops_flags(UMODE_ALL, L_ALL, "Dropping server %s due to command "
      "NICK with only %d arguments. (%s)", client_p->name, parc, tbuf);
    ilog(L_CRIT, "Insufficient parameters (%d) for command NICK from %s. "
      "(%s)", parc, client_p->name, tbuf);
    exit_client(client_p, client_p, client_p, "Not enough arguments to server "
      "command.");
    return;
  }
  /* fix the length of the nick */
  strlcpy(nick, parv[1], sizeof(nick));
  if (check_clean_nick(client_p, source_p, nick, parv[1], (parc == 9 ? parv[7] :
    (char *)source_p->user->server)))
    return;
  if (parc == 9) {
    if (check_clean_user(client_p, nick, parv[5], parv[7]) ||
      check_clean_host(client_p, nick, parv[6], parv[7]))
      return;
    /* check the length of the clients gecos */
    if (strlen(parv[8]) > REALLEN) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "Long realname from server %s for "
        "%s (%d > %d)", parv[7], parv[1], strlen(parv[8]), REALLEN);
      parv[8][REALLEN] = '\0';
    }
    if (IsServer(source_p))
      newts = atol(parv[3]);
  }
  else if (!IsServer(source_p))
      newts = atol(parv[2]);
  /* if the nick doesnt exist, allow it and process like normal */
  if (!(target_p = find_client(nick))) {
    nick_from_server(client_p, source_p, parc, parv, newts, nick);
    return;
  }
  /* we're not living in the past anymore, an unknown client is local only. */
  if (IsUnknown(target_p)) {
    exit_client(NULL, target_p, &me, "Overridden");
    nick_from_server(client_p, source_p, parc, parv, newts, nick);
    return;
  }
  if (target_p == source_p) {
    if (strcmp(target_p->name, nick)) {
      /* client changing case of nick */
      nick_from_server(client_p, source_p, parc, parv, newts, nick);
      return;
    }
    else
      return; /* client not changing nicks at all */
  }
  perform_nick_collides(source_p, client_p, target_p, parc, parv, newts, nick);
}

/* check_clean_nick
 * input        - pointer to source
 *              - nickname
 *              - truncated nickname
 *              - origin of client
 * side effects - if nickname is erroneous, or a different length to truncated
 *                nickname, return 1
 */
static int check_clean_nick(struct Client * client_p, struct Client * source_p,
   char *nick, char *newnick, char *server)
{
  /* the old code did some wacky stuff here, if the nick is invalid, kill it
   * and dont bother messing at all
   */
  /* Zero length nicks are bad too */
  if (!clean_nick_name(nick) || strcmp(nick, newnick) || nick[0] == '\0') {
    ServerStats->is_kill++;
    sendto_realops_flags(UMODE_DEBUG, L_ALL, "Bad Nick: %s from server %s, via "
      "%s", nick, server, client_p->name);
    sendto_one(client_p, ":%s KILL %s :%s (Bad Nickname)", me.name, newnick,
      me.name);
    /* bad nick change */
    if (source_p != client_p) {
      kill_client_serv_butone(client_p, source_p, "%s (Bad Nickname)", me.name);
      source_p->flags |= FLAGS_KILLED;
      exit_client(client_p, source_p, &me, "Bad Nickname");
    }
    return YES;
  }
  return NO;
}
/* check_clean_user
 * input        - pointer to client sending data
 *              - nickname
 *              - username to check
 *              - origin of NICK
 * side effects - if username is erroneous, return 1
 */
static int check_clean_user(struct Client * client_p, char *nick, char *user,
  char *server)
{
  if (strlen(user) > USERLEN) {
    ServerStats->is_kill++;
    sendto_realops_flags(UMODE_DEBUG, L_ALL, "Long username for %s from server "
      "%s, via %s (%d > %d)", nick, server, client_p->name, strlen(user),
      USERLEN);
    sendto_one(client_p, ":%s KILL %s :%s (Bad Username)", me.name, nick,
      me.name);
    return YES;
  }
  if (!clean_user_name(user))
    sendto_realops_flags(UMODE_DEBUG, L_ALL, "Bad username for %s from server "
      "%s, via %s (%d > %d)", nick, server, client_p->name, strlen(user),
      USERLEN);
  return NO;
}

/* check_clean_host
 * input        - pointer to client sending us data
 *              - nickname
 *              - hostname to check
 *              - source name
 * side effects - if hostname is erroneous, return 1
 */
static int check_clean_host(struct Client * client_p, char *nick, char *host,
  char *server)
{
  if (strlen(host) > HOSTLEN) {
    ServerStats->is_kill++;
    sendto_realops_flags(UMODE_DEBUG, L_ALL, "Long hostname for %s from server "
      "%s, via %s (%d > %d)", nick, server, client_p->name, strlen(host),
      HOSTLEN);
    sendto_one(client_p, ":%s KILL %s :%s (Bad Hostname)", me.name, nick,
      me.name);
    return YES;
  }
  if (!clean_host_name(host))
    sendto_realops_flags(UMODE_DEBUG, L_ALL, "Bad hostname for %s from server "
      "%s, via %s (%d > %d)", nick, server, client_p->name, strlen(host),
      HOSTLEN);
  return NO;
}

/* clean_nick_name
 * input        - nickname
 * output       - 0 if erroneous, 1 if okay
 */
static int clean_nick_name(char *nick)
{
  s_assert(nick);
  if (nick == NULL)
    return NO;
  /* nicks cant start with a digit or -, and must have a length */
  if (*nick == '-' || IsDigit(*nick) || !*nick)
    return NO;
  for (; *nick; nick++) {
    if (!IsNickChar(*nick))
      return NO;
  }
  return YES;
}

/* clean_user_name
 * input        - username
 * output       - 0 if erroneous, 1 if okay
 */
static int clean_user_name(char *user)
{
  s_assert(user);
  if (user == NULL)
    return NO;
  for (; *user; user++) {
    if (!IsUserChar(*user))
      return NO;
  }
  return YES;
}

/* clean_host_name
 * input        - hostname
 * output       - 0 if erroneous, 1 if okay
 */
static int clean_host_name(char *host)
{
  s_assert(host);
  if (host == NULL)
    return NO;
  for (; *host; host++) {
    if (!IsHostChar(*host))
      return NO;
  }
  return YES;
}

static int nick_from_server(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[], time_t newts, char *nick)
{
  if (IsServer(source_p)) {
    /* A server introducing a new client, change source */
    source_p = make_client(client_p);
    add_client_to_list(source_p);
    if (parc > 2)
      source_p->hopcount = atoi(parv[2]);
    if (newts)
      source_p->tsinfo = newts;
    else {
      newts = source_p->tsinfo = CurrentTime;
      ts_warn("Remote nick %s (%s) introduced without a TS", nick, parv[0]);
    }
    /* copy the nick in place */
    (void)strcpy(source_p->name, nick);
    (void)add_to_client_hash_table(nick, source_p);
    if (parc > 8) {
      int flag;
      char *m;
      /* parse usermodes */
      m = &parv[4][1];
      while (*m) {
        flag = user_modes_from_c_to_bitmask[(unsigned char)*m];
        if (!(source_p->umodes & UMODE_INVISIBLE) && (flag & UMODE_INVISIBLE))
          Count.invisi++;
        if (!(source_p->umodes & UMODE_OPER) && (flag & UMODE_OPER))
          Count.oper++;
        source_p->umodes |= flag & SEND_UMODES;
        m++;
      }
      return do_remote_user(nick, client_p, source_p, parv[5], parv[6], parv[7],
        parv[8]);
    }
  }
  else if (source_p->name[0]) {
    /* client changing their nick */
    if (irccmp(parv[0], nick))
      source_p->tsinfo = newts ? newts : CurrentTime;
    sendto_common_channels_local(source_p, ":%s!%s@%s NICK :%s", source_p->name,
      source_p->username, source_p->host, nick);
    if (source_p->user) {
      add_history(source_p, 1);
      sendto_server(client_p, NULL, NOCAPS, NOCAPS, ":%s NICK %s :%lu", parv[0],
        nick, (unsigned long)source_p->tsinfo);
    }
  }
  /* set the new nick name */
  if (source_p->name[0])
    del_from_client_hash_table(source_p->name, source_p);
  strcpy(source_p->name, nick);
  add_to_client_hash_table(nick, source_p);
  /* remove all accepts pointing to the client */
  del_all_accepts(source_p);
  return NO;
}
static int perform_nick_collides(struct Client * source_p,
  struct Client * client_p, struct Client * target_p, int parc, char *parv[],
  time_t newts, char *nick)
{
  int sameuser;
  /* server introducing new nick */
  if (IsServer(source_p)) {
    /* if we dont have a ts, or their TS's are the same, kill both */
    if (!newts || !target_p->tsinfo || (newts == target_p->tsinfo)) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "Nickname collision on %s (%s <- "
        "%s), killing both", target_p->name, target_p->from->name,
        client_p->name);
      kill_client_serv_butone(NULL, target_p, "%s (Nickname collision)",
        me.name);
      ServerStats->is_kill++;
      sendto_one(target_p, form_str(ERR_NICKCOLLISION), me.name, target_p->name,
        target_p->name);
      target_p->flags |= FLAGS_KILLED;
      exit_client(client_p, target_p, &me, "Nickname collision");
      return NO;
    }
    /* the timestamps are different */
    else {
      sameuser = (target_p->user) && !irccmp(target_p->username, parv[5])
        && !irccmp(target_p->host, parv[6]);
      /*
       * if the users are the same (loaded a client on a different server) and
       * the new users ts is older, or the users are different and the new
       * users ts is newer, ignore the new client and let it do the kill
       */
      if ((sameuser && newts < target_p->tsinfo) || (!sameuser && newts >
        target_p->tsinfo)) {
        return NO;
      }
      else {
        if (sameuser)
          sendto_realops_flags(UMODE_ALL, L_ALL, "Nickname collision on %s (%s "
            "<- %s), killing older", target_p->name, target_p->from->name,
            client_p->name);
        else
          sendto_realops_flags(UMODE_ALL, L_ALL, "Nickname collision on %s (%s "
            "<- %s), killing newer", target_p->name, target_p->from->name,
            client_p->name);
        ServerStats->is_kill++;
        sendto_one(target_p, form_str(ERR_NICKCOLLISION), me.name,
          target_p->name, target_p->name);
        kill_client_serv_butone(source_p, target_p, "%s (Nickname collision)",
          me.name);
        target_p->flags |= FLAGS_KILLED;
        (void)exit_client(client_p, target_p, &me, "Nickname collision");
        nick_from_server(client_p, source_p, parc, parv, newts, nick);
        return NO;
      }
    }
  }
  /* its a client changing nick and causing a collide */
  if (!newts || !target_p->tsinfo || (newts == target_p->tsinfo) ||
    !source_p->user) {
    sendto_realops_flags(UMODE_ALL, L_ALL, "Nickname change collision from %s "
      "to %s (%s <- %s), killing both", source_p->name, target_p->name,
      target_p->from->name, client_p->name);
    ServerStats->is_kill++;
    sendto_one(target_p, form_str(ERR_NICKCOLLISION),
      me.name, target_p->name, target_p->name);
    /* if we got the message from a LL, it knows about source_p */
    kill_client_serv_butone(NULL, source_p, "%s (Nickname collision)", me.name);
    ServerStats->is_kill++;
    kill_client_serv_butone(NULL, target_p, "%s (Nickname collision)", me.name);
    target_p->flags |= FLAGS_KILLED;
    exit_client(NULL, target_p, &me, "Nickname collision");
    source_p->flags |= FLAGS_KILLED;
    exit_client(client_p, source_p, &me, "Nickname collision");
    return NO;
  }
  else {
    sameuser = !irccmp(target_p->username, source_p->username) &&
      !irccmp(target_p->host, source_p->host);
    if ((sameuser && newts < target_p->tsinfo) || (!sameuser && newts >
      target_p->tsinfo)) {
      if (sameuser)
        sendto_realops_flags(UMODE_ALL, L_ALL, "Nickname change collision from "
          "%s to %s (%s <- %s), killing older", source_p->name, target_p->name,
          target_p->from->name, client_p->name);
      else
        sendto_realops_flags(UMODE_ALL, L_ALL, "Nickname change collision from "
          "%s to %s (%s <- %s), killing newer",
          source_p->name, target_p->name, target_p->from->name, client_p->name);
      ServerStats->is_kill++;
      /* this won't go back to the incoming link, so LL doesnt matter */
      kill_client_serv_butone(client_p, source_p, "%s (Nickname collision)",
        me.name);
      source_p->flags |= FLAGS_KILLED;
      if (sameuser)
        exit_client(client_p, source_p, &me, "Nickname collision");
      else
        exit_client(client_p, source_p, &me, "Nickname collision");
      return NO;
    }
    else {
      if (sameuser)
        sendto_realops_flags(UMODE_ALL, L_ALL, "Nickname collision on %s (%s "
          "<- %s), killing older", target_p->name, target_p->from->name,
          client_p->name);
      else
        sendto_realops_flags(UMODE_ALL, L_ALL, "Nickname collision on %s (%s "
          "<- %s), killing newer", target_p->name, target_p->from->name,
          client_p->name);
      kill_client_serv_butone(source_p, target_p, "%s (Nickname collision)",
        me.name);
      ServerStats->is_kill++;
      sendto_one(target_p, form_str(ERR_NICKCOLLISION), me.name, target_p->name,
        target_p->name);
      target_p->flags |= FLAGS_KILLED;
      (void)exit_client(client_p, target_p, &me, "Nickname collision");
    }
  }
  /* client changing nick, not a new client */
  nick_from_server(client_p, source_p, parc, parv, newts, nick);
  return NO;
}
