/* m_server.c: Introduces a server.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "tools.h"
#include "handlers.h"
#include "client.h"
#include "common.h"
#include "event.h"
#include "hash.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "s_stats.h"
#include "scache.h"
#include "send.h"
#include "motd.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#define SERVER_UNAUTH "Unauthorised server connection attempt from %s: "

static void mr_server(struct Client *, struct Client *, int, char **);
static void ms_server(struct Client *, struct Client *, int, char **);
static int set_server_gecos(struct Client *, char *);

struct Message server_msgtab = {
  "SERVER", 0, 0, 4, 0, MFLG_SLOW | MFLG_UNREG, 0,
  {mr_server, m_registered, ms_server, m_ignore, m_registered}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&server_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&server_msgtab);
}
const char *_version = "1.113.2.2";
#endif

int bogus_host(char *host);
struct Client *server_exists(char *);

/* mr_server - SERVER message handler
 *  parv[0] = sender prefix
 *  parv[1] = servername
 *  parv[2] = serverinfo/hopcount
 *  parv[3] = serverinfo
 */
static void mr_server(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  char info[REALLEN + 1];
  char *name;
  struct Client *target_p;
  int hop;
  if (parc < 4) {
    exit_client(client_p, client_p, client_p, "No servername");
    return;
  }
  name = parv[1];
  hop = atoi(parv[2]);
  strlcpy(info, parv[3], sizeof(info));
  /* Reject a direct nonTS server connection if we're TS_ONLY -orabidoo */
  if (!DoesTS(client_p)) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Link %s dropped because it does "
      "not support our protocol version", get_client_name(client_p, HIDE_IP,
      YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Link %s dropped because it does "
      "not support our protocol version", get_client_name(client_p, MASK_IP,
      YES));
    exit_client(client_p, client_p, &me, "Non-TS server");
    return;
  }
  if (bogus_host(name)) {
    exit_client(client_p, client_p, &me, "Invalid server name");
    return;
  }
  /* call check_server to check everything for us */
  switch (check_server(name, client_p, CHECK_SERVER_NOCRYPTLINK)) {
   case -1:
    if (ConfigFileEntry.warn_no_nline) {
      sendto_realops_flags(UMODE_ALL, L_ADMIN, SERVER_UNAUTH "No entry for "
        "servername %s", get_client_name(client_p, HIDE_IP, YES), name);
      sendto_realops_flags(UMODE_ALL, L_OPER, SERVER_UNAUTH "No entry for "
        "servername %s", get_client_name(client_p, MASK_IP, YES), name);
      ilog(L_NOTICE, "Access denied, no connect block for server %s",
        log_client_name(client_p));
    }
    exit_client(client_p, client_p, &me, "Invalid server name");
    return;
    break;
   case -2:
    sendto_realops_flags(UMODE_ALL, L_ADMIN, SERVER_UNAUTH "Bad password for "
      "server %s", get_client_name(client_p, HIDE_IP, YES), name);
    sendto_realops_flags(UMODE_ALL, L_OPER, SERVER_UNAUTH "Bad password for "
      "server %s", get_client_name(client_p, MASK_IP, YES), name);
    ilog(L_NOTICE, "Access denied, invalid password for server %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Invalid password");
    return;
    break;
   case -3:
    sendto_realops_flags(UMODE_ALL, L_ADMIN, SERVER_UNAUTH "Invalid host for "
      "server %s", get_client_name(client_p, HIDE_IP, YES), name);
    sendto_realops_flags(UMODE_ALL, L_OPER, SERVER_UNAUTH "Invalid host for "
      "server %s", get_client_name(client_p, MASK_IP, YES), name);
    ilog(L_NOTICE, "Access denied, invalid host for server %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Invalid hostname");
    return;
    break;
    /* servername is > HOSTLEN */
   case -4:
    sendto_realops_flags(UMODE_ALL, L_ADMIN, SERVER_UNAUTH "Server name "
      "exceeds maximum host length (%d > %d)", get_client_name(client_p,
      HIDE_IP, YES), strlen(name), HOSTLEN);
    sendto_realops_flags(UMODE_ALL, L_OPER, SERVER_UNAUTH "Server name exceeds "
      "maximum host length (%d > %d)", get_client_name(client_p,
      MASK_IP, YES), strlen(name), HOSTLEN);
    ilog(L_NOTICE, "Access denied, invalid servername from %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Invalid server name");
    return;
    break;
  }
  if ((target_p = server_exists(name))) {
    /* Reject the new server, it's from an unregistered connection */
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Attempt to re-introduce server "
      "%s from %s", name, get_client_name(client_p, HIDE_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Attempt to re-introduce server "
      "%s from %s", name, get_client_name(client_p, MASK_IP, YES));
    exit_client(client_p, client_p, &me, "Server Exists");
    return;
  }
  /* we're handshaking, so we already have the aconf name in client_p->name */
  strlcpy(client_p->name, name, sizeof(client_p->name));
  set_server_gecos(client_p, info);
  client_p->hopcount = hop;
  server_estab(client_p);
}
/* ms_server - SERVER message handler
 *  parv[0] = sender prefix
 *  parv[1] = servername
 *  parv[2] = serverinfo/hopcount
 *  parv[3] = serverinfo
 */
static void ms_server(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  char info[REALLEN + 1];
  /* same size as in s_misc.c */
  char *name;
  struct Client *target_p;
  struct Client *bclient_p;
  struct ConfItem *aconf;
  int hop;
  int hlined = 0;
  int llined = 0;
  dlink_node *ptr;
  /* Just to be sure -A1kmm. */
  if (!IsServer(source_p))
    return;

  if (parc < 4) {
    sendto_one(client_p, "ERROR :No servername");
    return;
  }
  name = parv[1];
  hop = atoi(parv[2]);
  strlcpy(info, parv[3], sizeof(info));
  if ((target_p = server_exists(name))) {
    /* This link is trying feed me a server that I already have access through
     * another path -- multiple paths not accepted currently, kill this link
     * immediately!!
     * 
     * Rather than KILL the link which introduced it, KILL the youngest of the two
     * links. -avalon
     * 
     * I think that we should exit the link itself, not the introducer, and we
     * should always exit the most recently received(i.e. the one we are
     * receiving this SERVER for. -A1kmm
     * 
     * You *cant* do this, if you link somewhere, it bursts you a server that
     * already exists, then sends you a client burst, you squit the server, but
     * you keep getting the burst of clients on a server that doesnt exist,
     * although ircd can handle it, its not a realistic solution.. --fl_
     */
    /* behind a host-masked server, ignore the server message */
   if (irccmp(target_p->name, name) && target_p->from == client_p)
      return;
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Link %s cancelled, server %s "
      "already exists", get_client_name(client_p, SHOW_IP, YES), name);
    sendto_realops_flags(UMODE_ALL, L_OPER, "Link %s cancelled, server %s "
      "already exists", get_client_name(client_p, MASK_IP, YES), name);
    exit_client(client_p, client_p, &me, "Server Exists");
    return;
  }
  /* User nicks never have '.' in them, servers always do */
  if (strchr(name, '.') == NULL) {
    /* Server trying to use the same name as a person. Would cause a fair bit
     * of confusion. Enough to make it hellish for a while and servers to send
     * stuff to the wrong place.
     */
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Link %s cancelled, "
      "server/nickname ambiguity on %s", get_client_name(client_p, HIDE_IP,
      YES), name);
    sendto_realops_flags(UMODE_ALL, L_OPER, "Link %s cancelled, "
      "server/nickname ambiguity on %s", get_client_name(client_p, MASK_IP,
      YES), name);
    exit_client(client_p, client_p, &me, "Invalid server name");
    return;
  }
  /* Server is informing about a new server behind this link. Create REMOTE
   * server structure, add it to list and propagate word to my other server
   * links...
   */
  if (parc == 1 || info[0] == '\0') {
    sendto_one(client_p, "ERROR :Server %s is missing info field", name);
    return;
  }
  /* if newly found server is behind a guaranteed leaf, close the link */
  for (aconf = ConfigItemList; aconf; aconf = aconf->next) {
    if ((aconf->status & (CONF_LEAF | CONF_HUB)) == 0)
      continue;
    if (match(aconf->name, client_p->name)) {
      if (aconf->status == CONF_HUB) {
        if (match(aconf->host, name))
          hlined++;
      }
      else if (aconf->status == CONF_LEAF) {
        if (match(aconf->host, name))
          llined++;
      }
    }
  }
  if (strlen(name) > HOSTLEN) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Link %s introduced server that "
      "exceeds maximum host length (%d > %d)", get_client_name(client_p,
      SHOW_IP, YES), strlen(name), HOSTLEN);
    sendto_realops_flags(UMODE_ALL, L_OPER, "Link %s introduced server that "
      "exceeds maximum host length (%d > %d)", get_client_name(client_p,
      MASK_IP, YES), strlen(name), HOSTLEN);
    ilog(L_NOTICE, "Server %s tried introducing a server that exceeds maximum "
      "host length (%d > %d)", log_client_name(client_p), strlen(name),
      HOSTLEN);
    exit_client(client_p, client_p, &me, "Invalid server introduction.");
    return;
  }
  /* Ok, check client_p can hub the new server, and make sure it's not a LL */
  if (!hlined) {
    /* OOOPs nope can't HUB */
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Server %s introduced %s, but is "
      "not a hub!", get_client_name(client_p, SHOW_IP, YES), name);
    sendto_realops_flags(UMODE_ALL, L_OPER, "Server %s introduced %s, but is "
      "not a hub!", get_client_name(client_p, MASK_IP, YES), name);
    ilog(L_NOTICE, "Server %s introduced %s, but is not a hub",
      log_client_name(client_p), name);
    exit_client(client_p, client_p, &me, "No matching hub_mask.");
    return;
  }
  /* Check for the new server being leafed behind this HUB */
  if (llined) {
    /* OOOPs nope can't HUB this leaf */
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Server %s introduced %s, but it "
      "is not permitted on this hub!", get_client_name(client_p, SHOW_IP, YES),
      name);
    sendto_realops_flags(UMODE_ALL, L_OPER, "Server %s introduced %s, but it "
      "is not permitted on this hub!", get_client_name(client_p, MASK_IP, YES),
       name);
    ilog(L_NOTICE, "Server %s introduced %s, but is not permitted to hub it",
      log_client_name(client_p), name);
    exit_client(client_p, client_p, &me, "Not permitted on this hub");
    return;
  }
  target_p = make_client(client_p);
  make_server(target_p);
  target_p->hopcount = hop;
  strlcpy(target_p->name, name, sizeof(target_p->name));
  set_server_gecos(target_p, info);
  target_p->serv->up = find_or_add(parv[0]);
  target_p->servptr = source_p;
  SetServer(target_p);
  Count.server++;
  add_client_to_list(target_p);
  add_server_to_list(target_p);
  add_to_client_hash_table(target_p->name, target_p);
  dlinkAdd(target_p, &target_p->lnode, &target_p->servptr->serv->servers);

  /* Old sendto_serv_but_one() call removed because we now need to send
   * different names to different servers (domain name matching)
   */
  DLINK_FOREACH(ptr, serv_list.head) {
    bclient_p = ptr->data;
    if (bclient_p == client_p)
      continue;
    if (!(aconf = bclient_p->serv->sconf)) {
      sendto_realops_flags(UMODE_ALL, L_ADMIN, "Lost connect configuration for "
        "%s on %s, closing link", get_client_name(client_p, SHOW_IP, YES),
        name);
      sendto_realops_flags(UMODE_ALL, L_OPER, "Lost connect configuration for "
        "%s on %s, closing link", get_client_name(client_p, MASK_IP, YES),
        name);
      ilog(L_NOTICE, "Lost connect configuration for %s on %s",
        log_client_name(client_p), name);
      exit_client(client_p, client_p, &me, "Lost connect configuration");
      return;
    }
    if (match(my_name_for_link(me.name, aconf), target_p->name))
      continue;
    sendto_one(bclient_p, ":%s SERVER %s %d :%s%s", parv[0], target_p->name,
      hop + 1, IsHidden(target_p) ? "(H) " : "", target_p->info);
  }
  sendto_realops_flags(UMODE_EXTERNAL, L_ALL, "Server %s being introduced by "
    "%s", target_p->name, source_p->name);
}

/* set_server_gecos
 * input        - pointer to client
 * side effects - servers gecos field is set
 */
static int set_server_gecos(struct Client * client_p, char *info)
{
  /* check the info for [IP] */
  if (info[0]) {
    char *p;
    char *s;
    char *t;
    s = info;
    /* we should only check the first word for an ip */
    if ((p = strchr(s, ' ')))
      *p = '\0';
    /* check for a ] which would symbolise an [IP] */
    if ((t = strchr(s, ']'))) {
      /* set s to after the first space */
      if (p)
        s = ++p;
      else
        s = NULL;
    }
    /* no ], put the space back */
    else if (p)
      *p = ' ';
    /* p may have been set to a trailing space, so check s exists & isn't \0 */
    if (s && (*s != '\0')) {
      /* a space? if not (H) could be the last part of info.. */
      if ((p = strchr(s, ' ')))
        *p = '\0';
      /* check for (H) which is a hidden server */
      if (!strcmp(s, "(H)")) {
        SetHidden(client_p);
        /* if there was no space.. theres nothing to set info to */
        if (p)
          s = ++p;
        else
          s = NULL;
      }
      else if (p)
        *p = ' ';
      /* if there was a trailing space, s could point to \0, so check */
      if (s && (*s != '\0')) {
        strlcpy(client_p->info, s, sizeof(client_p->info));
        return YES;
      }
    }
  }
  strlcpy(client_p->info, "(Unknown Location)", sizeof(client_p->info));
  return YES;
}

/* bogus_host
 * input  - hostname
 * output - 1 if a bogus hostname input, 0 if its valid
 */
int bogus_host(char *host)
{
  int bogus_server = NO;
  char *s;
  int dots = NO;
  for (s = host; *s; s++) {
    if (!IsServChar(*s)) {
      bogus_server = YES;
      break;
    }
    if ('.' == *s)
      ++dots;
  }
  if (!dots || bogus_server)
    return YES;
  return NO;
}

/* server_exists
 * input  - servername
 * output - 1 if server exists, 0 if doesnt exist
 */
struct Client * server_exists(char *servername)
{
  struct Client *target_p;
  dlink_node *ptr;
  DLINK_FOREACH(ptr, global_serv_list.head) {
    target_p = ptr->data;
    if (match(target_p->name, servername) || match(servername, target_p->name))
      return target_p;
  }
  return NULL;
}
