/* m_challenge.c: Allows an IRC Operator to securely authenticate.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "ircd.h"
#include "modules.h"
#include "numeric.h"
#include "send.h"
#include "s_conf.h"
#include "msg.h"
#include "parse.h"
#include "irc_string.h"
#include "s_log.h"

#ifdef HAVE_LIBCRYPTO
#include <openssl/rsa.h>
#include "rsa.h"

int oper_up(struct Client * source_p, struct ConfItem * aconf);
#endif

static void m_challenge(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[]);
static void mo_challenge(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[]);

struct Message challenge_msgtab = {
  "CHALLENGE", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_challenge, m_ignore, m_ignore, mo_challenge}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&challenge_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&challenge_msgtab);
}
const char *_version = "1.56.3.2";
#endif

#ifdef HAVE_LIBCRYPTO
/* m_challenge - generate RSA challenge for wouldbe oper
 *  parv[0] = sender prefix
 *  parv[1] = operator to challenge for, or +response
 */
static void m_challenge(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  static time_t last_used = 0L;
  char *challenge;
  dlink_node *ptr;
  struct ConfItem *aconf, *oconf;

  if ((last_used + ConfigFileEntry.pace_wait) > CurrentTime) {
    sendto_one(source_p, form_str(RPL_LOAD2HI), me.name, BadPtr(parv[0]) ?
      "*" : parv[0], "CHALLENGE");
    return;
  }
  else
    last_used = CurrentTime;

  if (!MyClient(source_p))
    return;

  if (*parv[1] == '+') {
    /* Ignore it if we aren't expecting this... -A1kmm */
    if (!source_p->user->response)
      return;
    if (irccmp(source_p->user->response, ++parv[1])) {
      sendto_one(source_p, form_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
      log_foper(source_p, source_p->user->auth_oper);
      if (ConfigFileEntry.failed_oper_notice)
        sendto_realops_flags(UMODE_ALL, L_ALL, "%s failed attempt to exit the "
          "Matrix (Incorrect Challenge Response)", get_client_name(source_p,
          HIDE_IP, YES));
      return;
    }
    if ((aconf = find_conf_by_name(source_p->user->auth_oper, CONF_OPERATOR)) ==
        NULL) {
      sendto_one(source_p, form_str(ERR_NOOPERHOST), me.name, parv[0]);
      log_foper(source_p, source_p->user->auth_oper);
      if (ConfigFileEntry.failed_oper_notice)
        sendto_realops_flags(UMODE_ALL, L_ALL, "%s failed attempt to exit the "
          "Matrix (Unrecognized Host)", get_client_name(source_p, HIDE_IP,
          YES));
      return;
    }
    ptr = source_p->localClient->confs.head;
    oconf = ptr->data;
    detach_conf(source_p, oconf);
    if (attach_conf(source_p, aconf) != 0) {
      sendto_one(source_p, ":%s NOTICE %s :Failed to attach configuration!",
        me.name, parv[0]);
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s failed attempt to exit the "
        "Matrix (Could not attach configuration!)", get_client_name(source_p,
        HIDE_IP, YES));
      log_foper(source_p, source_p->user->auth_oper);
      attach_conf(source_p, oconf);
      return;
    }
    oper_up(source_p, aconf);
    ilog(L_TRACE, "OPER %s (Public Key) by %s", source_p->user->auth_oper,
      log_client_name(source_p));
    log_oper(source_p, source_p->user->auth_oper);
    MyFree(source_p->user->response);
    MyFree(source_p->user->auth_oper);
    source_p->user->response = NULL;
    source_p->user->auth_oper = NULL;
    return;
  }
  MyFree(source_p->user->response);
  MyFree(source_p->user->auth_oper);
  source_p->user->response = NULL;
  source_p->user->auth_oper = NULL;
  if (!(aconf = find_conf_exact(parv[1], source_p->username, source_p->host,
      CONF_OPERATOR)) && !(aconf = find_conf_exact(parv[1], source_p->username,
      source_p->localClient->sockhost, CONF_OPERATOR))) {
    sendto_one(source_p, form_str(ERR_NOOPERHOST), me.name, parv[0]);
    log_foper(source_p, parv[1]);
    if (ConfigFileEntry.failed_oper_notice)
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s failed attempt to exit the "
        "Matrix (Unrecognized Host)", get_client_name(source_p, HIDE_IP, YES));
      return;
  }
  if (!aconf->rsa_public_key) {
    sendto_one(source_p, ":%s NOTICE %s :Public Key Challenge Authentication "
      "is disabled for your configuration.", me.name, parv[0]);
    return;
  }
  if (!generate_challenge(&challenge, &(source_p->user->response),
      aconf->rsa_public_key)) {
    sendto_one(source_p, form_str(RPL_RSACHALLENGE), me.name, parv[0],
      challenge);
  }
  DupString(source_p->user->auth_oper, aconf->name);
  MyFree(challenge);
  return;
}
#else
/*
 * m_challenge - if not compiled with HAVE_LIBCRYPTO,
 *               display a notice stating that challenge
 *               authentication is disabled.
 * parv[0] = sender prefix
 */
static void m_challenge(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  sendto_one(source_p, ":%s NOTICE %s :Public Key Challenge Authentication "
    "is not supported by this server.", me.name, parv[0]);
  return;
}
#endif /* HAVE_LIBCRYPTO */

static void mo_challenge(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  sendto_one(source_p, ":%s NOTICE %s :You are already an IRC Operator.",
    me.name, parv[0]);
  sendto_one(source_p, ":%s 381 %s :You are now an IRC Operator.", me.name,
    parv[0]);
  return;
}
