/* m_connect.c: Connects to a remote IRC server.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "ircd.h"
#include "irc_string.h"
#include "numeric.h"
#include "fdlist.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "hash.h"
#include "modules.h"

static void mo_connect(struct Client *, struct Client *, int, char **);
static void ms_connect(struct Client *, struct Client *, int, char **);

struct Message connect_msgtab = {
  "CONNECT", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, ms_connect, m_ignore, mo_connect}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&connect_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&connect_msgtab);
}
const char *_version = "1.57.2.2";
#endif

/* mo_connect - CONNECT command handler
 * by jto 11 Feb 1989
 *
 * m_connect
 *  parv[0] = sender prefix
 *  parv[1] = servername
 *  parv[2] = port number
 *  parv[3] = remote server
 */
static void mo_connect(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  int port;
  int tmpport;
  struct ConfItem *aconf;
  struct Client *target_p;
  /* always privileged with handlers */
  if (MyConnect(source_p) && !IsOperRemote(source_p) && parc > 3) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to perform remote server connections.", me.name, parv[0]);
    return;
  }
  if (hunt_server(client_p, source_p, ":%s CONNECT %s %s :%s", 3, parc, parv) !=
      HUNTED_ISME)
    return;

  if (*parv[1] == '\0') {
    sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
      "CONNECT");
    return;
  }
  if ((target_p = find_server(parv[1]))) {
    sendto_one(source_p, ":%s NOTICE %s :Server %s already exists from %s",
      me.name, parv[0], parv[1], target_p->from->name);
    return;
  }
  /* try to find the name, then host, if both fail notify ops and bail */
  if (!(aconf = find_conf_by_name(parv[1], CONF_SERVER))) {
    if (!(aconf = find_conf_by_host(parv[1], CONF_SERVER))) {
      sendto_one(source_p, ":%s NOTICE %s :Server %s is unknown", me.name,
        parv[0], parv[1]);
      return;
    }
  }
  s_assert(0 != aconf);
  /* Find port number from user, configuration or compiled default. */
  tmpport = port = aconf->port;
  if (parc > 2 && !EmptyString(parv[2])) {
    if ((port = atoi(parv[2])) <= 0) {
      sendto_one(source_p, ":%s NOTICE %s :Port number %s is invalid", me.name,
        parv[0], port);
      return;
    }
  }
  else if (port <= 0 && (port = PORTNUM) <= 0) {
    sendto_one(source_p, ":%s NOTICE %s :You must specify a port number",
      me.name, parv[0]);
    return;
  }
  ilog(L_TRACE, "CONNECT %s/%d by %s", parv[1], port,
    log_client_name(source_p));
  aconf->port = port;
  if (serv_connect(aconf, source_p)) {
#ifndef HIDE_SERVERS_IPS
    if (IsOperAdmin(source_p))
      sendto_one(source_p, ":%s NOTICE %s :Connecting to \002%s\002(%s/%d)",
        me.name, parv[0], aconf->name, aconf->host, aconf->port);
    else
#endif
      sendto_one(source_p, ":%s NOTICE %s :Connecting to \002%s\002(255.255"
        ".255.255/%d)", me.name, parv[0], aconf->name, aconf->port);
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Connection to \002%s\002(255.255"
      ".255.255/%d) failed", me.name, parv[0], aconf->name, aconf->port);
  /* client is either connecting with all data it needs or has been destroyed */
  aconf->port = tmpport;
}

/* ms_connect - CONNECT command handler
 *  by jto 11 Feb 1989
 *
 *  parv[0] = sender prefix
 *  parv[1] = servername
 *  parv[2] = port number
 *  parv[3] = remote server
 */
static void ms_connect(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  int port;
  int tmpport;
  struct ConfItem *aconf;
  struct Client *target_p;
  if (hunt_server(client_p, source_p, ":%s CONNECT %s %s :%s", 3, parc, parv) !=
      HUNTED_ISME)
    return;

  if (*parv[1] == '\0') {
    sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
      "CONNECT");
    return;
  }
  if ((target_p = find_server(parv[1]))) {
    sendto_one(source_p, ":%s NOTICE %s :Server %s already exists from %s",
      me.name, parv[0], parv[1], target_p->from->name);
    return;
  }
  /* try to find the name, then host, if both fail notify ops and bail */
  if (!(aconf = find_conf_by_name(parv[1], CONF_SERVER))) {
    if (!(aconf = find_conf_by_host(parv[1], CONF_SERVER))) {
      sendto_one(source_p, ":%s NOTICE %s :Server %s is unknown", me.name,
        parv[0], parv[1]);
      return;
    }
  }
  s_assert(0 != aconf);
  /* Get port number from user, if given. If not specified, use the default
   * form configuration structure. If missing from there, then use the
   * precompiled default.
   */
  tmpport = port = aconf->port;
  if (parc > 2 && !EmptyString(parv[2])) {
    port = atoi(parv[2]);
    /* if someone sends port 0, and we have a config port.. use it */
    if (port == 0 && aconf->port)
      port = aconf->port;
    else if (port <= 0) {
      sendto_one(source_p, ":%s NOTICE %s :Port number %s is invalid", me.name,
        parv[0], port);
      return;
    }
  }
  else if (port <= 0 && (port = PORTNUM) <= 0) {
    sendto_one(source_p, ":%s NOTICE %s :You must specify a port number",
      me.name, parv[0]);
    return;
  }
  /* Notify all operators about remote connect requests */
  sendto_wallops_flags(UMODE_WALLOP, &me, "%s is remotely requesting "
    "connection to \002%s\002(255.255.255.255/%d)", get_oper_name(source_p),
    aconf->name, port);
  sendto_server(NULL, NULL, NOCAPS, NOCAPS, ":%s WALLOPS :%s is remotely "
    "requesting connection to \002%s\002(255.255.255.255/%d)", me.name,
    get_oper_name(source_p), aconf->name, port);
  ilog(L_TRACE, "CONNECT %s/%d by %s", parv[1], port,
    log_client_name(source_p));
  aconf->port = port;

  /* we have all necessary connection information */
  if (serv_connect(aconf, source_p))
    sendto_one(source_p, ":%s NOTICE %s :Connecting to "
      "\002%s\002(255.255.255.255/%d)", me.name, parv[0], aconf->name,
      aconf->port);
  else
    sendto_one(source_p, ":%s NOTICE %s :Connection to "
      "\002%s\002(255.255.255.255/%d) failed", me.name, parv[0], aconf->name,
      aconf->port);
  /* client either has all info it needs or has been destroyed */
  aconf->port = tmpport;
}
