/* m_cryptlink.c: Used to negotiate an encrypted link.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "ircd.h"
#include "modules.h"
#include "numeric.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "irc_string.h"
#include "tools.h"
#include "memory.h"
#include "common.h"
#include "event.h"
#include "hash.h"
#include "s_conf.h"
#include "s_log.h" 
#include "s_serv.h"
#include "s_stats.h"
#include "scache.h"
#include "motd.h"

#ifdef HAVE_LIBCRYPTO
#include <openssl/rsa.h>
#include "rsa.h"

static int bogus_host(char *host);
static char *parse_cryptserv_args(struct Client * client_p, char *parv[],
  int parc, char *info, char *key);
static void mr_cryptlink(struct Client *, struct Client *, int, char **);
static void cryptlink_serv(struct Client *, struct Client *, int, char **);
static void cryptlink_auth(struct Client *, struct Client *, int, char **);

struct Message cryptlink_msgtab = {
  "CRYPTLINK", 0, 0, 4, 0, MFLG_SLOW | MFLG_UNREG, 0,
  {mr_cryptlink, m_ignore, m_error, m_ignore, m_ignore}
};

struct CryptLinkStruct {
  const char *cmd;    /* CRYPTLINK <command> to match */
  void (*handler) (); /* Function to call */
};
static struct CryptLinkStruct cryptlink_cmd_table[] = {
  /* command  function */
  {"AUTH", cryptlink_auth },
  {"SERV", cryptlink_serv },
  /* End of table */
  {(char *)0, (void (*) ())0,}
};
#endif

#ifndef STATIC_MODULES
void _modinit(void)
{
#ifdef HAVE_LIBCRYPTO
  mod_add_cmd(&cryptlink_msgtab);
#endif
}
void _moddeinit(void)
{
#ifdef HAVE_LIBCRYPTO
  mod_del_cmd(&cryptlink_msgtab);
#endif
}
const char *_version = "1.56.2.2";
#endif

#ifdef HAVE_LIBCRYPTO
/* mr_cryptlink - CRYPTLINK message handler
 *  parv[0] == CRYPTLINK
 *  parv[1] = command (SERV, AUTH)
 *  parv[2] = Parameters specific to each command (parv[1]):
 * SERV - parc must be >= 5
 *  parv[0] == CRYPTLINK
 *  parv[1] == SERV
 *  parv[2] == server name
 *  parv[3] == keyphrase
 *  parv[4] == :server info (M-line)
 * AUTH - parc must be >= 4
 *  parv[0] == CRYPTLINK
 *  parv[1] == AUTH
 *  parv[2] == cipher (eg. BF/256)
 *  parv[3] == keyphrase
 */
static void mr_cryptlink(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  int i;
  for (i = 0; cryptlink_cmd_table[i].handler; i++) {
    /* Traverse through the command table */
    if (!irccmp(cryptlink_cmd_table[i].cmd, parv[1])) {
      /* Match found.  Time to execute the function */
      cryptlink_cmd_table[i].handler(client_p, source_p, parc, parv);
    }
  }
  return;
}

/* cryptlink_auth - CRYPTLINK AUTH message handler
 *  parv[1] = secret key
 */
static void cryptlink_auth(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  struct EncCapability *ecap;
  struct ConfItem *aconf;
  int enc_len;
  int len;
  char *enc;
  char *key;
  if (parc < 4) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Invalid authorization parameters "
      "from %s, link cancelled", get_client_name(client_p, SHOW_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Invalid authorization parameters "
      "from %s, link cancelled", get_client_name(client_p, MASK_IP, YES));
    ilog(L_NOTICE, "Access denied, %s sent invalid authorization parameters",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Invalid authorization parameters");
    return;
  }
  if (!IsWaitAuth(client_p))
    return;
  for (ecap = CipherTable; ecap->name; ecap++) {
    if ((!irccmp(ecap->name, parv[2])) && (IsCapableEnc(client_p, ecap->cap))) {
      client_p->localClient->in_cipher = ecap;
      break;
    }
  }
  if (!client_p->localClient->in_cipher) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Unsupported encryption algorithm "
      "requested from %s, link cancelled", get_client_name(client_p, SHOW_IP,
      YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Unsupported encryption algorithm "
      "requested from %s, link cancelled", get_client_name(client_p, MASK_IP,
      YES));
    ilog(L_NOTICE, "Access denied, %s requested invalid cipher",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Unsupported encryption algorithm "
      "requested");
    return;
  }
  if (!(enc_len = unbase64_block(&enc, parv[3], strlen(parv[3])))) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "base64 decoding failed on "
      "keyphrase from %s, link cancelled", get_client_name(client_p, SHOW_IP,
      YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "base64 decoding failed on "
      "keyphrase from %s, link cancelled", get_client_name(client_p, MASK_IP,
      YES));
    ilog(L_NOTICE, "Access denied, %s sent invalid base64 keyphrase",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Failed to decode keyphrase");
    return;
  }
  if (verify_private_key() == NO) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "RSA private key validation "
      "failed, check the logs!");
    exit_client(client_p, client_p, &me, "Invalid authorization parameters");
    return;
  }

  key = MyMalloc(RSA_size(ServerInfo.rsa_private_key));
  len = RSA_private_decrypt(enc_len, (unsigned char *)enc, (unsigned char *)key,
    ServerInfo.rsa_private_key, RSA_PKCS1_PADDING);
  if (len < client_p->localClient->in_cipher->keylen) {
    report_crypto_errors();
    if (len < 0) {
      sendto_realops_flags(UMODE_ALL, L_ADMIN, "Decryption of data from %s "
        "failed, link cancelled", get_client_name(client_p, SHOW_IP, YES));
      sendto_realops_flags(UMODE_ALL, L_OPER, "Decryption of data from %s "
        "failed, link cancelled", get_client_name(client_p, MASK_IP, YES));
      ilog(L_NOTICE, "Access denied, corrupted data from %s",
        log_client_name(client_p));
      exit_client(client_p, client_p, &me, "Data decryption failed");
    }
    else {
      sendto_realops_flags(UMODE_ALL, L_ADMIN, "Invalid authorization "
        "parameters from %s, link cancelled", get_client_name(client_p, SHOW_IP,
        YES));
      sendto_realops_flags(UMODE_ALL, L_OPER, "Invalid authorization "
        "parameters from %s, link cancelled", get_client_name(client_p, MASK_IP,
        YES));
      ilog(L_NOTICE, "Access denied, %s sent invalid authorization parameters",
        log_client_name(client_p));
      exit_client(client_p, client_p, &me, "Invalid authorization parameters");
    }
    MyFree(enc);
    MyFree(key);
    return;
  }
  if (memcmp(key, client_p->localClient->in_key,
      client_p->localClient->in_cipher->keylen) != 0) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Unauthorised server connection "
      "attempt from %s: Invalid key", get_client_name(client_p, SHOW_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Unauthorised server connection "
      "attempt from %s: Invalid key", get_client_name(client_p, MASK_IP, YES));
    ilog(L_NOTICE, "Access denied, invalid key for server %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Invalid password");
    return;
  }
  aconf = find_conf_name(&client_p->localClient->confs, client_p->name,
    CONF_SERVER);
  if (!aconf) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Unauthorised server connection "
      "attempt from %s: Lost connect configuration", get_client_name(client_p,
      SHOW_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Unauthorised server connection "
      "attempt from %s: Lost connect configuration", get_client_name(client_p,
      MASK_IP, YES));
    ilog(L_NOTICE, "Access denied, lost connect block for server %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Lost connect configuration");
    return;
  }
  if (!(client_p->localClient->out_cipher ||
      (client_p->localClient->out_cipher = check_cipher(client_p, aconf)))) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "No available encryption "
      "algorithms are compatible with server %s, link cancelled",
      get_client_name(client_p, SHOW_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "No available encryption "
      "algorithms are compatible with server %s, link cancelled",
      get_client_name(client_p, MASK_IP, YES));
    ilog(L_NOTICE, "Access denied, no ciphers compatible with %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "No compatible encryption "
      "algorithms found");
    return;
  }
  /* set hopcount */
  client_p->hopcount = 1;
  SetCryptIn(client_p);
  ClearWaitAuth(client_p);
  server_estab(client_p);
}

/* cryptlink_serv - CRYPTLINK SERV message handler
 *  parv[0] == CRYPTLINK
 *  parv[1] == SERV
 *  parv[2] == server name
 *  parv[3] == keyphrase
 *  parv[4] == :server info (M-line)
 */
static void cryptlink_serv(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  char info[REALLEN + 1];
  char *name;
  struct Client *target_p;
  char *key = client_p->localClient->out_key;
  char *b64_key;
  struct ConfItem *aconf;
  char *encrypted;
  const char *p;
  int enc_len;
  /* if (client_p->name[0] != 0) return; */
  if ((parc < 5) || (*parv[4] == '\0') || ((name = 
      parse_cryptserv_args(client_p, parv, parc, info, key)) == NULL)) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Invalid authorization "
      "parameters from %s, link cancelled", get_client_name(client_p, SHOW_IP,
      YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Invalid authorization "
      "parameters from %s, link cancelled", get_client_name(client_p, MASK_IP,
      YES));
    ilog(L_NOTICE, "Access denied, %s sent invalid authorization parameters",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Invalid authorization parameters");
    return;
  }
  /* CRYPTLINK SERV support (TS support) */
  client_p->tsinfo = TS_DOESTS;
  if (bogus_host(name)) {
    exit_client(client_p, client_p, client_p, "Invalid server name");
    return;
  }
  /* Call check_server to check everything */
  switch (check_server(name, client_p, CHECK_SERVER_CRYPTLINK)) {
   case -1:
    if (ConfigFileEntry.warn_no_nline) {
      sendto_realops_flags(UMODE_ALL, L_ADMIN, "Unauthorised server connection "
        "attempt from %s: No entry for servername %s", get_client_name(client_p,
        SHOW_IP, YES), name);
      sendto_realops_flags(UMODE_ALL, L_OPER, "Unauthorised server connection "
        "attempt from %s: No entry for servername %s", get_client_name(client_p,
        MASK_IP, YES), name);
      ilog(L_NOTICE, "Access denied, no connect block for server %s",
        log_client_name(client_p));
    }
    exit_client(client_p, client_p, client_p, "Invalid server name");
    return;
    break;
   case -2:
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Unauthorised server connection "
      "attempt from %s: Data encryption not permitted for this configuration",
      get_client_name(client_p, SHOW_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Unauthorised server connection "
      "attempt from %s: Data encryption not permitted for this configuration",
      get_client_name(client_p, MASK_IP, YES));
    ilog(L_NOTICE, "Access denied, %s sent invalid authorization parameters",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Data encryption is not supported "
      "for your connection");
    return;
    break;
   case -3:
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Unauthorised server connection "
      "attempt from %s: Invalid host for server %s", get_client_name(client_p,
      SHOW_IP, YES), name);
    sendto_realops_flags(UMODE_ALL, L_OPER, "Unauthorised server connection "
      "attempt from %s: Invalid host for server %s", get_client_name(client_p,
      MASK_IP, YES), name);
    ilog(L_NOTICE, "Access denied, invalid host for server %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Invalid hostname");
    return;
    break;
  }
  if ((target_p = find_server(name))) {
    /* Reject the new server, it's from an unregistered connection */
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Attempt to re-introduce server "
      "%s from %s", name, get_client_name(client_p, SHOW_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Attempt to re-introduce server "
      "%s from %s", name, get_client_name(client_p, MASK_IP, YES));
    exit_client(client_p, client_p, &me, "Server Exists");
    return;
  }
  aconf = find_conf_name(&client_p->localClient->confs, name, CONF_SERVER);
  if (!aconf) {
      sendto_realops_flags(UMODE_ALL, L_ADMIN, "Lost connect configuration for "
        "%s on %s, closing link", get_client_name(client_p, SHOW_IP, YES),
        name);
      sendto_realops_flags(UMODE_ALL, L_OPER, "Lost connect configuration for "
        "%s on %s, closing link", get_client_name(client_p, MASK_IP, YES),
        name);
      ilog(L_NOTICE, "Lost connect configuration for %s on %s",
        log_client_name(client_p), name);
      exit_client(client_p, client_p, &me, "Lost connect configuration");
    return;
  }
  /* if we are connecting, we have the name from aconf */
  strlcpy(client_p->name, name, sizeof(client_p->name));
  p = info;
  if (!strncmp(info, "(H)", 3)) {
    SetHidden(client_p);
    if ((p = strchr(info, ' '))) {
      p++;
      if (*p == '\0')
        p = "(Unknown Location)";
    }
    else
      p = "(Unknown Location)";
  }
  strlcpy(client_p->info, p, sizeof(client_p->info));
  client_p->hopcount = 0;
  if (!(client_p->localClient->out_cipher ||
      (client_p->localClient->out_cipher = check_cipher(client_p, aconf)))) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "No available encryption "
      "algorithms are compatible with server %s, link cancelled",
      get_client_name(client_p, SHOW_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "No available encryption "
      "algorithms are compatible with server %s, link cancelled",
      get_client_name(client_p, MASK_IP, YES));
    ilog(L_NOTICE, "Access denied, no ciphers compatible with %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "No compatible encryption "
      "algorithms found");
    return;
  }
  encrypted = MyMalloc(RSA_size(ServerInfo.rsa_private_key));
  enc_len = RSA_public_encrypt(client_p->localClient->out_cipher->keylen,
    (unsigned char *)key,
    (unsigned char *)encrypted,
    aconf->rsa_public_key, RSA_PKCS1_PADDING);
  if (enc_len <= 0) {
    report_crypto_errors();
    MyFree(encrypted);
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Data encryption failure with "
       "server %s, link cancelled", get_client_name(client_p, SHOW_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Data encryption failure with "
       "server %s, link cancelled", get_client_name(client_p, MASK_IP, YES));
    ilog(L_NOTICE, "Data encryption failure for server %s, link cancelled",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "Data encryption failed");
    return;
  }
  base64_block(&b64_key, encrypted, enc_len);
  MyFree(encrypted);
  if (!IsWaitAuth(client_p))
    cryptlink_init(client_p, aconf, -1);

  sendto_one(client_p, "CRYPTLINK AUTH %s %s",
    client_p->localClient->out_cipher->name, b64_key);
  SetCryptOut(client_p);
  MyFree(b64_key);
}

/* parse_cryptserv_args
 * inputs       - parv parameters
 *              - parc count
 *              - info string (to be filled in by this routine)
 *              - key (to be filled in by this routine)
 * output       - NULL if invalid params, server name otherwise
 * side effects - parv[2] is trimmed to HOSTLEN size if needed.
 */
static char * parse_cryptserv_args(struct Client * client_p, char *parv[],
  int parc, char *info, char *key)
{
  char *name;
  unsigned char *tmp, *out;
  char **tmpx = (void *)&tmp;
  int len;
  int decoded_len;
  info[0] = '\0';
  name = parv[2];
  /* parv[2] contains encrypted auth data */
  if (!(decoded_len = unbase64_block(tmpx, parv[3], strlen(parv[3])))) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Unauthorised server connection "
      "attempt from %s: base64 key decoding failed", get_client_name(client_p,
      HIDE_IP, YES));
    sendto_realops_flags(UMODE_ALL, L_OPER, "Unauthorised server connection "
      "attempt from %s: base64 key decoding failed", get_client_name(client_p,
      MASK_IP, YES));
    ilog(L_NOTICE, "Access denied, could not base64 decode key for server %s",
      log_client_name(client_p));
    exit_client(client_p, client_p, &me, "base64 key decoding failed");
    return NULL;
  }
  if (verify_private_key() == NO) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "RSA private key validation "
      "failed, check the logs!");
    exit_client(client_p, client_p, &me, "Invalid authorization parameters");
    return NULL;
  }
  out = MyMalloc(RSA_size(ServerInfo.rsa_private_key));
  len = RSA_private_decrypt(decoded_len, tmp, out,
    ServerInfo.rsa_private_key, RSA_PKCS1_PADDING);
  MyFree(tmp);
  if (len < CIPHERKEYLEN) {
    report_crypto_errors();
    if (len < 0) {
      sendto_realops_flags(UMODE_ALL, L_ADMIN, "Data decryption failure with "
         "server %s, link cancelled", get_client_name(client_p, SHOW_IP, YES));
      sendto_realops_flags(UMODE_ALL, L_OPER, "Data decryption failure with "
         "server %s, link cancelled", get_client_name(client_p, MASK_IP, YES));
      ilog(L_NOTICE, "Data decryption failure for server %s, link cancelled",
        log_client_name(client_p));
      exit_client(client_p, client_p, &me, "Data decryption failed");
    }
    else {
      sendto_realops_flags(UMODE_ALL, L_ADMIN, "Data decryption failure with "
         "server %s, link cancelled", get_client_name(client_p, SHOW_IP, YES));
      sendto_realops_flags(UMODE_ALL, L_OPER, "Data decryption failure with "
         "server %s, link cancelled", get_client_name(client_p, MASK_IP, YES));
      ilog(L_NOTICE, "Data decryption failure for server %s, link cancelled",
        log_client_name(client_p));
      exit_client(client_p, client_p, &me, "Data decryption failed");
    }
    MyFree(out);
    return NULL;
  }
  memcpy(key, out, CIPHERKEYLEN);
  MyFree(out);
  strlcpy(info, parv[4], REALLEN + 1);
  if (strlen(name) > HOSTLEN)
    name[HOSTLEN] = '\0';
  return name;
}

/* bogus_host
 * inputs        - hostname
 * output        - 1 if a bogus hostname input, 0 if its valid
 */
static int bogus_host(char *host)
{
  int bogus_server = 0;
  int dots = 0;
  char *s;
  for (s = host; *s; s++) {
    if (!IsServChar(*s)) {
      bogus_server = 1;
      break;
    }
    if ('.' == *s) {
      ++dots;
    }
  }
  if ((!dots) || (bogus_server))
    return YES;
  return NO;
}
#endif /* HAVE_LIBCRYPTO */
