/*
 * Copyright (C) 2002 Hybrid Development Team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *  1.Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  2.Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  3.The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 *  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 *  IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "common.h"
#include "ircd.h"
#include "irc_string.h"
#include "numeric.h"
#include "fdlist.h"
#include "hash.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "channel.h"
#include "channel_mode.h"

/* Define whether you want FORCEJOIN/FORCEPART commands available */
#define FORCEJOIN 1

static void mo_ojoin(struct Client * client_p, struct Client * source_p,
   int parc, char *parv[]);
static void mo_forcejoin(struct Client * client_p, struct Client * source_p,
   int parc, char *parv[]);
static void mo_forcepart(struct Client * client_p, struct Client * source_p,
   int parc, char *parv[]);

struct Message ojoin_msgtab = {
  "OJOIN", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_ignore, m_not_oper, mo_ojoin, m_ignore, mo_ojoin}
};
#ifdef FORCEJOIN
struct Message forcejoin_msgtab = {
  "FORCEJOIN", 0, 0, 3, 0, MFLG_SLOW, 0,
  {m_ignore, m_not_oper, mo_forcejoin, m_ignore, mo_forcejoin}
};
struct Message forcepart_msgtab = {
  "FORCEPART", 0, 0, 3, 0, MFLG_SLOW, 0,
  {m_ignore, m_not_oper, mo_forcepart, m_ignore, mo_forcepart}
};
#endif

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&ojoin_msgtab);
#ifdef FORCEJOIN
  mod_add_cmd(&forcejoin_msgtab);
  mod_add_cmd(&forcepart_msgtab);
#endif
}
void _moddeinit(void)
{
  mod_del_cmd(&ojoin_msgtab);
#ifdef FORCEJOIN
  mod_del_cmd(&forcejoin_msgtab);
  mod_del_cmd(&forcepart_msgtab);
#endif
}
const char *_version = "1.21.3.1";
#endif

/* m_ojoin - points to m_forcejoin */
static void mo_ojoin(struct Client * client_p, struct Client * source_p,
   int parc, char *parv[])
{
  /* adjust parv[] so nickname = sender */
  parv[2] = parv[1];
  parv[1] = parv[0];
  mo_forcejoin(client_p, source_p, parc, parv);
  return;
}
/* m_forcejoin/m_forcepart
 *  parv[0] = sender prefix
 *  parv[1] = user to force
 *  parv[2] = channel to force them into/out of
 */
static void mo_forcejoin(struct Client * client_p, struct Client * source_p,
   int parc, char *parv[])
{
  struct Client *target_p;
  struct Channel *chptr;
  int type;
  char mode;
  char sjmode;
  char *newch;
  if (!IsAdmin(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to force users to join channels", me.name, parv[0]);
    return;
  }
  if ((hunt_server(client_p, source_p, ":%s FORCEJOIN %s %s", 1, parc, parv)) !=
      HUNTED_ISME)
    return;
  if ((target_p = find_client(parv[1])) == NULL) {
    sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
    return;
  }
  if (!IsClient(target_p))
    return;
  /* select our modes from parv[2] if they exist... */
  if (*parv[2] == '@') {
    type = CHFL_CHANOP;
    mode = 'o';
    sjmode = '@';
  }
  else if (*parv[2] == '+') {
    type = CHFL_VOICE;
    mode = 'v';
    sjmode = '+';
  }
  else {
    type = CHFL_PEON;
    mode = sjmode = '\0';
  }
  if (mode != '\0')
    parv[2]++;
  if ((chptr = hash_find_channel(parv[2])) != NULL) {
    if (IsMember(target_p, chptr)) {
      sendto_one(source_p, form_str(ERR_USERONCHANNEL), me.name, parv[0],
        target_p->name, chptr->chname);
      return;
    }
    sendto_wallops_flags(UMODE_WALLOP, &me, "%s is forcing %s to join channel "
      "%s", get_oper_name(source_p), parv[1], parv[2]);
    ilog(L_NOTICE, "FORCEJOIN %s to %s by %s", parv[1], parv[2],
      log_client_name(source_p));
    add_user_to_channel(chptr, target_p, type);
    if (*chptr->chname != '&') {
      sendto_server(NULL, NULL, NOCAPS, NOCAPS, ":%s WALLOPS :%s is forcing %s "
        "to join channel %s", me.name, get_oper_name(source_p), parv[1],
        parv[2]);
      sendto_server(target_p, chptr, NOCAPS, NOCAPS, ":%s SJOIN %lu %s + "
        ":%c%s", me.name, (unsigned long)chptr->channelts, chptr->chname,
        type ? sjmode : ' ', target_p->name);
    }
    sendto_channel_local(ALL_MEMBERS, chptr, ":%s!%s@%s JOIN :%s",
      target_p->name, target_p->username, target_p->host, chptr->chname);
    if (type)
      sendto_channel_local(ALL_MEMBERS, chptr, ":%s MODE %s +%c %s", me.name,
        chptr->chname, mode, target_p->name);
    if (chptr->topic != NULL) {
      sendto_one(target_p, form_str(RPL_TOPIC), me.name, target_p->name,
        chptr->chname, chptr->topic);
      sendto_one(target_p, form_str(RPL_TOPICWHOTIME), me.name, parv[0],
        chptr->chname, chptr->topic_info, chptr->topic_time);
    }
    channel_member_names(chptr, target_p, 1);
  }
  else {
    newch = parv[2];
    /* channel name must begin with & or #, channel name cannot be & if local
     * channels are disabled
     */
    if (!check_channel_name(newch) || !IsChannelName(newch) ||
        (ConfigServerHide.disable_local_channels && (*newch == '&'))) {
      sendto_one(source_p, form_str(ERR_BADCHANNAME), me.name, parv[0],
        (unsigned char *)newch);
      return;
    }
    /* newch can't be longer than CHANNELLEN */
    if (strlen(newch) > CHANNELLEN) {
      sendto_one(source_p, ":%s NOTICE %s :Channel name is too long", me.name,
        parv[0]);
      return;
    }
    /* let the oper know that the channel is being created */
    sendto_one(source_p, ":%s NOTICE %s :*** Notice -- Creating channel %s",
      me.name, parv[0], newch);
    chptr = get_or_create_channel(target_p, newch, NULL);
    add_user_to_channel(chptr, target_p, CHFL_CHANOP);
    sendto_wallops_flags(UMODE_WALLOP, &me, "%s is forcing %s to join channel "
      "%s", get_oper_name(source_p), parv[1], parv[2]);
    ilog(L_NOTICE, "FORCEJOIN %s to %s by %s", parv[1], parv[2],
      log_client_name(source_p));
    if (*chptr->chname != '&') {
      sendto_server(NULL, NULL, NOCAPS, NOCAPS, ":%s WALLOPS :%s is forcing %s "
        "to join channel %s", me.name, get_oper_name(source_p), parv[1],
        parv[2]);
      /* send out a join, make target_p join chptr */
      sendto_server(target_p, chptr, NOCAPS, NOCAPS, ":%s SJOIN %lu %s +nt "
         ":@%s", me.name, (unsigned long)chptr->channelts, chptr->chname,
        target_p->name);
    }
    sendto_channel_local(ALL_MEMBERS, chptr, ":%s!%s@%s JOIN :%s",
      target_p->name, target_p->username, target_p->host, chptr->chname);
    chptr->mode.mode |= MODE_TOPICLIMIT;
    chptr->mode.mode |= MODE_NOPRIVMSGS;
    sendto_channel_local(ALL_MEMBERS, chptr, ":%s MODE %s +nt", me.name,
      chptr->chname);
    target_p->localClient->last_join_time = CurrentTime;
    channel_member_names(chptr, target_p, 1);
  }
}
static void mo_forcepart(struct Client * client_p, struct Client * source_p,
   int parc, char *parv[])
{
  struct Client *target_p;
  struct Channel *chptr;
  struct membership *msptr;
  if (!IsAdmin(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to force users to leave channels", me.name, parv[0]);
    return;
  }
  if ((hunt_server(client_p, source_p, ":%s FORCEPART %s %s", 1, parc, parv)) !=
      HUNTED_ISME)
    return;
  /* if target_p == NULL then let the oper know */
  if ((target_p = find_client(parv[1])) == NULL) {
    sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, parv[0], parv[1]);
    return;
  }
  if (!IsClient(target_p))
    return;
  if ((chptr = hash_find_channel(parv[2])) == NULL) {
    sendto_one(source_p, form_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
      parv[2]);
    return;
  }
  msptr = find_channel_membership(chptr, target_p);
  if (msptr == NULL) {
    sendto_one(source_p, form_str(ERR_USERNOTINCHANNEL), me.name, parv[0],
      parv[1], parv[2]);
    return;
  }
  sendto_wallops_flags(UMODE_WALLOP, &me, "%s is forcing %s to leave channel "
    "%s", get_oper_name(source_p), parv[1], parv[2]);
  ilog(L_NOTICE, "FORCEPART %s from %s by %s", parv[1], parv[2], 
    log_client_name(source_p));
  if (*chptr->chname != '&') {
    sendto_server(NULL, NULL, NOCAPS, NOCAPS, ":%s WALLOPS :%s is forcing %s "
      "to leave channel %s", me.name, get_oper_name(source_p), parv[1],
      parv[2]);
    sendto_server(target_p, chptr, NOCAPS, NOCAPS, ":%s PART %s :%s",
      target_p->name, chptr->chname, target_p->name);
  }
  sendto_channel_local(ALL_MEMBERS, chptr, ":%s!%s@%s PART %s :%s",
    target_p->name, target_p->username, target_p->host, chptr->chname,
    target_p->name);
  remove_user_from_channel(msptr);
}
