/* m_info.c: Sends information about the server.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "tools.h"
#include "m_info.h"
#include "channel.h"
#include "client.h"
#include "common.h"
#include "irc_string.h"
#include "ircd.h"
#include "hook.h"
#include "numeric.h"
#include "s_serv.h"
#include "s_user.h"
#include "send.h"
#include "s_conf.h"
#include "handlers.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"

static void send_conf_options(struct Client * source_p);
static void send_birthdate_online_time(struct Client * source_p);
static void send_info_text(struct Client * source_p);
static void m_info(struct Client *, struct Client *, int, char **);
static void ms_info(struct Client *, struct Client *, int, char **);
static void mo_info(struct Client *, struct Client *, int, char **);

struct Message info_msgtab = {
  "INFO", 0, 0, 0, 0, MFLG_SLOW, 0,
  {m_unregistered, m_info, ms_info, m_ignore, mo_info}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&info_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&info_msgtab);
}
const char *_version = "1.88.2.6";
#endif

struct InfoStruct {
  const char *name; /* Displayed variable name */
  unsigned int output_type; /* type definition */
  void *option; /* Pointer reference to the value */
   const char *desc; /* ASCII description of the variable */
};
/* Types for output_type in InfoStruct */
#define OUTPUT_STRING     0x0001 /* Output option as %s w/ dereference */
#define OUTPUT_STRING_PTR 0x0002 /* Output option as %s w/out deference */
#define OUTPUT_DECIMAL    0x0004 /* Output option as decimal (%d) */
#define OUTPUT_BOOLEAN    0x0008 /* Output option as "ON" or "OFF" */
#define OUTPUT_BOOLEAN_YN 0x0010 /* Output option as "YES" or "NO" */
#define OUTPUT_BOOLEAN2   0x0020 /* Output option as "YES/NO/MASKED" */

static struct InfoStruct info_table[] = {
  /* --[  START OF TABLE  ]-------------------------------------------- */
  {
    "anti_nick_flood",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.anti_nick_flood,
    "NICK flood protection"
  },
  {
    "anti_spam_exit_message_time",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.anti_spam_exit_message_time,
    "Duration a client must be connected for to have an exit message"
  },
  {
    "caller_id_wait",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.caller_id_wait,
    "Minimum delay between notifying UMODE +g users of messages"
  },
  {
    "client_exit",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.client_exit,
    "Prepend 'Client Exit:' to user QUIT messages"
  },
  {
    "client_flood",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.client_flood,
    "Number of lines before a client Excess Flood's"
  },
  {
    "connect_timeout",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.connect_timeout,
    "Connect timeout for connections to servers"
  },
  {
    "default_floodcount",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.default_floodcount,
    "Startup value of FLOODCOUNT",
  },
  {
    "disable_auth",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.disable_auth,
    "Controls whether auth checking is disabled or not"
  },
  {
    "dot_in_ip6_addr",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.dot_in_ip6_addr,
    "Suffix a . to ip6 addresses",
  },
  {
    "dots_in_ident",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.dots_in_ident,
    "Number of permissable dots in an ident"
  },
  {
    "failed_oper_notice",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.failed_oper_notice,
    "Inform opers if someone /oper's with the wrong password"
  },
  {
    "fname_operlog",
    OUTPUT_STRING_PTR,
    &ConfigFileEntry.fname_operlog,
    "Operator log file"
  },
  {
    "fname_foperlog",
    OUTPUT_STRING_PTR,
    &ConfigFileEntry.fname_foperlog,
    "Operator failure log file"
  },
  {
    "fname_userlog",
    OUTPUT_STRING_PTR,
    &ConfigFileEntry.fname_userlog,
    "User log file"
  },
  {
    "glines",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.glines,
    "G-Line (network-wide ban) support"
  },
  {
    "gline_time",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.gline_time,
    "Expiry time for G-Lines"
  },
  {
    "gline_min_cidr",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.gline_min_cidr,
    "Minimum CIDR bitlen of ipv4 glines"
  },
  {
    "gline_min_cidr6",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.gline_min_cidr6,
    "Minimum CIDR bitlen of ipv6 glines"
  },
  {
    "hide_error_messages",
    OUTPUT_BOOLEAN2,
    &ConfigFileEntry.hide_error_messages,
    "Hide ERROR messages from servers"
  },
  {
    "hub",
    OUTPUT_BOOLEAN_YN,
    &ServerInfo.hub,
    "Can link multiple servers"
  },
  {
    "idletime",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.idletime,
    "Number of minutes before a client is considered idle"
  },
  {
    "kline_with_connection_closed",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.kline_with_connection_closed,
    "Display quit comment as 'Connection closed'"
  },
  {
    "kline_with_reason",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.kline_with_reason,
    "Show K-Line reason when rejecting users"
  },
  {
    "kline_delay",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.kline_delay,
    "K-Line checking delay"
  },
  {
    "max_accept",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.max_accept,
    "Maximum nicknames on accept list"
  },
  {
    "max_nick_changes",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.max_nick_changes,
    "Maximum nickname changes per interval"
  },
  {
    "max_nick_time",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.max_nick_time,
    "Nickname counter reset interval"
  },
  {
    "max_targets",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.max_targets,
    "Maximum message targets"
  },
  {
    "min_nonwildcard",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.min_nonwildcard,
    "Minimum non-wildcard chars in K/G lines"
  },
  {
    "min_nonwildcard_simple",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.min_nonwildcard_simple,
    "Minimum non-wildcard chars in X lines/resv's"
  },
  {
    "mask",
    OUTPUT_STRING,
    &ServerInfo.mask,
    "Masked server name"
  },
  {
    "network_name",
    OUTPUT_STRING,
    &ServerInfo.network_name,
    "Network name"
  },
  {
    "network_desc",
    OUTPUT_STRING,
    &ServerInfo.network_desc,
    "Network description"
  },
  {
    "no_oper_flood",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.no_oper_flood,
    "Disable flood control for operators"
  },
  {
    "non_redundant_klines",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.non_redundant_klines,
    "Check for and disallow redundant K-lines"
  },
  {
    "pace_wait",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.pace_wait,
    "Minimum delay between intensive commands"
  },
  {
    "pace_wait_simple",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.pace_wait_simple,
    "Minimum delay between less intensive commands"
  },
  {
    "ping_cookie",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.ping_cookie,
    "Require ping cookies to connect"
  },
  {
    "reject_after_count",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.reject_after_count,
    "Client rejection threshold setting"
  },
  {
    "reject_ban_time",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.reject_ban_time,
    "Client rejection time interval"
  },
  {
    "reject_duration",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.reject_duration,
    "Cached rejection lifetime"
  },
  {
    "short_motd",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.short_motd,
    "Do not show MOTD; only tell clients they should read it"
  },
  {
    "stats_c_oper_only",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.stats_c_oper_only,
    "Restrict STATS C to opers"
  },
  {
    "stats_h_oper_only",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.stats_h_oper_only,
    "Restrict STATS H to opers"
  },
  {
    "stats_i_oper_only",
    OUTPUT_BOOLEAN2,
    &ConfigFileEntry.stats_i_oper_only,
    "Restrict STATS I to opers"
  },
  {
    "stats_k_oper_only",
    OUTPUT_BOOLEAN2,
    &ConfigFileEntry.stats_k_oper_only,
    "Restrict STATS K to opers"
  },
  {
    "stats_o_oper_only",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.stats_o_oper_only,
    "Restrict STATS O to opers"
  },
  {
    "stats_P_oper_only",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.stats_P_oper_only,
    "Restrict STATS P to opers"
  },
  {
    "stats_y_oper_only",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.stats_y_oper_only,
    "Restrict STATS Y to opers"
  },
  {
    "map_oper_only",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.map_oper_only,
    "Restrict MAP to opers"
  },
  {
    "tkline_expire_notices",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.tkline_expire_notices,
    "TKLine expiry notices for opers"
  },
  {
    "ts_max_delta",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.ts_max_delta,
    "TS delta threshold"
  },
  {
    "ts_warn_delta",
    OUTPUT_DECIMAL,
    &ConfigFileEntry.ts_warn_delta,
    "TS delta warn trigger"
  },
  {
    "use_help",
    OUTPUT_BOOLEAN_YN,
    &ConfigFileEntry.use_help,
    "Restrict HELP to operators"
  },
  {
    "warn_no_nline",
    OUTPUT_BOOLEAN,
    &ConfigFileEntry.warn_no_nline,
    "Warn on unknown server links"
  },
  {
    "default_split_server_count",
    OUTPUT_DECIMAL,
    &ConfigChannel.default_split_server_count,
    "Startup value of SPLITNUM"
  },
  {
    "default_split_user_count",
    OUTPUT_DECIMAL,
    &ConfigChannel.default_split_user_count,
    "Startup value of SPLITUSERS"
  },
  {
    "knock_delay",
    OUTPUT_DECIMAL,
    &ConfigChannel.knock_delay,
    "Delay between a users KNOCK attempts"
  },
  {
    "knock_delay_channel",
    OUTPUT_DECIMAL,
    &ConfigChannel.knock_delay_channel,
    "Delay between KNOCK attempts to a channel"
  },
  {
    "max_bans",
    OUTPUT_DECIMAL,
    &ConfigChannel.max_bans,
    "Total +b/e/I modes allowed per channel"
  },
  {
    "max_chans_per_user",
    OUTPUT_DECIMAL,
    &ConfigChannel.max_chans_per_user,
    "Maximum channels per client"
  },
  {
    "no_create_on_split",
    OUTPUT_BOOLEAN_YN,
    &ConfigChannel.no_create_on_split,
    "Disallow creation of channels when split"
  },
  {
    "no_join_on_split",
    OUTPUT_BOOLEAN_YN,
    &ConfigChannel.no_join_on_split,
    "Disallow joining channels when split"
  },
  {
    "no_oper_resvs",
    OUTPUT_BOOLEAN_YN,
    &ConfigChannel.no_oper_resvs,
    "Disable channel resvs for operators"
  },
  {
    "quiet_on_ban",
    OUTPUT_BOOLEAN_YN,
    &ConfigChannel.quiet_on_ban,
    "Banned users may not send text to a channel"
  },
  {
    "use_except",
    OUTPUT_BOOLEAN_YN,
    &ConfigChannel.use_except,
    "Enable channel mode +e"
  },
  {
    "use_invex",
    OUTPUT_BOOLEAN_YN,
    &ConfigChannel.use_invex,
    "Enable channel mode +I"
  },
  {
    "use_knock",
    OUTPUT_BOOLEAN_YN,
    &ConfigChannel.use_knock,
    "Enable KNOCK"
  },
  {
    "disable_hidden",
    OUTPUT_BOOLEAN_YN,
    &ConfigServerHide.disable_hidden,
    "Disable server hiding"
  },
  {
    "disable_local_channels",
    OUTPUT_BOOLEAN_YN,
    &ConfigServerHide.disable_local_channels,
    "Disable &channels"
  },
  {
    "disable_remote_commands",
    OUTPUT_BOOLEAN_YN,
    &ConfigServerHide.disable_remote,
    "Prevent users issuing commands on remote servers"
  },
  {
    "flatten_links",
    OUTPUT_BOOLEAN_YN,
    &ConfigServerHide.flatten_links,
    "Flatten LINKS list"
  },
  {
    "hidden",
    OUTPUT_BOOLEAN_YN,
    &ConfigServerHide.hidden,
    "Hide this server from LINKS"
  },
  {
    "hide_servers",
    OUTPUT_BOOLEAN_YN,
    &ConfigServerHide.hide_servers,
    "Hide servernames from users"
  },
  {
    "links_delay",
    OUTPUT_DECIMAL,
    &ConfigServerHide.links_delay,
    "Links rehash delay"
  },
  /* --[  END OF TABLE  ]---------------------------------------------- */
  {
    (char *)0,
    (unsigned int)0,
    (void *)0,
    (char *)0
  }
};

/* m_info
 *  parv[0] = sender prefix
 *  parv[1] = servername
 */
static void m_info(struct Client * client_p, struct Client * source_p, int parc,
  char *parv[])
{
  static time_t last_used = 0L;
  if ((last_used + ConfigFileEntry.pace_wait) > CurrentTime) {
    /* safe enough to give this on a local connect only */
    sendto_one(source_p, form_str(RPL_LOAD2HI), me.name, parv[0], "INFO");
    sendto_one(source_p, form_str(RPL_ENDOFINFO), me.name, parv[0]);
    return;
  }
  else
    last_used = CurrentTime;

  if (!ConfigServerHide.disable_remote) {
    if (hunt_server(client_p, source_p, ":%s INFO :%s", 1, parc, parv) !=
        HUNTED_ISME)
      return;
  }
  hook_call_event(h_doing_info_id, source_p);
  send_info_text(source_p);
  send_birthdate_online_time(source_p);
}

/* mo_info
 *  parv[0] = sender prefix
 *  parv[1] = servername
 */
static void mo_info(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  if (hunt_server(client_p, source_p, ":%s INFO :%s", 1, parc, parv) ==
      HUNTED_ISME) {
    hook_call_event(h_doing_info_id, source_p);
    send_info_text(source_p);
    send_conf_options(source_p);
    send_birthdate_online_time(source_p);
  }
}

/* ms_info
 *  parv[0] = sender prefix
 *  parv[1] = servername
 */
static void ms_info(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  if (!IsClient(source_p))
    return;
  if (hunt_server(client_p, source_p, ":%s INFO :%s", 1, parc, parv) ==
      HUNTED_ISME) {
    hook_call_event(h_doing_info_id, source_p);
    send_info_text(source_p);
    /*  remote opers don't really need this */
    if (IsOper(source_p))
      send_conf_options(source_p);
    send_birthdate_online_time(source_p);
  }
}

/* send_info_text - send info text
 * input - client pointer to send info text to
 */
static void send_info_text(struct Client * source_p)
{
  char **text = infotext;
  while (*text)
    sendto_one(source_p, form_str(RPL_INFO), me.name, source_p->name, *text++);
  sendto_one(source_p, form_str(RPL_INFO), me.name, source_p->name, "");
}

/* send_birthdate_online_time - send birthdate and online time
 * input - client pointer to send to
 */
static void send_birthdate_online_time(struct Client * source_p)
{
  sendto_one(source_p, ":%s %d %s :Birth Date:   %s (compile #%s)", me.name,
    RPL_INFO, source_p->name, creation, generation);
  sendto_one(source_p, ":%s %d %s :Online since: %s", me.name, RPL_INFO,
    source_p->name, myctime(me.firsttime));
  sendto_one(source_p, form_str(RPL_ENDOFINFO), me.name, source_p->name);
}

/* send_conf_options - send selected configuration options
 * input - client pointer to send to
 */
static void send_conf_options(struct Client * source_p)
{
  Info *infoptr;
  int i = 0;

  sendto_one(source_p, form_str(RPL_INFO), me.name, source_p->name, "Server "
    "configuration options:");

  /* send them a list of configuration options */
  for (infoptr = MyInformation; infoptr->name; infoptr++) {
    if (infoptr->intvalue)
      sendto_one(source_p, ":%s %d %s :%-30s %-5d (%s)", me.name, RPL_INFO,
        source_p->name, infoptr->name, infoptr->intvalue, infoptr->desc);
    else
      sendto_one(source_p, ":%s %d %s :%-30s %-5s (%s)", me.name, RPL_INFO,
        source_p->name, infoptr->name, infoptr->strvalue, infoptr->desc);
  }
  /* Parse the info_table[] and do the magic. */
  for (i = 0; info_table[i].name; i++) {
    switch (info_table[i].output_type) {
     case OUTPUT_STRING:
     {
       char *option = *((char **)info_table[i].option);
       sendto_one(source_p, ":%s %d %s :%-30s %-5s (%s)", me.name, RPL_INFO,
         source_p->name, info_table[i].name, option ? option : "None",
         info_table[i].desc ? info_table[i].desc : "<none>");
       break;
     }
     case OUTPUT_STRING_PTR:
     {
       char *option = (char *)info_table[i].option;
       sendto_one(source_p, ":%s %d %s :%-30s %-5s (%s)", me.name, RPL_INFO,
         source_p->name, info_table[i].name, option ? option : "None",
         info_table[i].desc ? info_table[i].desc : "<none>");
       break;
     }
     case OUTPUT_DECIMAL:
     {
       int option = *((int *)info_table[i].option);
       sendto_one(source_p, ":%s %d %s :%-30s %-5d (%s)", me.name, RPL_INFO,
         source_p->name, info_table[i].name, option,
         info_table[i].desc ? info_table[i].desc : "<none>");
       break;
     }
     case OUTPUT_BOOLEAN:
     {
       int option = *((int *)info_table[i].option);
       sendto_one(source_p, ":%s %d %s :%-30s %-5s (%s)", me.name, RPL_INFO,
         source_p->name, info_table[i].name, option ? "On" : "Off",
         info_table[i].desc ? info_table[i].desc : "<none>");
       break;
     }
     case OUTPUT_BOOLEAN_YN:
     {
       int option = *((int *)info_table[i].option);
       sendto_one(source_p, ":%s %d %s :%-30s %-5s (%s)", me.name, RPL_INFO,
         source_p->name, info_table[i].name, option ? "Yes" : "No",
         info_table[i].desc ? info_table[i].desc : "<none>");
       break;
     }
     case OUTPUT_BOOLEAN2:
     {
       int option = *((int *)info_table[i].option);
       sendto_one(source_p, ":%s %d %s :%-30s %-5s (%s)", me.name, RPL_INFO,
         source_p->name, info_table[i].name, option ? ((option == 1) ? "Mask" :
         "Yes") : "No", info_table[i].desc ? info_table[i].desc : "<none>");
      }
    } /* switch */
  } /* for */

  /* Don't send oper_only_umodes... it's a bit mask, we will have to decode it
   * in order for it to show up properly to opers who issue INFO
   */
  sendto_one(source_p, form_str(RPL_INFO), me.name, source_p->name, "");
}
