/* m_list.c: Shows existing channels, regardless of modes.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "sprintf_irc.h"
#include "tools.h"
#include "handlers.h"
#include "channel.h"
#include "channel_mode.h"
#include "client.h"
#include "hash.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "s_conf.h"
#include "s_serv.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "newconf.h"

static void mo_olist(struct Client *, struct Client *, int, char **);
static void list_all_channels(struct Client * source_p);
static void list_limit_channels(struct Client *, char *param);
static void list_named_channel(struct Client * source_p, char *name);

struct Message olist_msgtab = {
  "OLIST", 0, 0, 0, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_ignore, m_ignore, mo_olist}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&olist_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&olist_msgtab);
}
const char *_version = "1.10.2.6";
#endif

/* mo_olist
 *  parv[0] = sender prefix
 *  parv[1] = channel
 */
static void mo_olist(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  if (!IsOperAdmin(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to view a full channel listing.", me.name, parv[0]);
    return;
  }
  sendto_wallops_flags(UMODE_WALLOP, &me, "%s is viewing the full channel "
    "listing", get_oper_name(source_p));
  sendto_server(NULL, NULL, NOCAPS, NOCAPS, ":%s WALLOPS :%s is viewing the "
    " full channel listing", me.name, get_oper_name(source_p));
  ilog(L_NOTICE, "OLIST by %s", log_client_name(source_p));
  /* If no arg, do all channels *whee*, else just one channel */
  if (parc < 2 || BadPtr(parv[1]))
    list_all_channels(source_p);
  else if(IsChannelName(parv[1]))
    list_named_channel(source_p, parv[1]);
  else
    list_limit_channels(source_p, parv[1]);
}
static void list_all_channels(struct Client * source_p)
{
  static char buf[BUFSIZE];
  static char modebuf[MODEBUFLEN];
  static char parabuf[MODEBUFLEN];
  struct Channel *chptr;
  dlink_node *ptr;
  sendto_one(source_p, form_str(RPL_LISTSTART), me.name, source_p->name);
  DLINK_FOREACH(ptr, global_channel_list.head) {
    chptr = ptr->data;
    channel_modes(chptr, source_p, modebuf, parabuf);
    ircsprintf(buf, "%s %s %s", modebuf, parabuf, chptr->topic == NULL ? "" :
      chptr->topic);
    sendto_one(source_p, form_str(RPL_LIST), me.name, source_p->name,
      chptr->chname, dlink_list_length(&chptr->members), buf);
  }
  sendto_one(source_p, form_str(RPL_LISTEND), me.name, source_p->name);
}
static void list_limit_channels(struct Client *source_p, char *param)
{
  static char buf[BUFSIZE];
  static char modebuf[MODEBUFLEN];
  static char parabuf[MODEBUFLEN];
  struct Channel *chptr;
  char *p;
  dlink_node *ptr;
  int max = INT_MAX;
  int min = 0;
  int i;

  for(i = 0; i < 2; i++)
  {
    if((p = strchr(param, ',')) != NULL)
      *p++ = '\0';
		if(*param == '<') {
      param++;
      if((max = atoi(param)) <= 0)
        max = INT_MAX;
    }
    else if(*param == '>') {
      param++;
      if((min = atoi(param)) < 0)
        min = 0;
    }
    if(EmptyString(p))
      break;
    else
      param = p;
  }
  sendto_one(source_p, form_str(RPL_LISTSTART), me.name, source_p->name);
  DLINK_FOREACH(ptr, global_channel_list.head)
  {
    chptr = ptr->data;
    if (dlink_list_length(&chptr->members) >= max ||
      dlink_list_length(&chptr->members) <= min)
      continue;
    channel_modes(chptr, source_p, modebuf, parabuf);
    ircsprintf(buf, "%s %s %s", modebuf, parabuf, chptr->topic == NULL ? "" :
      chptr->topic);
    sendto_one(source_p, form_str(RPL_LIST), me.name, source_p->name,
      chptr->chname, dlink_list_length(&chptr->members), buf);
	}
	sendto_one(source_p, form_str(RPL_LISTEND), me.name, source_p->name);
	return;
}
static void list_named_channel(struct Client * source_p, char *name)
{
  static char buf[BUFSIZE];
  static char modebuf[MODEBUFLEN];
  static char parabuf[MODEBUFLEN];
  struct Channel *chptr;
  char *p;
  sendto_one(source_p, form_str(RPL_LISTSTART), me.name, source_p->name);
  if ((p = strchr(name, ',')))
    *p = '\0';
  if (*name == '\0') {
    sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, source_p->name,
      name);
    sendto_one(source_p, form_str(RPL_LISTEND), me.name, source_p->name);
    return;
  }
  chptr = hash_find_channel(name);
  if (chptr == NULL) {
    sendto_one(source_p, form_str(ERR_NOSUCHNICK), me.name, source_p->name,
      name);
    sendto_one(source_p, form_str(RPL_LISTEND), me.name, source_p->name);
    return;
  }
  channel_modes(chptr, source_p, modebuf, parabuf);
  ircsprintf(buf, "%s %s %s", modebuf, parabuf, chptr->topic == NULL ? "" :
    chptr->topic);
  sendto_one(source_p, form_str(RPL_LIST), me.name, source_p->name,
    chptr->chname, dlink_list_length(&chptr->members), buf);
  sendto_one(source_p, form_str(RPL_LISTEND), me.name, source_p->name);
  return;
}
