/* m_oper.c: Makes a user an IRC Operator.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "tools.h"
#include "handlers.h"
#include "client.h"
#include "common.h"
#include "fdlist.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "fdlist.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_user.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "packet.h"

static struct ConfItem *find_password_aconf(char *name,
  struct Client * source_p);
static int match_oper_password(char *password, struct ConfItem * aconf);
int oper_up(struct Client * source_p, struct ConfItem * aconf);
extern char *crypt();

static void m_oper(struct Client *, struct Client *, int, char **);
static void ms_oper(struct Client *, struct Client *, int, char **);
static void mo_oper(struct Client *, struct Client *, int, char **);

struct Message oper_msgtab = {
  "OPER", 0, 0, 3, 0, MFLG_SLOW, 0,
  {m_unregistered, m_oper, ms_oper, m_ignore, mo_oper}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&oper_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&oper_msgtab);
}
const char *_version = "1.70.2.3";
#endif

/* m_oper
 *  parv[0] = sender prefix
 *  parv[1] = oper name
 *  parv[2] = oper password
 */
static void m_oper(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  static time_t last_used = 0L;
  struct ConfItem *aconf;
  struct ConfItem *oconf = NULL;
  char *name;
  char *password;
  dlink_node *ptr;
  name = parv[1];
  password = parv[2];

  if ((last_used + ConfigFileEntry.pace_wait) > CurrentTime) {
    sendto_one(source_p, form_str(RPL_LOAD2HI), me.name, BadPtr(parv[0]) ? "*" :
      parv[0], "OPER");
    return;
  }
  else
    last_used = CurrentTime;
 
  if (EmptyString(password)) {
    sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
      "OPER");
    return;
  }
  /* end the grace period */
  if (!IsFloodDone(source_p))
    flood_endgrace(source_p);
  if ((aconf = find_password_aconf(name, source_p)) == NULL) {
    sendto_one(source_p, form_str(ERR_NOOPERHOST), me.name, parv[0]);
    log_foper(source_p, name);
    if (ConfigFileEntry.failed_oper_notice)
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s failed attempt to exit the "
        "Matrix (Unrecognized Host)", get_client_name(source_p, HIDE_IP, YES));
    return;
  }
  if (match_oper_password(password, aconf)) {
    /* 20001216: detach old iline -einride */
    ptr = source_p->localClient->confs.head;
    if (ptr) {
      oconf = ptr->data;
      detach_conf(source_p, oconf);
    }
    if (attach_conf(source_p, aconf) != 0) {
      sendto_one(source_p, ":%s NOTICE %s :Failed to attach configuration!",
        me.name, parv[0]);
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s failed attempt to exit the "
        "Matrix (Could not attach configuration!)", get_client_name(source_p,
        HIDE_IP, YES));
      log_foper(source_p, name);
      /* 20001216: Reattach old iline -einride */
      attach_conf(source_p, oconf);
      return;
    }
    oper_up(source_p, aconf);
    ilog(L_TRACE, "OPER %s (Password) by %s", name, log_client_name(source_p));
    log_oper(source_p, name);
    return;
  }
  else {
    sendto_one(source_p, form_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
    log_foper(source_p, name);
    if (ConfigFileEntry.failed_oper_notice)
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s failed attempt to exit the "
        "Matrix (Incorrect Password)", get_client_name(source_p, HIDE_IP, YES));
  }
}
/* mo_oper
 *  parv[0] = sender prefix
 *  parv[1] = oper name
 *  parv[2] = oper password
 */
static void mo_oper(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  sendto_one(source_p, ":%s NOTICE %s :You are already an IRC Operator.",
    me.name, parv[0]);
  sendto_one(source_p, ":%s 381 %s :You are now an IRC Operator.", me.name,
    parv[0]);
  return;
}
/* ms_oper
 *  parv[0] = sender prefix
 *  parv[1] = oper name
 *  parv[2] = oper password
 */
static void ms_oper(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  /* if message arrived from server, trust it, and set to oper */
  if (!IsOper(source_p)) {
    if (source_p->status == STAT_CLIENT)
      source_p->handler = OPER_HANDLER;
    source_p->umodes |= UMODE_OPER;
    Count.oper++;
    sendto_server(client_p, NULL, NOCAPS, NOCAPS, ":%s MODE %s :+o", parv[0],
      parv[0]);
  }
}
/* find_password_aconf */
static struct ConfItem * find_password_aconf(char *name,
  struct Client * source_p)
{
  struct ConfItem *aconf;
  if (!(aconf = find_conf_exact(name, source_p->username, source_p->host,
    CONF_OPERATOR)) && !(aconf = find_conf_exact(name, source_p->username,
    source_p->localClient->sockhost, CONF_OPERATOR)))   
    return NO;

  return aconf;
}
/* match_oper_password
 *
 * input        - pointer to given password
 *              - pointer to Conf
 * output       - YES or NO if match
 */
static int match_oper_password(char *password, struct ConfItem * aconf)
{
  const char *encr;
  if (!aconf->status & CONF_OPERATOR)
    return NO;
  /* passwd may be NULL pointer. Head it off at the pass... */
  if (aconf->passwd == NULL)
    return NO;
  if (IsConfEncrypted(aconf)) {
    /* use first two chars of the password they send in as salt */
    /*
     * If the password in the conf is MD5, and ircd is linked to scrypt on
     * FreeBSD, or the standard crypt library on glibc Linux, then this code
     * will work fine on generating the proper encrypted hash for comparison.
     */
    if (password && *aconf->passwd)
      encr = crypt(password, aconf->passwd);
    else
      encr = "";
  }
  else {
    encr = password;
  }
  if (strcmp(encr, aconf->passwd) == 0)
    return YES;
  else
    return NO;
}
