/* m_rehash.c: Re-reads the configuration file.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "channel.h"
#include "common.h"
#include "irc_string.h"
#include "ircd.h"
#include "s_gline.h"
#include "numeric.h"
#include "res.h"
#include "s_conf.h"
#include "s_log.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "hostmask.h"
#include "reject.h"

static void mo_rehash(struct Client *, struct Client *, int, char **);
static void clear_temps(dlink_list *);
static void clear_pending_glines(void);

struct Message rehash_msgtab = {
  "REHASH", 0, 0, 0, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_ignore, m_ignore, mo_rehash}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&rehash_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&rehash_msgtab);
}
const char *_version = "1.66.2.4";
#endif

static void mo_rehash(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  if (!IsOperRehash(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You have insufficient permissions to "
      "perform a rehash.", me.name, parv[0]);
    return;
  }
  if (parc > 1) {
     if (irccmp(parv[1], "DNS") == 0) {
      sendto_one(source_p, form_str(RPL_REHASHING), me.name, parv[0], "DNS");
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is rehashing DNS",
        get_oper_name(source_p));
      restart_resolver(); /* re-read /etc/resolv.conf AGAIN */
    }
    else if (irccmp(parv[1], "MOTD") == 0) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is forcing rehash of Message "
        "of the Day file", get_oper_name(source_p));
      ReadMessageFile(&ConfigFileEntry.motd);
    }
    else if (irccmp(parv[1], "OMOTD") == 0) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is forcing rehash of Operator "
        "Message of the Day file", get_oper_name(source_p));
      ReadMessageFile(&ConfigFileEntry.opermotd);
    }
    else if (irccmp(parv[1], "GLINES") == 0) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is clearing all network bans", 
        get_oper_name(source_p));
      clear_temps(&glines);
    }
    else if (irccmp(parv[1], "PGLINES") == 0) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is clearing pending network " 
        "bans", get_oper_name(source_p));
      clear_pending_glines();
    }
    else if (irccmp(parv[1], "TKLINES") == 0) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is clearing temporary local "
        "server bans", get_oper_name(source_p));
      clear_temps(&tkline_min);
      clear_temps(&tkline_hour);
      clear_temps(&tkline_day);
      clear_temps(&tkline_week);
    }
    else if (irccmp(parv[1], "TDLINES") == 0) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is clearing temporary IP "
        "Address bans", get_oper_name(source_p));
      clear_temps(&tdline_min);
      clear_temps(&tdline_hour);
      clear_temps(&tdline_day);
      clear_temps(&tdline_week);
    }
    else if (irccmp(parv[1], "REJECTCACHE") == 0) {
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is clearing reject cache",
        get_oper_name(source_p));
      flush_reject();
    }
    else {
      sendto_one(source_p, ":%s NOTICE %s :You must specify what to rehash: "
        "DNS GLINES REJECTCACHE MOTD OMOTD PGLINES TKLINES TDLINES", me.name,
        parv[0]);
      return;
    }
    ilog(L_NOTICE, "REHASH %s by of %s", parv[1], log_client_name(source_p));
    return;
  }
  else {
    sendto_one(source_p, form_str(RPL_REHASHING), me.name, parv[0],
      ConfigFileEntry.configfile);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s is rehashing server "
      "configuration file", get_oper_name(source_p));
    ilog(L_NOTICE, "REHASH by of %s", log_client_name(source_p));
    rehash(0);
    return;
  }
}
static void clear_temps(dlink_list * tlist)
{
  dlink_node *ptr;
  dlink_node *next_ptr;
  struct ConfItem *aconf;
  DLINK_FOREACH_SAFE(ptr, next_ptr, tlist->head) {
    aconf = ptr->data;
    delete_one_address_conf(aconf->host, aconf);
    dlinkDestroy(ptr, tlist);
  }
}
static void clear_pending_glines(void)
{
  struct gline_pending *glp_ptr;
  dlink_node *ptr;
  dlink_node *next_ptr;
  DLINK_FOREACH_SAFE(ptr, next_ptr, pending_glines.head) {
    glp_ptr = ptr->data;
    MyFree(glp_ptr->reason1);
    MyFree(glp_ptr->reason2);
    MyFree(glp_ptr);
    dlinkDestroy(ptr, &pending_glines);
  }
}
