/* m_set.c: Sets a server parameter.
 *
 *  Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 *  Copyright (C) 1996-2002 Hybrid Development Team
 *  Copyright (C) 2002      jdc
 *  Copyright (C) 2002      ircd-ratbox development team
 *  Copyright (C) 2003-2004 ircd-immortal development team
 *
 *  This file is licensed under the GNU General Public License, which is
 *  distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "event.h"
#include "irc_string.h"
#include "sprintf_irc.h"
#include "ircd.h"
#include "numeric.h"
#include "fdlist.h"
#include "s_bsd.h"
#include "s_serv.h"
#include "send.h"
#include "common.h"
#include "channel.h"
#include "s_log.h"
#include "s_conf.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"

static void mo_set(struct Client *, struct Client *, int, char **);

struct Message set_msgtab = {
  "SET", 0, 0, 0, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_error, m_ignore, mo_set}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&set_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&set_msgtab);
}
const char *_version = "1.64.2.4";
#endif

/* Structure used for the SET table itself */
struct SetStruct {
  const char *name;
  void (*handler) ();
  int wants_char; /* 1 if it expects (char *, [int]) */
  int wants_int;  /* 1 if it expects ([char *], int) */
};

static void quote_autoconn(struct Client *, char *, int);
static void quote_autoconnall(struct Client *, int);
static void quote_floodcount(struct Client *, int);
static void quote_identtimeout(struct Client *, int);
static void quote_idletime(struct Client *, int);
static void quote_log(struct Client *, int);
static void quote_max(struct Client *, int);
static void quote_operstring(struct Client *, char *);
static void quote_adminstring(struct Client *, char *);
static void quote_spamnum(struct Client *, int);
static void quote_spamtime(struct Client *, int);
static void quote_splitmode(struct Client *, char *);
static void quote_splitnum(struct Client *, int);
static void quote_splitusers(struct Client *, int);
static void quote_splitdelay(struct Client *, int);
static void list_quote_commands(struct Client *);

static struct SetStruct set_cmd_table[] = {
  /* name, function, want_string, want_int */
  {"AUTOCONN", quote_autoconn, YES, YES},
  {"AUTOCONNALL", quote_autoconnall, NO, YES},
  {"FLOODCOUNT", quote_floodcount, NO, YES},
  {"IDENTTIMEOUT", quote_identtimeout, NO, YES},
  {"IDLETIME", quote_idletime, NO, YES},
  {"LOG", quote_log, NO, YES},
  {"MAX", quote_max, NO, YES},
  {"OPERSTRING", quote_operstring, YES, NO},
  {"ADMINSTRING", quote_adminstring, YES, NO},
  {"SPAMNUM", quote_spamnum, NO, YES},
  {"SPAMTIME", quote_spamtime, NO, YES},
  {"SPLITMODE", quote_splitmode, YES, NO},
  {"SPLITNUM", quote_splitnum, NO, YES},
  {"SPLITUSERS", quote_splitusers, NO, YES},
  {"SPLITDELAY", quote_splitdelay, NO, YES},
  {(char *)0, (void (*) ())0, 0, 0}
};

/* list_quote_commands() - sends the client all the available commands */
static void list_quote_commands(struct Client * source_p)
{
  int i;
  int j = 0;
  const char *names[6];
  sendto_one(source_p, ":%s NOTICE %s :Variables controlled by SET:", me.name,
    source_p->name);
  names[0] = names[1] = names[2] = names[3] = names[4] = names[5] = "";
  for (i = 0; set_cmd_table[i].handler; i++) {
    names[j++] = set_cmd_table[i].name;
    if (j > 5) {
      sendto_one(source_p, ":%s NOTICE %s :%s %s %s %s %s %s", me.name,
        source_p->name, names[0], names[1], names[2], names[3], names[4],
        names[5]);
      j = 0;
      names[0] = names[1] = names[2] = names[3] = names[4] = names[5] = "";
    }
  }
  if (j)
    sendto_one(source_p, ":%s NOTICE %s :%s %s %s %s %s %s", me.name,
      source_p->name, names[0], names[1], names[2], names[3], names[4],
      names[5]);
}

static void quote_autoconn(struct Client * source_p, char *arg, int newval)
{
  set_autoconn(source_p, arg, newval);
}

static void quote_autoconnall(struct Client * source_p, int newval)
{
  if (newval >= 0) {
    sendto_one(source_p, ":%s NOTICE %s :Changing automatic connection "
      "interval for all servers to %i seconds", me.name, source_p->name,
      newval);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed automatic "
      "connection interval for all servers to %i seconds",
      get_oper_name(source_p), newval);
    GlobalSetOptions.autoconn = newval;
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Automatic connection interval "
      "for all servers is currently %i seconds", me.name, source_p->name,
      GlobalSetOptions.autoconn);
}

static void quote_floodcount(struct Client * source_p, int newval)
{
  if (newval >= 0) {
    sendto_one(source_p, ":%s NOTICE %s :Changing number of lines "
      "required to trigger throttling to %i", me.name,
      source_p->name, newval);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed "
      "number of lines required to trigger throttling to %i",
      get_oper_name(source_p), newval);
    GlobalSetOptions.floodcount = newval;
  }
  else {
    sendto_one(source_p, ":%s NOTICE %s :Message throttling is currently "
      "triggered by %i lines of text", me.name,
      source_p->name, GlobalSetOptions.floodcount);
  }
}

static void quote_identtimeout(struct Client * source_p, int newval)
{
  if (newval > 0) {
    if (!IsOperAdmin(source_p)) {
      sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
        "permissions to modify the identity response timeout.",
        me.name, source_p->name);
      return;
    }
    sendto_one(source_p, ":%s NOTICE %s :Changing identity response timeout "
	  "to %i", me.name, source_p->name, newval);
	sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed identity response "
      "timeout to %i", get_oper_name(source_p), newval);
    GlobalSetOptions.ident_timeout = newval;
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Identity response timeout is "
      "currently %i seconds", me.name, source_p->name,
      GlobalSetOptions.ident_timeout);
}

static void quote_idletime(struct Client * source_p, int newval)
{
  if (newval == 0) {
    sendto_one(source_p, ":%s NOTICE %s :Disabling idle time threshold",
      me.name, source_p->name, newval);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has disabled idle time "
      "threshold", get_oper_name(source_p), newval);
    GlobalSetOptions.idletime = 0;
  }
  else if (newval > 0)
  {
    sendto_one(source_p, ":%s NOTICE %s :Changing idle time threshold to %i "
	  "minutes", me.name, source_p->name, newval);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed idle time threshold "
      "to %i minutes", get_oper_name(source_p), newval);
    GlobalSetOptions.idletime = (newval * 60);
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Idle time threshold is currently %i "
      "minutes", me.name, source_p->name, (GlobalSetOptions.idletime / 60));
}

static void quote_log(struct Client * source_p, int newval)
{
  const char *log_level_as_string;
  if (newval >= 0) {
    if (newval < L_WARN) {
      sendto_one(source_p, ":%s NOTICE %s :Log verbosity level must be at "
        "least %i (L_WARN)", me.name, source_p->name, L_WARN);
      return;
    }
    if (newval > L_DEBUG)
      newval = L_DEBUG;

    log_level_as_string = get_log_level_as_string(newval);
    sendto_one(source_p, ":%s NOTICE %s :Changing logging verbosity level to "
	  "%i (%s)", me.name, source_p->name, newval, log_level_as_string);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed logging verbosity "
      "level to %i (%s)", get_oper_name(source_p), newval, log_level_as_string);
    set_log_level(newval);
  }
  else {
    newval = get_log_level();
    log_level_as_string = get_log_level_as_string(newval);
    sendto_one(source_p, ":%s NOTICE %s :Logging verbosity level is currently "
      "%i (%s)", me.name, source_p->name, newval, log_level_as_string);
  }
}

static void quote_max(struct Client * source_p, int newval)
{
  if (newval > 0) {
    if (newval > MASTER_MAX) {
      sendto_one(source_p, ":%s NOTICE %s :You cannot set the maximum client "
        "capacity larger than the maximum allocated file descriptors (%d)",
        me.name, source_p->name, MASTER_MAX);
      return;
    }
    else if (newval < 32) {
      sendto_one(source_p, ":%s NOTICE %s :You cannot change the maximum "
        "client capacity to less than 32", me.name, source_p->name);
      return;
    }
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed the maximum client "
      "capacity to %i clients", get_oper_name(source_p), newval);
    GlobalSetOptions.maxclients = newval;
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :The current maximum client capacity "
      "is %i clients", me.name, source_p->name, GlobalSetOptions.maxclients);
}

static void quote_operstring(struct Client * source_p, char *operstring)
{
  if (!EmptyString(operstring)) {
    if (!IsAdmin(source_p)) {
      sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
        "permissions to modify the operator string", me.name,
        source_p->name);
      return;
    }
    sendto_one(source_p, ":%s NOTICE %s :Changing the operator string to "
      "\"%s\"", me.name, source_p->name, operstring);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed operator string to "
      "\"%s\"", get_oper_name(source_p), operstring);
    strlcpy(GlobalSetOptions.operstring, operstring,
      sizeof(GlobalSetOptions.operstring));
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Operator string is currently "
      "\"%s\"", me.name, source_p->name, GlobalSetOptions.operstring);
}

static void quote_adminstring(struct Client * source_p, char *adminstring)
{
  if (!EmptyString(adminstring)) {
    if (!IsAdmin(source_p)) {
      sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
        "permissions to modify the administrator string", me.name,
        source_p->name);
      return;
    }
    sendto_one(source_p, ":%s NOTICE %s :Changing the administrator string to "
      "\"%s\"", me.name, source_p->name, adminstring);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed administrator "
      "string to \"%s\"", get_oper_name(source_p), adminstring);
    strlcpy(GlobalSetOptions.adminstring, adminstring,
      sizeof(GlobalSetOptions.adminstring));
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Administrator string is currently "
      "\"%s\"", me.name, source_p->name, GlobalSetOptions.adminstring);
}

static void quote_spamnum(struct Client * source_p, int newval)
{
  if (newval > 0) {
    if (newval == 0) {
      sendto_one(source_p, ":%s NOTICE %s :Disabling spam bot detection "
        "routines", me.name, source_p->name);
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s has disabled spam bot "
        "detection routines", source_p->name);
      GlobalSetOptions.spam_num = newval;
      return;
    }
    else if (newval < MIN_SPAM_NUM)
      newval = MIN_SPAM_NUM;

    sendto_one(source_p, ":%s NOTICE %s :Changing the number of joins/parts "
      "within the spam bot threshold required to detect spam bots to %i",
      me.name, source_p->name, newval);
	sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed the number of "
      "joins/parts within the spam bot threshold to detect spam bots to %i",
      get_oper_name(source_p), newval);
	GlobalSetOptions.spam_num = newval;
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :The number of joins/parts within the "
      "spam bot threshold required to detect spam bots to %i", me.name,
      source_p->name, GlobalSetOptions.spam_num);
}

static void quote_spamtime(struct Client * source_p, int newval)
{
  if (newval > 0) {
    if (newval < MIN_SPAM_TIME)
      newval = MIN_SPAM_TIME;

    sendto_one(source_p, ":%s NOTICE %s :Changing the spam bot detection "
      "threshold to %i seconds", me.name, source_p->name, newval);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed the spam bot "
      "detection threshold to %i seconds", source_p->name, newval);   
    GlobalSetOptions.spam_time = newval;
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :The spam bot detection threshold is "
      "currently %i", me.name, source_p->name, GlobalSetOptions.spam_time);
}

/* this table is what splitmode may be set to */
static const char *splitmode_values[] = {
  "OFF",
  "ON",
  "AUTO",
  NULL
};
/* this table is what splitmode may be */
static const char *splitmode_status[] = {
  "inactive",
  "automatically triggered (inactive)",
  "active",
  "automatically triggered (active)",
  NULL
};
static void quote_splitmode(struct Client * source_p, char *charval)
{
  if (charval) {
    int newval;
    for (newval = 0; splitmode_values[newval]; newval++) {
      if (!irccmp(splitmode_values[newval], charval))
        break;
    }
    if (newval == 0) {
      sendto_one(source_p, ":%s NOTICE %s :Deactivating network split mode "
        "protection routines", me.name, source_p->name);
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s has deactivated network split "
        "mode protection routines", get_oper_name(source_p));
      splitmode = 0;
      splitchecking = 0;
      eventDelete(check_splitmode, NULL);
    }
    else if (newval == 1) {
      sendto_one(source_p, ":%s NOTICE %s :Activating network split mode "
        "protection routines", me.name, source_p->name);
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s has activated network split "
        "mode protection routines", get_oper_name(source_p));
      splitmode = 1;
      splitchecking = 0;
      eventDelete(check_splitmode, NULL);
    }
    /* AUTO */
    else {
      sendto_one(source_p, ":%s NOTICE %s :Enabling automated network split "
        "detection and protection routines", me.name, source_p->name);
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is enabling automated network "
        "split detection and protection routines", get_oper_name(source_p));
      splitchecking = 1;
      check_splitmode(NULL);
    }
  }
  else
    /* if we add splitchecking to splitmode*2 we get a unique table to pull
     * values back out of, splitmode can be four states - but you can only set
     * to three, which means we cant use the same table --fl_
     */
    sendto_one(source_p, ":%s NOTICE %s :Network split protection mode is "
      "currently %s", me.name, source_p->name,
      splitmode_status[(splitchecking + (splitmode * 2))]);
}

static void quote_splitnum(struct Client * source_p, int newval)
{
  if (newval >= 0) {
    sendto_one(source_p, ":%s NOTICE %s :Changing network split detection "
      "server trigger to %i", me.name, source_p->name, newval);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed network split "
      "detection server trigger to %i", get_oper_name(source_p), newval);
    split_servers = newval;
    if (splitchecking)
      check_splitmode(NULL);
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Network split detection server "
      "trigger is currently %i", me.name, source_p->name, split_servers);
}

static void quote_splitusers(struct Client * source_p, int newval)
{
  if (newval >= 0) {
    sendto_one(source_p, ":%s NOTICE %s :Changing network split detection "
      "server trigger to %i", me.name, source_p->name, newval);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed network split "
      "detection user trigger to %i", get_oper_name(source_p), newval);
    split_users = newval;
    if (splitchecking)
      check_splitmode(NULL);
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Network split detection user trigger "
      "is currently %i", me.name, source_p->name, split_users);
}

static void quote_splitdelay(struct Client * source_p, int newval)
{
  if (newval >= 0) {
    sendto_one(source_p, ":%s NOTICE %s :Changing network split checking "
      "delay to %i", me.name, source_p->name, newval);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s has changed network split "
      "checking delay to %i", get_oper_name(source_p), newval);
    GlobalSetOptions.split_delay = newval;
  }
  else
    sendto_one(source_p, ":%s NOTICE %s :Network split detection delay is "
      "currently %i", me.name, source_p->name, GlobalSetOptions.split_delay);
}

/* mo_set - set options while running */
static void mo_set(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  int newval;
  int i, n;
  char *arg = NULL;
  char *intarg = NULL;
  if (parc > 1) {
    /* Go through all the commands in set_cmd_table, until one is matched. */
    for (i = 0; set_cmd_table[i].handler; i++) {
      if (!irccmp(set_cmd_table[i].name, parv[1])) {
        /* Command found; now execute the code */
        n = 2;
        if (set_cmd_table[i].wants_char)
          arg = parv[n++];
        if (set_cmd_table[i].wants_int)
          intarg = parv[n++];
        if ((n - 1) > parc) {
          if (parc > 2)
            sendto_one(source_p, ":%s NOTICE %s :Variable %s expects (\"%s%s\")"
              "arguments", me.name, source_p->name, set_cmd_table[i].name,
              (set_cmd_table[i].wants_char ? "string, " : ""),
              (set_cmd_table[i].wants_char ? "int" : ""));
        }
        if (parc <= 2) {
          arg = NULL;
          intarg = NULL;
        }
        if (set_cmd_table[i].wants_int && (parc > 2)) {
          if (intarg) {
            if (!irccmp(intarg, "yes") || !irccmp(intarg, "on"))
              newval = 1;
            else if (!irccmp(intarg, "no") || !irccmp(intarg, "off"))
              newval = 0;
            else
              newval = atoi(intarg);
          }
          else {
            newval = -1;
          }
          if (newval < 0) {
            sendto_one(source_p, ":%s NOTICE %s :Values less than 0 are "
              "invalid for %s", me.name, source_p->name, set_cmd_table[i].name);
            return;
          }
        }
        else
          newval = -1;
        if (set_cmd_table[i].wants_char) {
          if (set_cmd_table[i].wants_int)
            set_cmd_table[i].handler(source_p, arg, newval);
          else
            set_cmd_table[i].handler(source_p, arg);
          return;
        }
        else {
          if (set_cmd_table[i].wants_int)
            set_cmd_table[i].handler(source_p, newval);
          else
            set_cmd_table[i].handler(source_p);
          return;
        }
      }
    }
    /* Code here will be executed when a /QUOTE SET command is not found within
     * set_cmd_table.
     */
    sendto_one(source_p, ":%s NOTICE %s :Variable not found.", me.name,
      parv[0]);
    return;
  }
  list_quote_commands(source_p);
}
