/* m_svinfo.c: Sends TS information for clock & compatibility checks.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "common.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "send.h"
#include "s_conf.h"
#include "s_log.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"

static void ms_svinfo(struct Client *, struct Client *, int, char **);

struct Message svinfo_msgtab = {
  "SVINFO", 0, 0, 4, 0, MFLG_SLOW, 0,
  {m_unregistered, m_ignore, ms_svinfo, m_ignore, m_ignore}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&svinfo_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&svinfo_msgtab);
}
const char *_version = "1.47.2.2";
#endif

/* ms_svinfo - SVINFO message handler
 *  parv[0] = sender prefix
 *  parv[1] = TS_CURRENT for the server
 *  parv[2] = TS_MIN for the server
 *  parv[3] = server is standalone or connected to non-TS only
 *  parv[4] = server's idea of UTC time
 */
static void ms_svinfo(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  time_t deltat;
  time_t theirtime;
  if (MyConnect(source_p) && IsUnknown(source_p)) {
    exit_client(source_p, source_p, source_p, "Received SVINFO before SERVER");
    return;
  }
  if (!IsServer(source_p) || !MyConnect(source_p) || parc < 5)
    return;
  if (TS_CURRENT < atoi(parv[2]) || atoi(parv[1]) < TS_MIN) {
    /* wrong TS version! can't fall back to non-TS, drop link! */
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Link %s dropped, wrong TS "
      "protocol version (%s,%s)", get_client_name(source_p, SHOW_IP, YES),
      parv[1], parv[2]);
    sendto_realops_flags(UMODE_ALL, L_OPER, "Link %s dropped, wrong TS "
      "protocol version (%s,%s)", get_client_name(source_p, MASK_IP, YES),
       parv[1], parv[2]);
    exit_client(source_p, source_p, source_p, "Incompatible TS version");
    return;
  }
  /* might as well set CurrentTime while we're at it */
  set_time();
  theirtime = atol(parv[4]);
  deltat = abs(theirtime - CurrentTime);
  if (deltat > ConfigFileEntry.ts_max_delta) {
    sendto_realops_flags(UMODE_ALL, L_ADMIN, "Link %s dropped, excessive TS "
      "delta (my TS=%lu, their TS=%lu, delta=%d)", get_client_name(source_p,
      SHOW_IP, YES), (unsigned long)CurrentTime,
      (unsigned long)theirtime, (int)deltat);
    sendto_realops_flags(UMODE_ALL, L_OPER, "Link %s dropped, excessive TS "
      "delta (my TS=%lu, their TS=%lu, delta=%d)", get_client_name(source_p,
      MASK_IP, YES), (unsigned long)CurrentTime,
      (unsigned long)theirtime, (int)deltat);
    ilog(L_NOTICE, "%s dropped, excessive TS delta (my TS=%lu, their TS=%lu, "
      "delta=%d)", log_client_name(source_p), CurrentTime, theirtime,
      (int)deltat);
    exit_client(source_p, source_p, source_p, "Excessive TS delta");
    return;
  }
  if (deltat > ConfigFileEntry.ts_warn_delta)
    sendto_realops_flags(UMODE_ALL, L_ALL, "Link %s notable TS delta (my "
      "TS=%lu, their TS=%lu, delta=%d)", source_p->name, CurrentTime,
      theirtime, (int)deltat);
}
