/*
 *  m_takeover.c: allows one to 'take over' channels
 *
 *  Copyright (C) 1996-2002 Hybrid Development Team
 *  Copyright (C) 2003-2004 ircd-immortal development team
 *
 *  This file is licensed under the GNU General Public License, which is
 *  distributed as LICENSE.
 */

#include "stdinc.h"
#include "tools.h"
#include "handlers.h"
#include "channel.h"
#include "channel_mode.h"
#include "client.h"
#include "ircd.h"
#include "numeric.h"
#include "s_log.h"
#include "s_serv.h"
#include "send.h"
#include "whowas.h"
#include "irc_string.h"
#include "sprintf_irc.h"
#include "hash.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"

static void mo_takeover(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[]);

struct Message takeover_msgtab = {
  "TAKEOVER", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_ignore, m_ignore, mo_takeover}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&takeover_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&takeover_msgtab);
}
const char *_version = "1.45.2.5";
#endif

/* mo_takeover
 *  parv[0] = sender prefix
 *  parv[1] = channel
 */
static void mo_takeover(struct Client * client_p, struct Client * source_p,
   int parc, char *parv[])
{
  struct Channel *chptr;
  struct membership *msptr;
  struct Client *target_p;
  dlink_node *ptr;
  dlink_node *next_ptr;
  /* admins only */
  if (!IsOperAdmin(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to take over channels", me.name, parv[0]);
    return;
  }
  chptr = hash_find_channel(parv[1]);
  if (chptr == NULL) {
    sendto_one(source_p, form_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
      parv[1]);
    return;
  }
  if (IsMember(source_p, chptr)) {
    sendto_one(source_p, ":%s NOTICE %s :You cannot take over a channel in "
      "which you are a member of.", me.name, parv[0], parv[1]);
    return;
  }
  /* quickly make everyone a peon */
  DLINK_FOREACH(ptr, chptr->members.head) {
    msptr = ptr->data;
    msptr->flags &= ~CHFL_CHANOP | CHFL_VOICE;
  }
  sendto_wallops_flags(UMODE_WALLOP, &me, "%s is taking over channel %s",
    get_oper_name(source_p), parv[1]);
  ilog(L_NOTICE, "TAKEOVER %s by %s", parv[1], log_client_name(source_p));
  if (*chptr->chname != '&') {
    sendto_server(NULL, NULL, NOCAPS, NOCAPS, ":%s WALLOPS :%s is taking over "
      "channel %s", me.name, get_oper_name(source_p), parv[1]);
    /* SJOIN the user to give them ops, and lock the channel */
    sendto_server(client_p, chptr, NOCAPS, NOCAPS, ":%s SJOIN %lu %s +ntsi "
      ":@%s", me.name, (unsigned long)(chptr->channelts - 1), chptr->chname,
      parv[0]);
  }
  sendto_channel_local(ALL_MEMBERS, chptr, ":%s!%s@%s JOIN %s", parv[0],
    source_p->username, source_p->host, chptr->chname);
  sendto_channel_local(ALL_MEMBERS, chptr, ":%s MODE %s +o %s", me.name,
    chptr->chname, parv[0]);
  add_user_to_channel(chptr, source_p, CHFL_CHANOP);
  /* Take the TS down by 1, so we don't see the channel taken over again */
  if (chptr->channelts)
    chptr->channelts--;
  chptr->mode.mode = MODE_SECRET | MODE_TOPICLIMIT | MODE_INVITEONLY |
    MODE_NOPRIVMSGS;
  free_topic(chptr);
  chptr->mode.key[0] = '\0';
  DLINK_FOREACH_SAFE(ptr, next_ptr, chptr->members.head) {
    msptr = ptr->data;
    target_p = msptr->client_p;
    /* skip the person we just added.. */
    if (is_chanop(msptr))
      continue;
    sendto_channel_local(ALL_MEMBERS, chptr, ":%s KICK %s %s :Takeover",
      parv[0], chptr->chname, target_p->name);
    if (*chptr->chname != '&')
      sendto_server(NULL, chptr, NOCAPS, NOCAPS, ":%s KICK %s %s :Takeover",
        parv[0], chptr->chname, target_p->name);
    remove_user_from_channel(msptr);
  }
  /* Join the user themself to the channel now, so they dont see a nicklist or
   * people being kicked
   */
  sendto_one(source_p, ":%s!%s@%s JOIN %s", parv[0], source_p->username,
    source_p->host, chptr->chname);
  channel_member_names(chptr, source_p, 1);
}
