/* m_testline.c: Tests a hostmask to see what will happen to it.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "handlers.h"
#include "client.h"
#include "common.h"
#include "irc_string.h"
#include "ircd_defs.h"
#include "ircd.h"
#include "restart.h"
#include "s_conf.h"
#include "send.h"
#include "msg.h"
#include "hostmask.h"
#include "numeric.h"
#include "parse.h"
#include "modules.h"

static void mo_testline(struct Client *, struct Client *, int, char **);

struct Message testline_msgtab = {
  "TESTLINE", 0, 0, 0, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_ignore, m_ignore, mo_testline}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&testline_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&testline_msgtab);
}
const char *_version = "1.43.3.1";
#endif

/* mo_testline - tests user@host|ip for aconf
 * input - pointer to physical connection request is coming from
 *       - pointer to source connection request is comming from
 *       - parc arg count
 *       - parv actual arguments
 */
static void mo_testline(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  struct ConfItem *aconf;
  struct irc_inaddr ip;
  int host_mask;
  char *host, *pass, *user, *name, *classname, *given_host, *given_name,
  *p;
  int port, t;
  if (parc > 1) {
    given_name = parv[1];
    if (!(p = (char *)strchr(given_name, '@'))) {
      if ((t = parse_netmask(given_name, &ip, &host_mask)) != HM_HOST) {
        aconf = find_dline(&ip,
#ifdef IPV6
          (t == HM_IPV6) ? AF_INET6 :
#endif
          AF_INET);
        if (aconf) {
          get_printable_conf(aconf, &name, &host, &pass, &user, &port,
            &classname);
          if (aconf->status & CONF_EXEMPTDLINE) {
            sendto_one(source_p, ":%s NOTICE %s :Exempt D-Line Information "
              "for address %s", me.name, parv[0], parv[1]);
            sendto_one(source_p, ":%s NOTICE %s :Address: %s", me.name, parv[0],
              host);
            sendto_one(source_p, ":%s NOTICE %s :Password: %s", me.name,
              parv[0], pass);
          }
          else {
            sendto_one(source_p, ":%s NOTICE %s :D-Line Information for "
              "address %s", me.name, parv[0], parv[1]);
            sendto_one(source_p, ":%s NOTICE %s :Address: %s", me.name, parv[0],
              host);
            sendto_one(source_p, ":%s NOTICE %s :Reason: %s", me.name, parv[0],
              pass);
          }
        }
        else
          sendto_one(source_p, ":%s NOTICE %s :No D-Line matching the provided "
            "IP Address was found", me.name, parv[0]);
      }
      else
        sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
          "TESTLINE");
      return;
    }
    *p = '\0';
    p++;
    given_host = p;
    if ((t = parse_netmask(given_host, &ip, &host_mask)) != HM_HOST)
      aconf = find_address_conf(given_host, given_name, &ip,
#ifdef IPV6
        (t == HM_IPV6) ? AF_INET6 : 
#endif
        AF_INET);
    else
      aconf = find_address_conf(given_host, given_name, NULL, 0);
    if (aconf) {
      get_printable_conf(aconf, &name, &host, &pass, &user, &port, &classname);
      if (aconf->status & CONF_KILL) {
        sendto_one(source_p, ":%s NOTICE %s :%c-Line Information", me.name,
          parv[0], (aconf->flags & CONF_FLAGS_TEMPORARY) ? 'k' : 'K');
        sendto_one(source_p, ":%s NOTICE %s :Hostmask: %s@%s", me.name, parv[0],
          user, host);
        sendto_one(source_p, ":%s NOTICE %s :Reason: %s", me.name, parv[0],
          pass);
        sendto_one(source_p, ":%s NOTICE %s :Type: %s", (aconf->flags &
          CONF_FLAGS_TEMPORARY) ? "Temporary" : "Permanent");
      }
      else if (aconf->status & CONF_CLIENT) {
        sendto_one(source_p, ":%s NOTICE %s :I-Line Information", me.name,
          parv[0]);
        sendto_one(source_p, ":%s NOTICE %s :Flags: %s", me.name, parv[0],
          show_iline_prefix(source_p, aconf, user));
        sendto_one(source_p, ":%s NOTICE %s :Class: %s", me.name, parv[0],
          classname);
        sendto_one(source_p, ":%s NOTICE %s :Hostmask: %s@%s", me.name, parv[0],
          user, host);
        if (name != "NOMATCH")
          sendto_one(source_p, ":%s NOTICE %s :Spoof: %s", me.name, parv[0],
            name);
        sendto_one(source_p, ":%s NOTICE %s :Port: %d", me.name, parv[0], port);
      }
    }
    else
      sendto_one(source_p, ":%s NOTICE %s :The specified mask specified would "
        "revert to the default configuration.", me.name, parv[0]);
  }
  else
    sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
      "TESTLINE");
}
