/* m_topic.c: Sets a channel topic.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "tools.h"
#include "handlers.h"
#include "channel.h"
#include "channel_mode.h"
#include "client.h"
#include "hash.h"
#include "irc_string.h"
#include "sprintf_irc.h"
#include "ircd.h"
#include "numeric.h"
#include "send.h"
#include "s_conf.h"
#include "s_serv.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "packet.h"

static void m_topic(struct Client *, struct Client *, int, char **);
static void ms_topic(struct Client *, struct Client *, int, char **);

struct Message topic_msgtab = {
  "TOPIC", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_topic, ms_topic, m_ignore, m_topic}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&topic_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&topic_msgtab);
}
const char *_version = "1.77.2.3";
#endif

/* m_topic
 *  parv[0] = sender prefix
 *  parv[1] = channel name
 *  parv[2] = new topic, if setting topic
 */
static void m_topic(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  struct Channel *chptr = NULL;
  struct membership *msptr;
  char *p = NULL;
  if (EmptyString(parv[1])) {
    sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
      "TOPIC");
    return;
  }
  if ((p = strchr(parv[1], ',')))
    *p = '\0';
  if (MyClient(source_p) && !IsFloodDone(source_p))
    flood_endgrace(source_p);
  if (parv[1] && IsChannelName(parv[1])) {
    chptr = hash_find_channel(parv[1]);
    if (chptr == NULL) {
      sendto_one(source_p, form_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
        parv[1]);
      return;
    }
    /* setting topic */
    if (parc > 2) {
      msptr = find_channel_membership(chptr, source_p);
      if (msptr == NULL) {
        sendto_one(source_p, form_str(ERR_NOTONCHANNEL), me.name, parv[0],
          parv[1]);
        return;
      }
      if ((chptr->mode.mode & MODE_TOPICLIMIT) == 0 || is_chanop(msptr)) {
        char topic_info[USERHOST_REPLYLEN];
        ircsprintf(topic_info, "%s!%s@%s", parv[0], source_p->username,
          source_p->host);
        set_channel_topic(chptr, parv[2], topic_info, CurrentTime);
        sendto_server(client_p, chptr, NOCAPS, NOCAPS, ":%s TOPIC %s :%s",
          parv[0], chptr->chname, chptr->topic == NULL ? "" : chptr->topic);
        sendto_channel_local(ALL_MEMBERS, chptr, ":%s!%s@%s TOPIC %s :%s",
          parv[0], source_p->username, source_p->host, chptr->chname,
          chptr->topic == NULL ? "" : chptr->topic);
      }
      else
        sendto_one(source_p, form_str(ERR_CHANOPRIVSNEEDED), me.name, parv[0],
          parv[1]);
    }
    else { /* only asking for topic */
      if (!IsMember(source_p, chptr) && SecretChannel(chptr)) {
        sendto_one(source_p, form_str(ERR_NOTONCHANNEL), me.name, parv[0],
          parv[1]);
        return;
      }
      if (chptr->topic == NULL)
        sendto_one(source_p, form_str(RPL_NOTOPIC), me.name, parv[0], parv[1]);
      else {
        sendto_one(source_p, form_str(RPL_TOPIC), me.name, parv[0],
          chptr->chname, chptr->topic);
        sendto_one(source_p, form_str(RPL_TOPICWHOTIME), me.name, parv[0],
          chptr->chname, chptr->topic_info, chptr->topic_time);
      }
    }
  }
  else
    sendto_one(source_p, form_str(ERR_NOSUCHCHANNEL), me.name, parv[0],
      parv[1]);
}
/* ms_topic
 *  parv[0] = sender prefix
 *  parv[1] = channel name
 *  parv[2] = topic_info
 *  parv[3] = topic_info time
 *  parv[4] = new channel topic
 *
 * Let servers always set a topic
 */
static void ms_topic(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  struct Channel *chptr = NULL;
  if (!IsServer(source_p)) {
    m_topic(client_p, source_p, parc, parv);
    return;
  }
  if (parc < 5)
    return;
  if (parv[1] && IsChannelName(parv[1])) {
    if ((chptr = hash_find_channel(parv[1])) == NULL)
      return;
    set_channel_topic(chptr, parv[4], parv[2], atoi(parv[3]));
    if (ConfigServerHide.hide_servers)
      sendto_channel_local(ALL_MEMBERS, chptr, ":%s TOPIC %s :%s", me.name,
        parv[1], chptr->topic == NULL ? "" : chptr->topic);
    else
      sendto_channel_local(ALL_MEMBERS, chptr, ":%s TOPIC %s :%s", parv[0],
        parv[1], chptr->topic == NULL ? "" : chptr->topic);
  }
}
