/* m_unkline.c: Unklines a user.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "tools.h"
#include "handlers.h"
#include "channel.h"
#include "client.h"
#include "common.h"
#include "fileio.h"
#include "irc_string.h"
#include "sprintf_irc.h"
#include "ircd.h"
#include "hostmask.h"
#include "numeric.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_misc.h"
#include "send.h"
#include "msg.h"
#include "s_gline.h"
#include "parse.h"
#include "modules.h"
#include "s_serv.h"
#include "cluster.h"

static void mo_unkline(struct Client *, struct Client *, int, char **);
static void me_unkline(struct Client *, struct Client *, int, char **);
static void ms_unkline(struct Client *, struct Client *, int, char **);
static void mo_undline(struct Client *, struct Client *, int, char **);
static void mo_ungline(struct Client *, struct Client *, int, char **);
static void handle_remote_unkline(struct Client *, const char *, const char *);
static void remove_permkline_match(struct Client *, const char *, const char *);
static int flush_write(struct Client *, FBFILE *, char *, char *);
static int remove_temp_kline(const char *, const char *);
static int remove_temp_dline(const char *);
static int remove_temp_gline(const char *, const char *);

struct Message kline_msgtab = {
  "UNKLINE", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, ms_unkline, m_ignore, mo_unkline},
};
struct Message dline_msgtab = {
  "UNDLINE", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_error, m_ignore, mo_undline},
};
struct Message gline_msgtab = {
  "UNGLINE", 0, 0, 2, 0, MFLG_SLOW, 0,
  {m_unregistered, m_not_oper, m_error, m_ignore, mo_ungline}
};

#ifndef STATIC_MODULES
void _modinit(void)
{
  mod_add_cmd(&kline_msgtab);
  mod_add_cmd(&dline_msgtab);
  mod_add_cmd(&gline_msgtab);
}
void _moddeinit(void)
{
  mod_del_cmd(&kline_msgtab);
  mod_del_cmd(&dline_msgtab);
  mod_del_cmd(&gline_msgtab);
}
const char *_version = "1.90.3.5";
#endif

/* mo_unkline
 * Added Aug 31, 1997
 * common (Keith Fralick) fralick@gate.net
 *  parv[0] = sender
 *  parv[1] = address to remove
*/
static void mo_unkline(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  const char *user;
  char *host;
  if (!IsOperUnkline(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to remove K-Lines", me.name, parv[0]);
    return;
  }
  if (parc < 2 || EmptyString(parv[1])) {
    sendto_one(source_p, form_str(ERR_NEEDMOREPARAMS), me.name, parv[0],
      "UNKLINE");
    return;
  }
  if ((host = strchr(parv[1], '@')) || *parv[1] == '*') {
    /* Explicit user@host mask given */
    if (host) {
      user = parv[1]; /* user */
      *(host++) = '\0'; /* host */
    }
    else {
      user = "*"; /* no @ found, assume user is * */
      host = parv[1];
    }
  }
  else {
    sendto_one(source_p, ":%s NOTICE %s :Invalid parameters", me.name, parv[0]);
    return;
  }
  /* possible remote kline */
  if (parc > 3 && (irccmp(parv[2], "ON") == 0)) {
    propagate_generic(source_p, "UNKLINE", parv[3], CAP_UNKLN, "%s %s", user,
      host);
    if (match(parv[3], me.name) == 0)
      return;
  }
  else if (dlink_list_length(&cluster_list) > 0)
    cluster_generic(source_p, "UNKLINE", CLUSTER_UNKLINE, CAP_UNKLN, "%s %s",
      user, host);
  if (remove_temp_kline(user, host)) {
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s removed K-Line for %s@%s",
      get_oper_name(source_p), user, host);
    ilog(L_NOTICE, "UNKLINE %s@%s by %s", user, host,
      log_client_name(source_p));
    sendto_one(source_p, ":%s NOTICE %s :Temporary K-Line for %s@%s removed",
      me.name, parv[0], user, host);
    return;
  }
  remove_permkline_match(source_p, host, user);
}

/* ms_unkline
 * input        - pointer to servere
 *              - pointer to client
 *              - parm count
 *              - params
 * side effects - kline is removed if matching shared {} is found.
 */
static void ms_unkline(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  if (parc != 4)
    return;
  /* parv[0] = oper
   * parv[1] = target server
   * parv[2] = user
   * parv[3] = host
   */
  propagate_generic(source_p, "UNKLINE", parv[1], CAP_UNKLN, "%s %s", parv[2],
    parv[3]);
  if (!match(parv[1], me.name) || !IsPerson(source_p))
    return;
  handle_remote_unkline(source_p, parv[2], parv[3]);
}

static void me_unkline(struct Client *client_p, struct Client *source_p,
  int parc, char *parv[])
{
  if (parc != 3)
    return;

  handle_remote_unkline(source_p, parv[1], parv[2]);
}

static void handle_remote_unkline(struct Client *source_p, const char *kuser,
  const char *khost)
{
  if (find_cluster(source_p->user->server, CLUSTER_UNKLINE) ||
      find_u_conf((char *) source_p->user->server, source_p->username,
      source_p->host, OPER_UNKLINE)) {
    if (remove_temp_kline(kuser, khost)) {
      sendto_one(source_p, ":%s NOTICE %s :Temporary K-Line for %s@%s removed",
        me.name, source_p->name, kuser, khost);
      sendto_realops_flags(UMODE_ALL, L_ALL, "%s is removing K-Line for %s@%s",
        get_oper_name(source_p), kuser, khost);
      ilog(L_NOTICE, "UNKLINE %s@%s by %s", kuser, khost,
        log_client_name(source_p));
      return;
    }
    remove_permkline_match(source_p, khost, kuser);
  }
}

/* remove_permkline_match - removes a permanent k-line */
static void remove_permkline_match(struct Client * source_p, const char *host,
  const char *user)
{
  FBFILE *in, *out;
  int pairme = 0;
  int error_on_write = NO;
  char buf[BUFSIZE];
  char buff[BUFSIZE];
  char temppath[BUFSIZE];
  const char *filename;
  mode_t oldumask;
  char *p;
  snprintf(temppath, sizeof(temppath),
    "%s.tmp", ConfigFileEntry.klinefile);
  filename = get_conf_name(KLINE_TYPE);
  if ((in = fbopen(filename, "r")) == 0) {
    sendto_one(source_p, ":%s NOTICE %s :Couldn't open %s", me.name,
      source_p->name, filename);
    return;
  }
  oldumask = umask(0);
  if ((out = fbopen(temppath, "w")) == 0) {
    sendto_one(source_p, ":%s NOTICE %s :Couldn't open %s", me.name,
      source_p->name, temppath);
    fbclose(in);
    umask(oldumask);
    return;
  }
  umask(oldumask);
  while (fbgets(buf, sizeof(buf), in)) {
    char *found_host, *found_user;
    strlcpy(buff, buf, sizeof(buff));
    if ((p = strchr(buff, '\n')) != NULL)
      *p = '\0';
    if ((*buff == '\0') || (*buff == '#')) {
      if (!error_on_write)
        flush_write(source_p, out, buf, temppath);
      continue;
    }
    if ((found_user = getfield(buff)) == NULL) {
      if (!error_on_write)
        flush_write(source_p, out, buf, temppath);
      continue;
    }
    if ((found_host = getfield(NULL)) == NULL) {
      if (!error_on_write)
        flush_write(source_p, out, buf, temppath);
      continue;
    }
    if ((irccmp(host, found_host) == 0) && (irccmp(user, found_user) == 0))
      pairme++;
    else {
      if (!error_on_write)
        flush_write(source_p, out, buf, temppath);
    }
  }
  fbclose(in);
  fbclose(out);
/* The result of the rename should be checked too... oh well */
/* If there was an error on a write above, then its been reported
 * and I am not going to trash the original kline /conf file
 */
  if (!error_on_write) {
    (void)rename(temppath, filename);
    rehash(0);
  }
  else {
    sendto_one(source_p, ":%s NOTICE %s :Couldn't write temporary kline file, "
      "aborted", me.name, source_p->name);
	  return;
  }
  if (!pairme) {
    sendto_one(source_p, ":%s NOTICE %s :There is no K-Line for %s@%s", me.name,
      source_p->name, user, host);
    return;
  }
  sendto_one(source_p, ":%s NOTICE %s :Removing K-Line for %s@%s", me.name,
    source_p->name, user, host);
  sendto_realops_flags(UMODE_ALL, L_ALL, "%s is removing K-Line for %s@%s",
    get_oper_name(source_p), user, host);
  ilog(L_NOTICE, "UNKLINE %s@%s by %s", user, host, log_client_name(source_p));
  return;
}

/* flush_write
 * inputs       - pointer to client structure of oper requesting unkline
 *              - out is the file descriptor
 *              - buf is the buffer to write
 *              - ntowrite is the expected number of character to be written
 *              - temppath is the temporary file name to be written
 * output       - YES for error on write, NO for success
 * side effects - success, the buf is written to output file
 *                failure, and the file pointed to by temppath is removed
 */
static int flush_write(struct Client * source_p, FBFILE * out, char *buf,
  char *temppath)
{
  int error_on_write = (fbputs(buf, out) < 0) ? YES : NO;
  if (error_on_write) {
    sendto_one(source_p, ":%s NOTICE %s :Failed to open %s: K-Line may "
      "not have been properly removed.", me.name, source_p->name, temppath);
    fbclose(out);
    if (temppath != (char *)NULL)
      (void)unlink(temppath);
  }
  return error_on_write;
}
static dlink_list *tkline_list[] = {
  &tkline_hour,
  &tkline_day,
  &tkline_min,
  &tkline_week,
  NULL
};
/* remove_temp_kline
 * inputs       - username, hostname to unkline
 * side effects - tries to unkline anything that matches
 */
static int remove_temp_kline(const char *user, const char *host)
{
  dlink_list *tklist;
  struct ConfItem *aconf;
  dlink_node *ptr;
  struct irc_inaddr addr, caddr;
  int nm_t, cnm_t, bits, cbits;
  int i;
  nm_t = parse_netmask(host, &addr, &bits);
  for (i = 0; tkline_list[i] != NULL; i++) {
    tklist = tkline_list[i];
    DLINK_FOREACH(ptr, tklist->head) {
      aconf = (struct ConfItem *) ptr->data;
      cnm_t = parse_netmask(aconf->host, &caddr, &cbits);
      if (cnm_t != nm_t || (user && irccmp(user, aconf->user)))
        continue;
      if ((nm_t == HM_HOST && !irccmp(aconf->host, host)) ||
        (nm_t == HM_IPV4 && bits == cbits
          && comp_with_mask(&IN_ADDR(addr), &IN_ADDR(caddr), bits))
#ifdef IPV6
        || (nm_t == HM_IPV6 && bits == cbits && comp_with_mask(&IN_ADDR(addr),
          &IN_ADDR(caddr), bits))
#endif
        ) {
        dlinkDestroy(ptr, tklist);
        delete_one_address_conf(aconf->host, aconf);
        return YES;
      }
    }
  }
  return NO;
}
static dlink_list *tdline_list[] = {
  &tdline_hour,
  &tdline_day,
  &tdline_min,
  &tdline_week,
  NULL
};
/* remove_temp_dline
 * inputs       - hostname to undline
 * side effects - tries to undline anything that matches
 */
static int remove_temp_dline(const char *host)
{
  dlink_list *tdlist;
  struct ConfItem *aconf;
  dlink_node *ptr;
  struct irc_inaddr addr, caddr;
  int nm_t, cnm_t, bits, cbits;
  int i;
  nm_t = parse_netmask(host, &addr, &bits);
  for (i = 0; tdline_list[i] != NULL; i++) {
    tdlist = tdline_list[i];
    DLINK_FOREACH(ptr, tdlist->head) {
      aconf = ptr->data;
      cnm_t = parse_netmask(aconf->host, &caddr, &cbits);
      if (cnm_t != nm_t)
        continue;
      if ((nm_t == HM_HOST && !irccmp(aconf->host, host)) || (nm_t == HM_IPV4 ==
        HM_IPV4 && bits == cbits && comp_with_mask(&IN_ADDR(addr),
        &IN_ADDR(caddr), bits))
#ifdef IPV6
        || (nm_t == HM_IPV6 && bits == cbits && comp_with_mask(&IN_ADDR(addr),
          &IN_ADDR(caddr), bits))
#endif
        ) {
        dlinkDestroy(ptr, tdlist);
        delete_one_address_conf(aconf->host, aconf);
        return YES;
      }
    }
  }
  return NO;
}
/* remove_temp_gline
 * inputs       - username, hostname to ungline
 * side effects - tries to ungline anything that matches
 */
static int remove_temp_gline(const char *user, const char *host)
{
  struct ConfItem *aconf;
  dlink_node *ptr;
  struct irc_inaddr addr, caddr;
  int nm_t, cnm_t, bits, cbits;
  nm_t = parse_netmask(host, &addr, &bits);
  DLINK_FOREACH(ptr, glines.head) {
    aconf = (struct ConfItem *) ptr->data;
    cnm_t = parse_netmask(aconf->host, &caddr, &cbits);
    if (cnm_t != nm_t || (user && irccmp(user, aconf->user)))
      continue;
    if ((nm_t == HM_HOST && !irccmp(aconf->host, host)) || (nm_t == HM_IPV4 &&
        bits == cbits && comp_with_mask(&IN_ADDR(addr), &IN_ADDR(caddr), bits))
#ifdef IPV6
      || (nm_t == HM_IPV6 && bits == cbits
        && comp_with_mask(&IN_ADDR(addr), &IN_ADDR(caddr), bits))
#endif
      ) {
      dlinkDestroy(ptr, &glines);
      delete_one_address_conf(aconf->host, aconf);
      return YES;
    }
  }
  return NO;
}
/* m_undline
 * added May 28th 2000 by Toby Verrall <toot@melnet.co.uk>
 * based totally on m_unkline
 * added to hybrid-7 7/11/2000 -is
 *  parv[0] = sender nick
 *  parv[1] = dline to remove
 */
static void mo_undline(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  FBFILE *in;
  FBFILE *out;
  char buf[BUFSIZE], buff[BUFSIZE], temppath[BUFSIZE], *p;
  const char *filename, *found_cidr;
  char *cidr;
  int pairme = NO, error_on_write = NO;
  mode_t oldumask;
  snprintf(temppath, sizeof(temppath), "%s.tmp", ConfigFileEntry.dlinefile);
  if (!IsOperUnkline(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to remove D-Lines", me.name, parv[0]);
    return;
  }
  cidr = parv[1];
#if 0
  if ((type = parse_netmask(cidr, &ip_host, &ip_mask)) == HM_HOST) {
    sendto_one(source_p, ":%s NOTICE %s :Invalid parameters", me.name, parv[0]);
    return;
  }
#endif
  if (remove_temp_dline(cidr)) {
    sendto_one(source_p, ":%s NOTICE %s :Temporary D-Line for %s removed",
      me.name, parv[0], cidr);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s removed Temporary D-Line for %s",
      get_oper_name(source_p), cidr);
    ilog(L_NOTICE, "UNDLINE %s by %s", cidr, parv[0]);
    return;
  }
  filename = get_conf_name(DLINE_TYPE);
  if ((in = fbopen(filename, "r")) == 0) {
    sendto_one(source_p, ":%s NOTICE %s :Couldn't write %s", me.name, parv[0],
      filename);
    return;
  }
  oldumask = umask(0); /* ircd is normally too paranoid */
  if ((out = fbopen(temppath, "w")) == 0) {
    sendto_one(source_p, ":%s NOTICE %s :Couldn't write D-Line file, aborted",
      me.name, parv[0]);
    fbclose(in);
    umask(oldumask); /* Restore the old umask */
    return;
  }
  umask(oldumask); /* Restore the old umask */
  while (fbgets(buf, sizeof(buf), in)) {
    strlcpy(buff, buf, sizeof(buff));
    if ((p = strchr(buff, '\n')) != NULL)
      *p = '\0';
    if ((*buff == '\0') || (*buff == '#')) {
      if (!error_on_write)
        flush_write(source_p, out, buf, temppath);
      continue;
    }
    if ((found_cidr = getfield(buff)) == NULL) {
      if (!error_on_write)
        flush_write(source_p, out, buf, temppath);
      continue;
    }
    if (irccmp(found_cidr, cidr) == 0)
      pairme++;
    else {
      if (!error_on_write)
        flush_write(source_p, out, buf, temppath);
      continue;
    }
  }
  fbclose(in);
  fbclose(out);
  if (!error_on_write) {
    (void)rename(temppath, filename);
    rehash(0);
  }
  else {
    sendto_one(source_p, ":%s NOTICE %s :Couldn't write D-line file, aborted",
      me.name, parv[0]);
    return;
  }
  if (!pairme) {
    sendto_one(source_p, ":%s NOTICE %s :There is no D-Line for %s", me.name,
      parv[0], cidr);
    return;
  }
  sendto_one(source_p, ":%s NOTICE %s :D-Line for %s removed", me.name, parv[0],
    cidr);
  sendto_realops_flags(UMODE_ALL, L_ALL, "%s removed D-Line for %s",
    get_oper_name(source_p), cidr);
  ilog(L_NOTICE, "UNDLINE %s by %s", cidr, log_client_name(source_p));
}
/* m_ungline
 * added May 29th 2000 by Toby Verrall <toot@melnet.co.uk>
 * added to hybrid-7 7/11/2000 -is
 *  parv[0] = sender nick
 *  parv[1] = gline to remove
 */
static void mo_ungline(struct Client * client_p, struct Client * source_p,
  int parc, char *parv[])
{
  const char *user;
  char *host;
  if (!ConfigFileEntry.glines) {
    sendto_one(source_p, ":%s NOTICE %s :This command has been disabled.",
      me.name, parv[0]);
    return;
  }
  if (!IsOperUnkline(source_p) || !IsOperGline(source_p)) {
    sendto_one(source_p, ":%s NOTICE %s :You do not have sufficient "
      "priviledges to remove G-Lines", me.name, parv[0]);
    return;
  }
  if ((host = strchr(parv[1], '@')) || *parv[1] == '*') {
    /* Explicit user@host mask given */
    if (host) {
      user = parv[1]; /* user */
      *(host++) = '\0'; /* host */
    }
    else {
      user = "*"; /* no @, assume *@somehost */
      host = parv[1];
    }
  }
  else {
    sendto_one(source_p, ":%s NOTICE %s :Invalid parameters", me.name, parv[0]);
    return;
  }
  if (remove_temp_gline(user, host)) {
    sendto_one(source_p, ":%s NOTICE %s :G-Line for %s@%s removed", me.name,
      parv[0], user, host);
    sendto_realops_flags(UMODE_ALL, L_ALL, "%s removed G-Line for %s@%s",
      get_oper_name(source_p), user, host);
    ilog(L_NOTICE, "UNGLINE %s@%s by %s@%s", user, host,
      get_oper_name(source_p));
    return;
  }
  else {
    sendto_one(source_p, ":%s NOTICE %s :There is no G-Line for %s@%s", me.name,
      parv[0], user, host);
    return;
  }
}
