/* control.c: main servlink control mechanism
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "setup.h"
#include <sys/types.h>
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#ifdef HAVE_LIBCRYPTO
#include <openssl/evp.h>
#include <openssl/err.h>
#endif
#ifdef HAVE_LIBZ
#include <zlib.h>
#endif
#include "servlink.h"
#include "io.h"
#include "control.h"

static cmd_handler cmd_set_zip_out_level;
static cmd_handler cmd_start_zip_out;
static cmd_handler cmd_start_zip_in;
static cmd_handler cmd_set_crypt_in_cipher;
static cmd_handler cmd_set_crypt_in_key;
static cmd_handler cmd_start_crypt_in;
static cmd_handler cmd_set_crypt_out_cipher;
static cmd_handler cmd_set_crypt_out_key;
static cmd_handler cmd_start_crypt_out;
static cmd_handler cmd_init;
struct command_def command_table[] = {
  {CMD_SET_ZIP_OUT_LEVEL, cmd_set_zip_out_level, COMMAND_FLAG_DATA},
  {CMD_START_ZIP_OUT, cmd_start_zip_out, 0},
  {CMD_START_ZIP_IN, cmd_start_zip_in, 0},
  {CMD_SET_CRYPT_IN_CIPHER, cmd_set_crypt_in_cipher, COMMAND_FLAG_DATA},
  {CMD_SET_CRYPT_IN_KEY, cmd_set_crypt_in_key, COMMAND_FLAG_DATA},
  {CMD_START_CRYPT_IN, cmd_start_crypt_in, 0},
  {CMD_SET_CRYPT_OUT_CIPHER, cmd_set_crypt_out_cipher, COMMAND_FLAG_DATA},
  {CMD_SET_CRYPT_OUT_KEY, cmd_set_crypt_out_key, COMMAND_FLAG_DATA},
  {CMD_START_CRYPT_OUT, cmd_start_crypt_out, 0},
  {CMD_INJECT_RECVQ, process_recvq, COMMAND_FLAG_DATA},
  {CMD_INJECT_SENDQ, process_sendq, COMMAND_FLAG_DATA},
  {CMD_INIT, cmd_init, 0},
  {CMD_ZIPSTATS, send_zipstats, 0},
  {0, 0, 0}
};

void cmd_set_zip_out_level(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBZ
  out_state.zip_state.level = *cmd->data;
  if (out_state.zip_state.level < -1 || out_state.zip_state.level > 9)
    send_error("invalid compression level %d", out_state.zip_state.level);
#else
  send_error("zlib is unsupported, can't set compression level");
#endif
}

void cmd_start_zip_out(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBZ
  int ret;
  if (out_state.zip)
    send_error("compression active, can't start again");
  out_state.zip_state.z_stream.total_in = 0;
  out_state.zip_state.z_stream.total_out = 0;
  out_state.zip_state.z_stream.zalloc = (alloc_func) 0;
  out_state.zip_state.z_stream.zfree = (free_func) 0;
  out_state.zip_state.z_stream.data_type = Z_ASCII;
  if (out_state.zip_state.level <= 0)
    out_state.zip_state.level = Z_DEFAULT_COMPRESSION;
  if ((ret = deflateInit(&out_state.zip_state.z_stream,
      out_state.zip_state.level)) != Z_OK)
    send_error("initialization of deflation functions failed: %s", zError(ret));
  out_state.zip = 1;
#else
   send_error("zlib is unsupported, can't start compression");
#endif
}

void cmd_start_zip_in(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBZ
  int ret;
  if (in_state.zip)
    send_error("decompression active, can't start again");
  in_state.zip_state.z_stream.total_in = 0;
  in_state.zip_state.z_stream.total_out = 0;
  in_state.zip_state.z_stream.zalloc = (alloc_func) 0;
  in_state.zip_state.z_stream.zfree = (free_func) 0;
  in_state.zip_state.z_stream.data_type = Z_ASCII;
  if ((ret = inflateInit(&in_state.zip_state.z_stream)) != Z_OK)
    send_error("initialization of inflation functions failed: %s", zError(ret));
  in_state.zip = 1;
#else
   send_error("zlib is unsupported, can't start decompression");
#endif
}

void cmd_set_crypt_in_cipher(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBCRYPTO
  const unsigned int cipher = *cmd->data;
  if (in_state.crypt_state.cipher)
    send_error("decryption cipher can't be selected twice");
  switch (cipher) {
#ifdef HAVE_EVP_BF_CFB
   case CIPHER_BF:
    in_state.crypt_state.cipher = EVP_bf_cfb();
    break;
#endif
#ifdef HAVE_EVP_CAST5_CFB
   case CIPHER_CAST:
    in_state.crypt_state.cipher = EVP_cast5_cfb();
    break;
#endif
#ifdef HAVE_EVP_DES_CFB
   case CIPHER_DES:
    in_state.crypt_state.cipher = EVP_des_cfb();
    break;
#endif
#ifdef HAVE_EVP_DES_EDE3_CFB
   case CIPHER_3DES:
    in_state.crypt_state.cipher = EVP_des_ede3_cfb();
    break;
#endif
#ifdef HAVE_EVP_IDEA_CFB
   case CIPHER_IDEA:
    in_state.crypt_state.cipher = EVP_idea_cfb();
    break;
#endif
#ifdef HAVE_EVP_RC5_32_12_16_CFB
   case CIPHER_RC5_8:
    in_state.crypt_state.cipher = EVP_rc5_32_12_16_cfb();
    in_state.crypt_state.rounds = 8;
    break;
   case CIPHER_RC5_12:
    in_state.crypt_state.cipher = EVP_rc5_32_12_16_cfb();
    in_state.crypt_state.rounds = 12;
    break;
   case CIPHER_RC5_16:
    in_state.crypt_state.cipher = EVP_rc5_32_12_16_cfb();
    in_state.crypt_state.rounds = 16;
    break;
#endif
   default:
    send_error("decryption cipher %d is unsupported by this installation of "
      "OpenSSL.", cipher);
    break;
  }
#else
  send_error("OpenSSL is unsupported, can't set decryption cipher");
#endif
}

void cmd_set_crypt_in_key(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBCRYPTO
  if (in_state.crypt_state.key)
    send_error("decryption key can't be set twice");
  in_state.crypt_state.keylen = cmd->datalen;
  in_state.crypt_state.key = malloc(cmd->datalen);
  memcpy(in_state.crypt_state.key, cmd->data, cmd->datalen);
#else
  send_error("OpenSSL is unsupported, can't set decryption key");
#endif
}

void cmd_start_crypt_in(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBCRYPTO
  if (in_state.crypt)
    send_error("decryption active, can't start again");
  if (!in_state.crypt_state.cipher)
    send_error("preferred decryption cipher unknown");
  if (!in_state.crypt_state.key)
    send_error("decryption key missing");
  in_state.crypt = 1;
  if (!EVP_DecryptInit(&in_state.crypt_state.ctx, in_state.crypt_state.cipher,
      NULL, NULL))
    send_error("initialization of decryption functions (1) failed: %s",
      ERR_error_string(ERR_get_error(), NULL));

  /* XXX - ugly hack to work around OpenSSL bug and makes BF/256 work. */
  /* cast to avoid warning */
  *(unsigned int *)(&in_state.crypt_state.ctx.cipher->flags) |=
    EVP_CIPH_VARIABLE_LENGTH;
  if (!EVP_CIPHER_CTX_set_key_length(&in_state.crypt_state.ctx,
      in_state.crypt_state.keylen))
    send_error("can't start decryption - set_key_length failed: %s",
      ERR_error_string(ERR_get_error(), NULL));
  in_state.crypt_state.ivlen =
    EVP_CIPHER_CTX_iv_length(&in_state.crypt_state.ctx);
  if (in_state.crypt_state.ivlen)
    in_state.crypt_state.iv = calloc(in_state.crypt_state.ivlen, 1);
  if (in_state.crypt_state.rounds) {
    if (!EVP_CIPHER_CTX_ctrl(&in_state.crypt_state.ctx, EVP_CTRL_SET_RC5_ROUNDS,
        in_state.crypt_state.rounds, NULL))
      send_error("can't start decryption - SET_RC5_ROUNDS failed: %s",
        ERR_error_string(ERR_get_error(), NULL));
  }
  if (!EVP_DecryptInit(&in_state.crypt_state.ctx, NULL,
      in_state.crypt_state.key, in_state.crypt_state.iv))
    send_error("initialization of decryption functions (2) failed: %s",
      ERR_error_string(ERR_get_error(), NULL));
#else
  send_error("OpenSSL is unsupported, can't start decryption");
#endif
}

void cmd_set_crypt_out_cipher(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBCRYPTO
  const unsigned int cipher = *cmd->data;
  if (out_state.crypt_state.cipher)
    send_error("encryption cipher can't be selected twice");
  switch (cipher) {
#ifdef HAVE_EVP_BF_CFB
   case CIPHER_BF:
    out_state.crypt_state.cipher = EVP_bf_cfb();
    break;
#endif
#ifdef HAVE_EVP_CAST5_CFB
   case CIPHER_CAST:
    out_state.crypt_state.cipher = EVP_cast5_cfb();
    break;
#endif
#ifdef HAVE_EVP_DES_CFB
   case CIPHER_DES:
    out_state.crypt_state.cipher = EVP_des_cfb();
    break;
#endif
#ifdef HAVE_EVP_DES_EDE3_CFB
   case CIPHER_3DES:
    out_state.crypt_state.cipher = EVP_des_ede3_cfb();
    break;
#endif
#ifdef HAVE_EVP_IDEA_CFB
   case CIPHER_IDEA:
    out_state.crypt_state.cipher = EVP_idea_cfb();
    break;
#endif
#ifdef HAVE_EVP_RC5_32_12_16_CFB
   case CIPHER_RC5_8:
    out_state.crypt_state.cipher = EVP_rc5_32_12_16_cfb();
    out_state.crypt_state.rounds = 8;
    break;
   case CIPHER_RC5_12:
    out_state.crypt_state.cipher = EVP_rc5_32_12_16_cfb();
    out_state.crypt_state.rounds = 12;
    break;
   case CIPHER_RC5_16:
    out_state.crypt_state.cipher = EVP_rc5_32_12_16_cfb();
    out_state.crypt_state.rounds = 16;
    break;
#endif
   default:
    send_error("encryption cipher %d unsupported by this installation of "
      "OpenSSL", cipher);
  }
#else
  send_error("OpenSSL is unsupported, can't set encryption cipher");
#endif
}

void cmd_set_crypt_out_key(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBCRYPTO
  if (out_state.crypt_state.key)
    send_error("encryption key can't be set twice");
  out_state.crypt_state.keylen = cmd->datalen;
  out_state.crypt_state.key = malloc(cmd->datalen);
  memcpy(out_state.crypt_state.key, cmd->data, cmd->datalen);
#else
  send_error("OpenSSL is unsupported, can't set encryption key");
#endif
}

void cmd_start_crypt_out(struct ctrl_command * cmd)
{
#ifdef HAVE_LIBCRYPTO
  if (out_state.crypt)
    send_error("encryption active, can't start again");
  if (!out_state.crypt_state.cipher)
    send_error("preferred encryption cipher unknown");
  if (!out_state.crypt_state.key)
    send_error("encryption key missing");
  out_state.crypt = 1;
  if (!EVP_EncryptInit(&out_state.crypt_state.ctx, out_state.crypt_state.cipher,
      NULL, NULL))
    send_error("initialization of encryption functions (1) failed: %s",
      ERR_error_string(ERR_get_error(), NULL));

  /* XXX - ugly hack to work around OpenSSL bug and makes BF/256 work. */
  /* cast to avoid warning */
  *(unsigned int *)(&out_state.crypt_state.ctx.cipher->flags) |=
    EVP_CIPH_VARIABLE_LENGTH;
  if (!EVP_CIPHER_CTX_set_key_length(&out_state.crypt_state.ctx,
      out_state.crypt_state.keylen))
    send_error("can't start encryption - set_key_length failed: %s",
      ERR_error_string(ERR_get_error(), NULL));
  out_state.crypt_state.ivlen = EVP_CIPHER_CTX_iv_length(&out_state.crypt_state.ctx);
  if (out_state.crypt_state.ivlen)
    out_state.crypt_state.iv = calloc(out_state.crypt_state.ivlen, 1);
  if (out_state.crypt_state.rounds) {
    if (!EVP_CIPHER_CTX_ctrl(&out_state.crypt_state.ctx,
        EVP_CTRL_SET_RC5_ROUNDS,
        out_state.crypt_state.rounds, NULL))
      send_error("can't start encryption - SET_RC5_ROUNDS failed: %s",
        ERR_error_string(ERR_get_error(), NULL));
  }
  if (!EVP_EncryptInit(&out_state.crypt_state.ctx,
      NULL, out_state.crypt_state.key, out_state.crypt_state.iv))
    send_error("initialization of encryption functions (2) failed: %s",
      ERR_error_string(ERR_get_error(), NULL));
#else
  send_error("OpenSSL is unsupported, can't start encryption");
#endif
}

void cmd_init(struct ctrl_command * cmd)
{
  if (in_state.active || out_state.active)
    send_error("command initialization must only be performed once!");
  in_state.active = 1;
  out_state.active = 1;
  CONTROL_R.read_cb = read_ctrl;
  CONTROL_W.write_cb = NULL;
  LOCAL_R.read_cb = read_data;
  LOCAL_W.write_cb = NULL;
  REMOTE_R.read_cb = read_net;
  REMOTE_W.write_cb = NULL;
}
