/* balloc.c: A block allocator.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 *  File:   blalloc.c
 *  Owner:  Wohali (Joan Touzet)
 *
 *  Modified 2001/11/29 for mmap() support by Aaron Sethman <androsyn@ratbox.org>
 *
 *  This file is licensed under the GNU General Public License, which is
 *  distributed as LICENSE.
 */
/*
 * About the block allocator
 *
 * Basically we have three ways of getting memory off of the operating
 * system. Below are this list of methods and the order of preference.
 *
 * 1. mmap() anonymous pages with the MMAP_ANON flag.
 * 2. mmap() via the /dev/zero trick.
 * 3. malloc()
 *
 * The advantages of 1 and 2 are this.  We can munmap() the pages which will
 * return the pages back to the operating system, thus reducing the size of the
 * process as the memory is unused.  malloc() on many systems just keeps a heap
 * of memory to itself, which never gets given back to the OS, except on exit.
 * This of course is bad, if say we have an event that causes us to allocate
 * say, 200MB of memory, while our normal memory consumption would be 15MB.
 *
 * Of course it is up to the caller to make sure BlockHeapGarbageCollect() gets
 * called periodically to do this cleanup, otherwise you'll keep the memory in
 * the process.
 */

#include "stdinc.h"
#define WE_ARE_MEMORY_C
#include "setup.h"
#include "ircd_defs.h"
#include "ircd.h"
#include "memory.h"
#include "balloc.h"
#include "irc_string.h"
#include "tools.h"
#include "s_log.h"
#include "client.h"
#include "fdlist.h"
#include "event.h"
#ifdef HAVE_MMAP
#include <sys/mman.h>
/* HP-UX sucks */
#ifdef MAP_ANONYMOUS
#ifndef MAP_ANON
#define MAP_ANON MAP_ANONYMOUS
#endif
#endif
#endif

static int newblock(BlockHeap * bh);
static void block_heap_gc(void *unused);
static dlink_list heap_lists;
static int BlockHeapGarbageCollect(BlockHeap *);
#if defined(HAVE_MMAP) && !defined(MAP_ANON)
static int zero_fd = -1;
#endif
#define blockheap_fail(x) _blockheap_fail(x, __FILE__, __LINE__)
static void _blockheap_fail(const char *reason, const char *file, int line)
{
  ilog(L_CRIT, "Blockheap failure: %s (%s:%d)", reason, file, line);
  abort();
}

/* inline void free_block(void *ptr, size_t size)
 *
 * Inputs: The block and its size
 * Side Effects: Returns memory for the block back to the OS
 */
static inline void free_block(void *ptr, size_t size)
{
#ifdef HAVE_MMAP
  munmap(ptr, size);
#else
  free(ptr);
#endif
}

/* initBlockHeap - initializes block heap */
void initBlockHeap(void)
{
#if defined(HAVE_MMAP) && !defined(MAP_ANON)
  zero_fd = open("/dev/zero", O_RDWR);
  if (zero_fd < 0)
    blockheap_fail("Failed opening /dev/zero");
  fd_open(zero_fd, FD_FILE, "Anonymous mmap()");
#endif
  eventAddIsh("block_heap_gc", block_heap_gc, NULL, 30);
}
/* static inline void *get_block(size_t size)
 * Input: Size of block to allocate
 * Output: Pointer to new block
 */
static inline void * get_block(size_t size)
{
  void *ptr;
#ifdef HAVE_MMAP
#ifdef MAP_ANON
  ptr = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON, -1, 0);
#else
  ptr = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_PRIVATE, zero_fd, 0);
#endif
  if (ptr == MAP_FAILED) {
    ptr = NULL;
  }
#else
  ptr = malloc(size);
#endif
  return ptr;
}

static void block_heap_gc(void *unused)
{
  dlink_node *ptr;
  DLINK_FOREACH(ptr, heap_lists.head) {
    BlockHeapGarbageCollect((BlockHeap *) ptr->data);
  }
}
/* newblock - allocates new block for addition to a blockheap
 * input  - pointer to parent blockheap
 * output - success, 0
 *        - failure, 1
 */
static int newblock(BlockHeap * bh)
{
  MemBlock *newblk;
  Block *b;
  unsigned long i;
  void *offset;
  /* Setup the initial data structure. */
  b = (Block *) calloc(1, sizeof(Block));
  if (b == NULL) {
    return 1;
  }
  b->free_list.head = b->free_list.tail = NULL;
  b->used_list.head = b->used_list.tail = NULL;
  b->next = bh->base;
  b->alloc_size = (bh->elemsPerBlock + 1) * (bh->elemSize + sizeof(MemBlock));
  b->elems = get_block(b->alloc_size);
  if (b->elems == NULL)
    return 1;
  offset = b->elems;
  /* Setup our blocks now */
  for (i = 0; i < bh->elemsPerBlock; i++) {
    void *data;
    newblk = (void *)offset;
    newblk->block = b;
    data = (void *)((size_t) offset + sizeof(MemBlock));
    newblk->block = b;
    dlinkAdd(data, &newblk->self, &b->free_list);
    offset = (unsigned char *)((unsigned char *)offset + bh->elemSize +
      sizeof(MemBlock));
  }
  ++bh->blocksAllocated;
  bh->freeElems += bh->elemsPerBlock;
  bh->base = b;
  return 0;
}

/* BlockHeapCreate - creates new blockheap from which smaller blocks can be
 *                   allocated. used instead of multiple calls to malloc() for
 *                   performance reasons.
 * input  - size of basic element to be stored
 *        - number of elements to be stored in a single block of memory.
 * output - success, pointer to new BlockHeap
 *        - failure, NULL
 *
 * when the blockheap runs out of free memory, it will allocate elemsize *
 * elemsperblock more
 */
BlockHeap * BlockHeapCreate(size_t elemsize, int elemsperblock)
{
  BlockHeap *bh;
  s_assert(elemsize > 0 && elemsperblock > 0);
  /* Catch idiotic requests up front */
  if ((elemsize <= 0) || (elemsperblock <= 0)) {
    blockheap_fail("Attempting to BlockHeapCreate idiotic sizes");
  }
  /* Allocate our new BlockHeap */
  bh = (BlockHeap *) calloc(1, sizeof(BlockHeap));
  if (bh == NULL) {
    ilog(L_CRIT, "Attempt to calloc() failed: (%s:%d)", __FILE__, __LINE__);
    outofmemory(); /* die.. out of memory */
  }
  if ((elemsize % sizeof(void *)) != 0) {
    /* Pad to even pointer boundary */
    elemsize += sizeof(void *);
    elemsize &= ~(sizeof(void *) - 1);
  }
  bh->elemSize = elemsize;
  bh->elemsPerBlock = elemsperblock;
  bh->blocksAllocated = 0;
  bh->freeElems = 0;
  bh->base = NULL;
  /* Be sure our malloc was successful */
  if (newblock(bh)) {
    if (bh != NULL)
      free(bh);
    ilog(L_CRIT, "newblock() failed");
    outofmemory(); /* die.. out of memory */
  }
  if (bh == NULL)
    blockheap_fail("bh == NULL when it shouldn't be");
  dlinkAdd(bh, &bh->hlist, &heap_lists);
  return bh;
}

/* BlockHeapAlloc - finds a free struct within our BlockHeap
 * input  - pointer to the BlockHeap
 * output - success, pointer to structure within BlockHeap
 *        - failure, NULL
 */
void * BlockHeapAlloc(BlockHeap * bh)
{
  Block *walker;
  dlink_node *new_node;
  s_assert(bh != NULL);
  if (bh == NULL) {
    blockheap_fail("Cannot allocate if bh == NULL");
  }
  if (bh->freeElems == 0) {
    /* Allocate new block and assign */
    /* newblock returns 1 if unsuccessful, 0 if not */
    if (newblock(bh)) {
      /* That didn't work..try to garbage collect */
      BlockHeapGarbageCollect(bh);
      if (bh->freeElems == 0) {
        ilog(L_CRIT, "newblock() failed and garbage collection didn't help");
        outofmemory(); /* Well that didn't work either...bail */
      }
    }
  }
  for (walker = bh->base; walker != NULL; walker = walker->next) {
    if (dlink_list_length(&walker->free_list) > 0) {
      bh->freeElems--;
      new_node = walker->free_list.head;
      dlinkMoveNode(new_node, &walker->free_list, &walker->used_list);
      s_assert(new_node->data != NULL);
      if (new_node->data == NULL)
        blockheap_fail("new_node->data is NULL and that shouldn't happen!!!");
      return new_node->data;
    }
  }
  blockheap_fail("BlockHeapAlloc failed, giving up");
  return NULL;
}

/* BlockHeapFree - returns element to free pool without free()'ing
 * input  - pointer to the BlockHeap
 *        - pointer to element being 'freed'
 * output - success, 0
 *        - failure, 1
 */
int BlockHeapFree(BlockHeap * bh, void *ptr)
{
  Block *block;
  struct MemBlock *memblock;
  s_assert(bh != NULL);
  s_assert(ptr != NULL);
  if (bh == NULL) {
    ilog(L_NOTICE, "balloc.c:BlockHeapFree() bh == NULL");
    return 1;
  }
  if (ptr == NULL) {
    ilog(L_NOTICE, "balloc.BlockHeapFree() ptr == NULL");
    return 1;
  }
  memblock = (void *)((size_t) ptr - sizeof(MemBlock));
  s_assert(memblock->block != NULL);
  if (memblock->block == NULL) {
    blockheap_fail("memblock->block == NULL, not a valid block?");
    outofmemory();
  }
  block = memblock->block;
  bh->freeElems++;
  mem_frob(ptr, bh->elemSize);
  dlinkMoveNode(&memblock->self, &block->used_list, &block->free_list);
  return 0;
}

/* BlockHeapGarbageCollect - performs garbage collection on the block heap.
 *                           removes unallocated blocks from the heap. the root
 *                           node is never removed by garbage collection
 * input  - pointer to the BlockHeap
 * output - success, 0
 *        - failure, 1
 */
static int BlockHeapGarbageCollect(BlockHeap * bh)
{
  Block *walker, *last;
  if (bh == NULL) {
    return 1;
  }
  if (bh->freeElems < bh->elemsPerBlock || bh->blocksAllocated == 1) {
    /* There couldn't possibly be an entire free block.  Return. */
    return 0;
  }
  last = NULL;
  walker = bh->base;
  while (walker != NULL) {
    if ((dlink_list_length(&walker->free_list) == bh->elemsPerBlock) != 0) {
      free_block(walker->elems, walker->alloc_size);
      if (last != NULL) {
        last->next = walker->next;
        if (walker != NULL)
          free(walker);
        walker = last->next;
      }
      else {
        bh->base = walker->next;
        if (walker != NULL)
          free(walker);
        walker = bh->base;
      }
      bh->blocksAllocated--;
      bh->freeElems -= bh->elemsPerBlock;
    }
    else {
      last = walker;
      walker = walker->next;
    }
  }
  return 0;
}

/* BlockHeapDestroy - completely free()'s a BlockHeap.
 * input  - pointer to the BlockHeap
 * output - success, 0
 *        - failure, 1
 */
int BlockHeapDestroy(BlockHeap * bh)
{
  Block *walker, *next;
  if (bh == NULL) {
    return 1;
  }
  for (walker = bh->base; walker != NULL; walker = next) {
    next = walker->next;
    free_block(walker->elems, walker->alloc_size);
    if (walker != NULL)
      free(walker);
  }
  dlinkDelete(&bh->hlist, &heap_lists);
  free(bh);
  return 0;
}
void BlockHeapUsage(BlockHeap * bh, size_t * bused, size_t * bfree, size_t * bmemusage)
{
  size_t used;
  size_t freem;
  size_t memusage;
  if (bh == NULL) {
    return;
  }
  freem = bh->freeElems;
  used = (bh->blocksAllocated * bh->elemsPerBlock) - bh->freeElems;
  memusage = used * (bh->elemSize + sizeof(MemBlock));
  if (bused != NULL)
    *bused = used;
  if (bfree != NULL)
    *bfree = freem;
  if (bmemusage != NULL)
    *bmemusage = memusage;
}
