/* event.c: Event functions.
 *
 * Copyright (C) 1998-2000 Regents of the University of California
 * Copyright (C) 2001-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * Code borrowed from the squid web cache by Adrian Chadd.
 *
 * Original header:
 *
 * DEBUG: section 41   Event Processing
 * AUTHOR: Henrik Nordstrom
 *
 * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
 * ----------------------------------------------------------
 * Squid is the result of efforts by numerous individuals from the Internet
 * community.  Development is led by Duane Wessels of the National Laboratory
 * for Applied Network Research and funded by the National Science Foundation.
 * Squid is Copyrighted (C) 1998 by the Regents of the University of California.
 * Please see the COPYRIGHT file for full details. Squid incorporates software
 * developed and/or copyrighted by other sources. Please see the CREDITS file
 * for full details.
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

/* Events are added to the static array event_table with a frequency telling
 * eventRun how often to execute it.
 */

#include "stdinc.h"
#include "config.h"
#include "ircd.h"
#include "event.h"
#include "client.h"
#include "send.h"
#include "memory.h"
#include "s_log.h"
#include "numeric.h"

static const char *last_event_ran = NULL;
struct ev_entry event_table[MAX_EVENTS];
static time_t event_time_min = -1;

/* eventAdd[Once] - add event to event list
 * input        - Name of event, function to call, arguments to pass, and
 *                frequency of the event.
 */
void eventAdd(const char *name, EVH * func, void *arg, time_t when)
{
  int i;
  /* find first inactive index */
  for (i = 0; i < MAX_EVENTS; i++) {
    if (event_table[i].active == 0) {
      event_table[i].func = func;
      event_table[i].name = name;
      event_table[i].arg = arg;
      event_table[i].when = CurrentTime + when;
      event_table[i].frequency = when;
      event_table[i].active = 1;
      if ((event_table[i].when < event_time_min) || (event_time_min == -1))
        event_time_min = event_table[i].when;
      return;
    }
  }
  sendto_realops_flags(UMODE_DEBUG, L_ALL, "Failed to add event \"%s\" to "
    "event table!", name);
}
void eventAddOnce(const char *name, EVH * func, void *arg, time_t when)
{
  int i;
  /* find first inactive index */
  for (i = 0; i < MAX_EVENTS; i++) {
    if (event_table[i].active == 0) {
      event_table[i].func = func;
      event_table[i].name = name;
      event_table[i].arg = arg;
      event_table[i].when = CurrentTime + when;
      event_table[i].frequency = 0;
      event_table[i].active = 1;
      if ((event_table[i].when < event_time_min) || (event_time_min == -1))
        event_time_min = event_table[i].when;
      return;
    }
  }
  sendto_realops_flags(UMODE_DEBUG, L_ALL, "Failed to add event \"%s\" to "
    "event table!", name);
}

/* eventAddIsh - less precise timing (+-1/3 of specified frequency)
 * input - Name of event, function to call, arguments to pass, and frequency of
 *         the event.
 */
void eventAddIsh(const char *name, EVH * func, void *arg, time_t delta_ish)
{
  if (delta_ish >= 3.0) {
    const time_t two_third = (2 * delta_ish) / 3;
    delta_ish = two_third + ((rand() % 1000) * two_third) / 1000;
  }
  eventAdd(name, func, arg, delta_ish);
}

/* eventDelete
 * input        - Function handler, argument that was passed.
 * side effects - Removes the event from the event list
 */
void eventDelete(EVH * func, void *arg)
{
  int i;
  i = eventFind(func, arg);
  if (i == -1)
    return;
  event_table[i].name = NULL;
  event_table[i].func = NULL;
  event_table[i].arg = NULL;
  event_table[i].active = 0;
}

/* eventRun - runs pending events */
void eventRun(void)
{
  int i;
  for (i = 0; i < MAX_EVENTS; i++) {
    if (event_table[i].active && (event_table[i].when <= CurrentTime)) {
      last_event_ran = event_table[i].name;
      event_table[i].func(event_table[i].arg);
      /* if the event is only scheduled to run once, remove it from the table */
      if (event_table[i].frequency)
        event_table[i].when = CurrentTime + event_table[i].frequency;
      else {
        event_table[i].name = NULL;
        event_table[i].func = NULL;
        event_table[i].arg = NULL;
        event_table[i].active = 0;
      }
      event_time_min = -1;
    }
  }
}

/* eventNextTime - return time of next eventRun() */
time_t eventNextTime(void)
{
  int i;
  if (event_time_min == -1) {
    for (i = 0; i < MAX_EVENTS; i++) {
      if (event_table[i].active && ((event_table[i].when < event_time_min) ||
          (event_time_min == -1)))
        event_time_min = event_table[i].when;
    }
  }
  return event_time_min;
}

/* eventInit - initializes event system */
void eventInit(void)
{
  last_event_ran = NULL;
  memset((void *)event_table, 0, sizeof(event_table));
}

/* eventFind
 * input  - event function and argument passed to it
 * output - index to the slow in the event_table
 */
int eventFind(EVH * func, void *arg)
{
  int i;
  for (i = 0; i < MAX_EVENTS; i++) {
    if ((event_table[i].func == func) && (event_table[i].arg == arg) &&
        event_table[i].active)
      return i;
  }
  return -1;
}

/* show_events
 * input  - client requesting event
 * output - list of events
 */
void show_events(struct Client * source_p)
{
  int i;
  if (last_event_ran)
    sendto_one(source_p, ":%s %d %s E :Last event to run: %s", me.name,
      RPL_STATSDEBUG, source_p->name, last_event_ran);
  sendto_one(source_p, ":%s %d %s E :Operation                    "
    "Next Execution", me.name, RPL_STATSDEBUG, source_p->name);
  for (i = 0; i < MAX_EVENTS; i++) {
    if (event_table[i].active) {
      sendto_one(source_p, ":%s %d %s E :%-28s %d seconds", me.name,
        RPL_STATSDEBUG, source_p->name, event_table[i].name,
        (int)(event_table[i].when - CurrentTime));
    }
  }
}
/* set_back_events - sets back all events by n seconds
 * input - Time to set back events by
 */
void set_back_events(time_t by)
{
  int i;
  for (i = 0; i < MAX_EVENTS; i++) {
    if (event_table[i].when > by)
      event_table[i].when -= by;
    else
      event_table[i].when = 0;
  }
}
