/* fdlist.c: Maintains a list of file descriptors.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "config.h"
#include "fdlist.h"
#include "client.h"
#include "event.h"
#include "ircd.h"
#include "s_bsd.h"
#include "send.h"
#include "memory.h"
#include "numeric.h"
#include "s_log.h"

fde_t *fd_table = NULL;
static void fdlist_update_biggest(int fd, int opening);

/* Highest FD and number of open FDs */
int highest_fd = -1;
int number_fd = 0;
static void fdlist_update_biggest(int fd, int opening)
{
  if (fd < highest_fd)
    return;
  s_assert(fd < MAXCONNECTIONS);
  if (fd > highest_fd) {
    /* s_assert that we are not closing a FD bigger than our known biggest FD */
    s_assert(opening);
    highest_fd = fd;
    return;
  }
  /* s_assert that we are closing the biggest FD; we can't be re-opening it */
  s_assert(!opening);
  while (highest_fd >= 0 && !fd_table[highest_fd].flags.open)
    highest_fd--;
}

void fdlist_init(void)
{
  static int initialized = 0;
  if (!initialized) {
    /* Since we're doing this once .. */
    fd_table = MyMalloc((MAXCONNECTIONS + 1) * sizeof(fde_t));
    initialized = 1;
  }
}
/* Called to open a given filedescriptor */
void fd_open(int fd, unsigned int type, const char *desc)
{
  fde_t *F = &fd_table[fd];
  s_assert(fd >= 0);
  if (F->flags.open) {
#ifdef NOTYET
    debug(51, 1) ("WARNING: Closing open FD %4d\n", fd);
#endif
    fd_close(fd);
  }
  s_assert(!F->flags.open);
#ifdef NOTYET
  debug(51, 3) ("fd_open FD %d %s\n", fd, desc);
#endif
  F->fd = fd;
  F->type = type;
  F->flags.open = 1;
#ifdef NOTYET
  F->defer.until = 0;
  F->defer.n = 0;
  F->defer.handler = NULL;
#endif
  fdlist_update_biggest(fd, 1);
  F->comm_index = -1;
  F->list = FDLIST_NONE;
  if (desc)
    strncpy(F->desc, desc, FD_DESC_SZ);
  number_fd++;
}

/* Called to close a given filedescriptor */
void fd_close(int fd)
{
  fde_t *F = &fd_table[fd];
  s_assert(F->flags.open);
  /* All disk fd's MUST go through file_close() ! */
  s_assert(F->type != FD_FILE);
  if (F->type == FD_FILE) {
    s_assert(F->read_handler == NULL);
    s_assert(F->write_handler == NULL);
  }
#ifdef NOTYET
  debug(51, 3) ("fd_close FD %d %s\n", fd, F->desc);
#endif
  comm_setselect(fd, FDLIST_NONE, COMM_SELECT_WRITE | COMM_SELECT_READ, NULL,
    NULL, 0);
  if (F->dns_query != NULL) {
    delete_adns_queries(F->dns_query);
    MyFree(F->dns_query);
    F->dns_query = NULL;
  }
  F->flags.open = 0;
  fdlist_update_biggest(fd, 0);
  number_fd--;
  memset(F, '\0', sizeof(fde_t));
  F->timeout = 0;
  /* Unlike squid, we're actually closing the FD here! -- adrian */
  close(fd);
}

/* fd_dump() - dump the list of active filedescriptors */
void fd_dump(struct Client * source_p)
{
  int i;
  for (i = 0; i <= highest_fd; i++) {
    if (!fd_table[i].flags.open)
      continue;
    sendto_one(source_p, ":%s %d %s F :fd %-3d (%s)", me.name, RPL_STATSDEBUG,
      source_p->name, i, fd_table[i].desc);
  }
}
/* fd_note() - set the fd note. be careful not to overflow fd_table[fd].desc */
void fd_note(int fd, const char *format, ...)
{
  va_list args;
  if (format) {
    va_start(args, format);
    vsnprintf(fd_table[fd].desc, FD_DESC_SZ, format, args);
    va_end(args);
  }
  else
    fd_table[fd].desc[0] = '\0';
}
