/*
 *  hook.c: Provides a generic event hooking interface.
 *
 *  Copyright (C) 2000-2002 Edward Brocklesby, Hybrid Development Team
 *  Copyright (C) 2003-2004 ircd-immortal development team
 *
 *  This file is licensed under the GNU General Public License, which is
 *  distributed as LICENSE.
 */
/* used by modules to hook into events called by other parts of code
 * this allows modules like m_spy.c to know when commands are received and
 * processed by other modules (for example, m_motd.c)
 */

#include "stdinc.h"
#include "memory.h"
#include "tools.h"
#include "hook.h"
#include "common.h"

dlink_list hooks;
static hook* find_hook_byname(const char *);

#ifndef NDEBUG
int h_iosend_id;
int h_iorecv_id;
int h_iorecvctrl_id;
#endif
int h_burst_channel_id;
int h_doing_admin_id;
int h_doing_info_id;
int h_doing_links_id;
int h_doing_motd_id;
int h_doing_stats_id;
int h_doing_trace_id;
static int hook_curid = 0;

void init_hooks(void)
{
  memset(&hooks, 0, sizeof(hooks));
#ifndef NDEBUG
  hook_add_event("iosend", &h_iosend_id);
  hook_add_event("iorecv", &h_iorecv_id);
  hook_add_event("iorecvctrl", &h_iorecvctrl_id);
#endif
  hook_add_event("burst_channel", &h_burst_channel_id);
  hook_add_event("doing_admin", &h_doing_admin_id);
  hook_add_event("doing_info", &h_doing_info_id);
  hook_add_event("doing_links", &h_doing_links_id);
  hook_add_event("doing_motd", &h_doing_motd_id);
  hook_add_event("doing_stats", &h_doing_stats_id);
  hook_add_event("doing_trace", &h_doing_trace_id);
}

static hook * new_hook(const char *name, int *id)
{
  hook *h;
  h = MyMalloc(sizeof(hook));
  DupString(h->name, name);
  h->id = *id = hook_curid++;
  return h;
}

int hook_add_event(const char *name, int *id)
{
  hook *newhook;
	if((newhook = find_hook_byname(name)))
	{
		*id = newhook->id;
		return NO;
	}
  newhook = new_hook(name, id);
  dlinkAddAlloc(newhook, &hooks);
  return NO;
}

int hook_del_event(const char *name)
{
  dlink_node *node;
  hook *h;
  DLINK_FOREACH(node, hooks.head) {
    h = node->data;
    if (!strcmp(h->name, name)) {
      dlinkDelete(node, &hooks);
      MyFree(h);
      return NO;
    }
  }
  return NO;
}

static hook * find_hook_byname(const char *name)
{
  dlink_node *node;
  hook *h;

  DLINK_FOREACH(node, hooks.head)
  {
    h = node->data;
    if(!strcmp(h->name, name))
      return h;
  }
  return NULL;
}

static hook * find_hook_byid(int id)
{
	dlink_node *ptr;
	hook *h;

	DLINK_FOREACH(ptr, hooks.head)
	{
		h = ptr->data;
		if(h->id == id)
			return h;
	}
	return NULL;
}

int hook_del_hook(const char *event, hookfn fn)
{
  hook *h;
  dlink_node *node, *nnode;
  h = find_hook_byname(event);
  if (!h)
    return -1;
  DLINK_FOREACH_SAFE(node, nnode, h->hooks.head) {
    if (fn == (hookfn) node->data)
      dlinkDestroy(node, &h->hooks);
  }
  return NO;
}

int hook_add_hook(const char *event, hookfn fn)
{
  hook *h;
  h = find_hook_byname(event);
  if (!h)
    return -1;
  dlinkAddAlloc(fn, &h->hooks);
  return NO;
}

int hook_call_event(int id, void *data)
{
  hook *h;
  dlink_node *node;
  hookfn fn;
  h = find_hook_byid(id);
  if (!h)
    return -1;
  DLINK_FOREACH(node, h->hooks.head) {
    fn = (hookfn) (uintptr_t) node->data;
    if (fn(data) != 0)
      return NO;
  }
  return NO;
}
