/* hostmask.c: Code to efficiently find IP & hostmask based configs.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "memory.h"
#include "ircd_defs.h"
#include "s_conf.h"
#include "hostmask.h"
#include "numeric.h"
#include "send.h"
#include "irc_string.h"

#ifdef IPV6
static int try_parse_v6_netmask(const char *, struct irc_inaddr *, int *);
static unsigned long hash_ipv6(struct irc_inaddr *, int);
#endif
static int try_parse_v4_netmask(const char *, struct irc_inaddr *, int *);
static unsigned long hash_ipv4(struct irc_inaddr *, int);
#define DigitParse(ch) \
  do { \
    if (ch >= '0' && ch <= '9') \
      ch = ch - '0'; \
    else if (ch >= 'A' && ch <= 'F') \
      ch = ch - 'A' + 10; \
    else if (ch >= 'a' && ch <= 'f') \
      ch = ch - 'a' + 10; \
  } while(0);

/* The mask parser/type determination code */
/* try_parse_v6_netmask
 * input  - A possible IPV6 address as a string.
 * output - An integer describing whether it is an IPV6 or hostmask
 *
 * Called from parse_netmask
 */
/* Fixed ::/0 (any IPv6) & DigitParse -Gozem 2002-07-19 gozem@linux.nu */
#ifdef IPV6
static int try_parse_v6_netmask(const char *text, struct irc_inaddr * addr,
  int *b)
{
  const char *p;
  char c;
  int d[8] = {0, 0, 0, 0, 0, 0, 0, 0}, dp = 0, nyble = 4, finsert = -1,
    bits = 128, deficit = 0;
  short dc[8];
  for (p = text; (c = *p); p++)
    if (IsXDigit(c)) {
      if (nyble == 0)
        return HM_HOST;
      DigitParse(c);
      d[dp] |= c << (4 * --nyble);
    }
    else if (c == ':') {
      if (p > text && *(p - 1) == ':') {
        if (finsert >= 0)
          return HM_HOST;
        finsert = dp;
      }
      else {
        /* If there were less than 4 hex digits, e.g. :ABC: shift right so we
         * don't interpret it as ABC0 -A1kmm
         */
        d[dp] = d[dp] >> 4 * nyble;
        nyble = 4;
        if (++dp >= 8)
          return HM_HOST;
      }
    }
    else if (c == '*') {
      /* * must be last, and * is ambiguous if there is a :: -A1kmm */
      if (finsert >= 0 || *(p + 1) || dp == 0 || *(p - 1) != ':')
        return HM_HOST;
      bits = dp * 16;
    }
    else if (c == '/') {
      char *after;
      d[dp] = d[dp] >> 4 * nyble;
      dp++;
      bits = strtoul(p + 1, &after, 10);
      if (bits < 0 || *after)
        return HM_HOST;
      if (bits > dp * 4 && !(finsert >= 0 && bits <= 128))
        return HM_HOST;
      break;
    }
    else
      return HM_HOST;
  d[dp] = d[dp] >> 4 * nyble;
  if (c == 0)
    dp++;
  if (finsert < 0 && bits == 0)
    bits = dp * 16;
  /* How many words are missing? -A1kmm */
  deficit = bits / 16 + ((bits % 16) ? 1 : 0) - dp;
  /* Now fill in the gaps (from ::) in the copied table -A1kmm */
  for (dp = 0, nyble = 0; dp < 8; dp++) {
    if (nyble == finsert && deficit) {
      dc[dp] = 0;
      deficit--;
    }
    else
      dc[dp] = d[nyble++];
  }
  /* Set unused bits to 0 -A1kmm */
  if (bits < 128 && (bits % 16 != 0))
    dc[bits / 16] &= ~((1 << (15 - bits % 16)) - 1);
  for (dp = bits / 16 + (bits % 16 ? 1 : 0); dp < 8; dp++)
    dc[dp] = 0;
  /* And assign -A1kmm */
  if (addr)
    for (dp = 0; dp < 8; dp++)
      /* The cast is a kludge to make netbsd work */
      ((unsigned short *)&addr->sins.sin6)[dp] = htons(dc[dp]);
  if (b != NULL)
    *b = bits;
  return HM_IPV6;
}
#endif

/* try_parse_v4_netmask
 * input  - A possible IPV4 address as a string.
 * output - An integer describing whether it is an IPV4 or hostmask
 *
 * Called from parse_netmask
 */
static int try_parse_v4_netmask(const char *text, struct irc_inaddr * addr, int *b)
{
  const char *p;
  const char *digits[4];
  unsigned char addb[4];
  int n = 0, bits = 0;
  char c;
  digits[n++] = text;
  for (p = text; (c = *p); p++)
    if (c >= '0' && c <= '9') /* empty */
      ;
    else if (c == '.') {
      if (n >= 4)
        return HM_HOST;
      digits[n++] = p + 1;
    }
    else if (c == '*') {
      if (*(p + 1) || n == 0 || *(p - 1) != '.')
        return HM_HOST;
      bits = (n - 1) * 8;
      break;
    }
    else if (c == '/') {
      char *after;
      bits = strtoul(p + 1, &after, 10);
      if (!bits || *after)
        return HM_HOST;
      if (bits > n * 8)
        return HM_HOST;
      break;
    }
    else
      return HM_HOST;
  if (n < 4 && bits == 0)
    bits = n * 8;
  if (bits)
    while (n < 4)
      digits[n++] = "0";
  for (n = 0; n < 4; n++)
    addb[n] = strtoul(digits[n], NULL, 10);
  if (bits == 0)
    bits = 32;
  /* Set unused bits to 0... -A1kmm */
  if (bits < 32 && bits % 8)
    addb[bits / 8] &= ~((1 << (8 - bits % 8)) - 1);
  for (n = bits / 8 + (bits % 8 ? 1 : 0); n < 4; n++)
    addb[n] = 0;
  if (addr)
    addr->sins.sin.s_addr =
      htonl(addb[0] << 24 | addb[1] << 16 | addb[2] << 8 | addb[3]);
  if (b != NULL)
    *b = bits;
  return HM_IPV4;
}

/* parse_netmask
 * input  - A hostmask, or an IPV4/6 address.
 * output - An integer describing whether it is an IPV4, IPV6 address or a
 *          hostmask
 */
int parse_netmask(const char *text, struct irc_inaddr * addr, int *b)
{
#ifdef IPV6
  if (strchr(text, ':'))
    return try_parse_v6_netmask(text, addr, b);
#endif
  if (strchr(text, '.'))
    return try_parse_v4_netmask(text, addr, b);
  return HM_HOST;
}

/* Hashtable stuff...now external as its used in m_stats.c */
struct AddressRec *atable[ATABLE_SIZE];

void init_host_hash(void)
{
  memset(&atable, 0, sizeof(atable));
}

/* hash_ipv4 - determine hash value of input IPv4 address */
static unsigned long hash_ipv4(struct irc_inaddr * addr, int bits)
{
  if (bits != 0) {
    unsigned long av = ntohl(addr->sins.sin.s_addr) & ~((1 << (32 - bits)) - 1);
    return (av ^ (av >> 12) ^ (av >> 24)) & (ATABLE_SIZE - 1);
  }
  return NO;
}

#ifdef IPV6
/* hash_ipv6 - determine hash value of input IPv6 address */
static unsigned long hash_ipv6(struct irc_inaddr * addr, int bits)
{
  unsigned long v = 0, n;
  for (n = 0; n < 16; n++) {
    if (bits >= 8) {
      v ^= addr->sins.sin6.s6_addr[n];
      bits -= 8;
    }
    else if (bits) {
      v ^= addr->sins.sin6.s6_addr[n] & ~((1 << (8 - bits)) - 1);
      return v & (ATABLE_SIZE - 1);
    }
    else
      return v & (ATABLE_SIZE - 1);
  }
  return v & (ATABLE_SIZE - 1);
}
#endif
/* hash_text
 * input  - The start of the text to hash.
 * output - The hash of the string between 1 and (TH_MAX-1)
 */
static int hash_text(const char *start)
{
  const char *p = start;
  unsigned long h = 0;
  while (*p)
    h = (h << 4) - (h + (unsigned char)ToLower(*p++));
  return (h & (ATABLE_SIZE - 1));
}
/* get_hash_mask
 * input  - The text to hash.
 * output - The hash of the string right of the first '.' past the last
 *          wildcard in the string.
 */
static unsigned long get_mask_hash(const char *text)
{
  const char *hp = "", *p;
  for (p = text + strlen(text) - 1; p >= text; p--) {
    if (*p == '*' || *p == '?')
      return hash_text(hp);
    else if (*p == '.')
      hp = p + 1;
  }
  return hash_text(text);
}

/* find_conf_by_address
 * input  - The hostname, the address, the type of mask to find, the address
 *          family, the username.
 * output - The matching value with the highest precedence.
 *
 * Setting bit 0 of the type means that the username is ignored.
 */
struct ConfItem * find_conf_by_address(const char *name,
  struct irc_inaddr * addr, int type, int fam, const char *username)
{
  unsigned long hprecv = 0;
  struct ConfItem *hprec = NULL;
  struct AddressRec *arec;
  int b;
  if (username == NULL)
    username = "";
  if (addr) {
    /* Check for IPV6 matches... */
#ifdef IPV6
    if (fam == AF_INET6) {
      for (b = 128; b >= 0; b -= 16) {
        for (arec = atable[hash_ipv6(addr, b)]; arec; arec = arec->next) {
          if (arec->type == (type & ~0x1) && arec->masktype == HM_IPV6 &&
              comp_with_mask(&PIN_ADDR(addr), &IN_ADDR(arec->Mask.ipa.addr),
              arec->Mask.ipa.bits) && (type & 0x1 || match(arec->username,
              username)) && arec->precedence > hprecv) {
            hprecv = arec->precedence;
            hprec = arec->aconf;
          }
        }
      }
    }
    else
#endif
    if (fam == AF_INET) {
      for (b = 32; b >= 0; b -= 8) {
        for (arec = atable[hash_ipv4(addr, b)]; arec; arec = arec->next) {
          if (arec->type == (type & ~0x1) && arec->masktype == HM_IPV4 &&
              comp_with_mask(&PIN_ADDR(addr), &IN_ADDR(arec->Mask.ipa.addr),
              arec->Mask.ipa.bits) && (type & 0x1 || match(arec->username,
              username)) && arec->precedence > hprecv) {
            hprecv = arec->precedence;
            hprec = arec->aconf;
          }
        }
      }
    }
  }
  if (name != NULL) {
    const char *p;
    /* check p after strchr and p after increment for NULL -kre */
    for (p = name; p != NULL;) {
      for (arec = atable[hash_text(p)]; arec; arec = arec->next)
        if ((arec->type == (type & ~0x1)) && (arec->masktype == HM_HOST) &&
          match(arec->Mask.hostname, name) && (type & 0x1 ||
          match(arec->username, username)) && (arec->precedence > hprecv)) {
          hprecv = arec->precedence;
          hprec = arec->aconf;
        }
      p = strchr(p, '.');
      if (p != NULL)
        p++;
      else
        break;
    }
    for (arec = atable[0]; arec; arec = arec->next)
      if (arec->type == (type & ~0x1) && arec->masktype == HM_HOST &&
          match(arec->Mask.hostname, name) && (type & 0x1 ||
          match(arec->username, username)) && arec->precedence > hprecv) {
        hprecv = arec->precedence;
        hprec = arec->aconf;
      }
  }
  return hprec;
}

/* find_kline
 * input        - pointer to client
 * side effects - matching kline is returned, if found
 */
struct ConfItem * find_kline(struct Client * client_p)
{
  struct ConfItem *aconf;
  aconf = find_conf_by_address(client_p->host, &client_p->localClient->ip,
    CONF_KILL, client_p->localClient->aftype, client_p->username);
  return aconf;
}

/* find_gline
 * input        - pointer to client
 * side effects - matching dline is returned, if found
 */
struct ConfItem * find_gline(struct Client * client_p)
{
  struct ConfItem *aconf;
  aconf = find_conf_by_address(client_p->host, &client_p->localClient->ip,
    CONF_GLINE, client_p->localClient->aftype, client_p->username);
  return aconf;
}

/* find_address_conf
 * input  - The hostname, username, address, address family.
 * output - The applicable ConfItem.
 */
struct ConfItem * find_address_conf(const char *host, const char *user,
  struct irc_inaddr * ip, int aftype)
{
  struct ConfItem *iconf, *kconf;
  /* Find the best I-line. If none, return NULL -A1kmm */
  if (!(iconf = find_conf_by_address(host, ip, CONF_CLIENT, aftype, user)))
    return NULL;
  /* If they are exempt from K-lines, return the best I-line. -A1kmm */
  if (IsConfExemptKline(iconf))
    return iconf;
  /* Find the best K-line -A1kmm */
  kconf = find_conf_by_address(host, ip, CONF_KILL, aftype, user);
  /* If they are K-lined, return the K-line */
  if (kconf)
    return kconf;
  /* hunt for a gline */
  if (ConfigFileEntry.glines) {
    kconf = find_conf_by_address(host, ip, CONF_GLINE, aftype, user);
    if ((kconf != NULL) && !IsConfExemptGline(iconf))
      return kconf;
  }
  return iconf;
}

/* find_dline
 * input  - An address, an address family.
 * output - The best matching D-line or exempt line.
 */
struct ConfItem *
 find_dline(struct irc_inaddr * addr, int aftype)
{
  struct ConfItem *eline;
  eline = find_conf_by_address(NULL, addr, CONF_EXEMPTDLINE | 1, aftype, NULL);
  if (eline)
    return eline;
  return find_conf_by_address(NULL, addr, CONF_DLINE | 1, aftype, NULL);
}
/* add_conf_by_address - adds entry to hash table */
void add_conf_by_address(const char *address, int type, const char *username,
  struct ConfItem * aconf)
{
  static unsigned long prec_value = 0xFFFFFFFF;
  int masktype, bits;
  unsigned long hv;
  struct AddressRec *arec;
  if (address == NULL)
    address = "/NOMATCH!/";
  arec = MyMalloc(sizeof(struct AddressRec));
  masktype = parse_netmask(address, &arec->Mask.ipa.addr, &bits);
  arec->Mask.ipa.bits = bits;
  arec->masktype = masktype;
#ifdef IPV6
  if (masktype == HM_IPV6) {
    /* We have to do this, since we do not re-hash for every bit -A1kmm */
    bits -= bits % 16;
    arec->next = atable[(hv = hash_ipv6(&arec->Mask.ipa.addr, bits))];
    atable[hv] = arec;
  }
  else
#endif
  if (masktype == HM_IPV4) {
    /* We have to do this, since we do not re-hash for every bit -A1kmm */
    bits -= bits % 8;
    arec->next = atable[(hv = hash_ipv4(&arec->Mask.ipa.addr, bits))];
    atable[hv] = arec;
  }
  else {
    arec->Mask.hostname = address;
    arec->next = atable[(hv = get_mask_hash(address))];
    atable[hv] = arec;
  }
  arec->username = username;
  arec->aconf = aconf;
  arec->precedence = prec_value--;
  arec->type = type;
}
/* delete_one_address
 * input        - An address string, the associated ConfItem.
 * side effects - Deletes an address record. Frees the ConfItem if there
 *                is nothing referencing it, sets it as illegal otherwise.
 */
void delete_one_address_conf(const char *address, struct ConfItem * aconf)
{
  int masktype, bits;
  unsigned long hv;
  struct AddressRec *arec, *arecl = NULL;
  struct irc_inaddr addr;
  masktype = parse_netmask(address, &addr, &bits);
#ifdef IPV6
  if (masktype == HM_IPV6) {
    /* We have to do this, since we do not re-hash for every bit -A1kmm. */
    bits -= bits % 16;
    hv = hash_ipv6(&addr, bits);
  }
  else
#endif
  if (masktype == HM_IPV4) {
    /* We have to do this, since we do not re-hash for every bit -A1kmm. */
    bits -= bits % 8;
    hv = hash_ipv4(&addr, bits);
  }
  else
    hv = get_mask_hash(address);
  for (arec = atable[hv]; arec; arec = arec->next) {
    if (arec->aconf == aconf) {
      if (arecl)
        arecl->next = arec->next;
      else
        atable[hv] = arec->next;
      aconf->status |= CONF_ILLEGAL;
      if (!aconf->clients)
        free_conf(aconf);
      MyFree(arec);
      return;
    }
    arecl = arec;
  }
}
/* clear_out_address_conf - clear out address records in hash tables */
void clear_out_address_conf(void)
{
  int i;
  struct AddressRec **store_next;
  struct AddressRec *arec, *arecn;
  for (i = 0; i < ATABLE_SIZE; i++) {
    store_next = &atable[i];
    for (arec = atable[i]; arec; arec = arecn) {
      arecn = arec->next;
      /* We keep the temporary K-lines and destroy the permanent ones */
      if (arec->aconf->flags & CONF_FLAGS_TEMPORARY) {
        *store_next = arec;
        store_next = &arec->next;
      }
      else {
        arec->aconf->status |= CONF_ILLEGAL;
        if (!arec->aconf->clients)
          free_conf(arec->aconf);
        MyFree(arec);
      }
    }
    *store_next = NULL;
  }
}

/* show_iline_prefix
 * inputs       - pointer to struct Client requesting output
 *              - pointer to struct ConfItem
 *              - name to which iline prefix will be prefixed to
 * output       - pointer to static string with prefixes listed in ascii form
 */
char * show_iline_prefix(struct Client * sptr, struct ConfItem * aconf,
  char *name)
{
  static char prefix_of_host[USERLEN + 15];
  char *prefix_ptr;
  prefix_ptr = prefix_of_host;
  if (IsNoTilde(aconf))
    *prefix_ptr++ = '-';
  if (IsLimitIp(aconf))
    *prefix_ptr++ = '!';
  if (IsNeedIdentd(aconf))
    *prefix_ptr++ = '+';
  if (IsPassIdentd(aconf))
    *prefix_ptr++ = '$';
  if (IsNoMatchIp(aconf))
    *prefix_ptr++ = '%';
  if (IsConfDoSpoofIp(aconf))
    *prefix_ptr++ = '=';
  if (MyOper(sptr) && IsConfExemptKline(aconf))
    *prefix_ptr++ = '^';
  if (MyOper(sptr) && IsConfExemptLimits(aconf))
    *prefix_ptr++ = '>';
  if (MyOper(sptr) && IsConfIdlelined(aconf))
    *prefix_ptr++ = '<';
  *prefix_ptr = '\0';
  strncpy(prefix_ptr, name, USERLEN);
  return (prefix_of_host);
}

/* report_auth
 * input        - pointer to client to report to
 * side effects - Reports configured auth{} blocks to client_p
 */
void report_auth(struct Client * client_p)
{
  char *name, *host, *pass, *user, *classname;
  struct AddressRec *arec;
  struct ConfItem *aconf;
  int i, port;
  for (i = 0; i < ATABLE_SIZE; i++)
    for (arec = atable[i]; arec; arec = arec->next)
      if (arec->type == CONF_CLIENT) {
        aconf = arec->aconf;
        if (!MyOper(client_p) && IsConfDoSpoofIp(aconf))
          continue;
        get_printable_conf(aconf, &name, &host, &pass, &user, &port,
          &classname);
        sendto_one(client_p, form_str(RPL_STATSILINE), me.name, client_p->name,
          (IsConfRestricted(aconf)) ? 'i' : 'I', name, show_iline_prefix(
          client_p, aconf, user),
#ifdef HIDE_SPOOF_IPS
          IsConfDoSpoofIp(aconf) ? "255.255.255.255" :
#endif
          host, port, classname);
      }
}

/* report_Klines
 * inputs       - Client to report to, mask
 * side effects - Reports configured K-lines to client_p
 */
void report_Klines(struct Client * client_p)
{
  char *name, *host, *pass, *user, *classname;
  struct AddressRec *arec;
  struct ConfItem *aconf = NULL;
  int i, port;
  for (i = 0; i < ATABLE_SIZE; i++) {
    for (arec = atable[i]; arec; arec = arec->next) {
      if (arec->type == CONF_KILL) {
        aconf = arec->aconf;
        /* its a tempkline, theyre reported elsewhere */
        if (aconf->flags & CONF_FLAGS_TEMPORARY)
          continue;
        get_printable_conf(aconf, &name, &host, &pass, &user, &port,
          &classname);
        sendto_one(client_p, form_str(RPL_STATSKLINE), me.name, client_p->name,
          'K', host, user, pass);
      }
    }
  }
}
