/* match.c: string matching functions.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"
#include "irc_string.h"
#include "common.h"
#include "config.h"
#include "client.h"
#include "ircd.h"

/* match - compare given string against mask (wild cards * and ?)
 *  return 1, if match
 *         0, if no match
 *
 *  -Douglas A Lewis (dalewis@acsu.buffalo.edu)
 */
/* Dies when it's less than this, we've got long lines to parse */
#define MATCH_MAX_CALLS 512

int match(const char *mask, const char *name)
{
  const unsigned char *m = (const unsigned char *)mask;
  const unsigned char *n = (const unsigned char *)name;
  const unsigned char *ma = (const unsigned char *)mask;
  const unsigned char *na = (const unsigned char *)name;
  int wild = 0;
  int calls = 0;
  s_assert(mask != NULL);
  s_assert(name != NULL);
  if (!mask || !name)
    return NO;

	/* if the mask is "*", it matches everything */
  if((*m == '*') && (*(m+1) == '\0'))
    return YES;

  while (calls++ < MATCH_MAX_CALLS) {
    if (*m == '*') {
      while (*m == '*')
        m++;
      wild = 1;
      ma = m;
      na = n;
    }
    if (!*m) {
      if (!*n)
        return YES;
      for (m--; (m > (const unsigned char *)mask) && (*m == '?'); m--);
   
      if (*m == '*' && (m > (const unsigned char *)mask))
        return YES;
      if (!wild)
        return NO;
      m = ma;
      n = ++na;
    }
    else if (!*n) {
      while (*m == '*')
        m++;
      return (*m == 0);
    }
    if (ToLower(*m) != ToLower(*n) && *m != '?') {
      if (!wild)
        return NO;
      m = ma;
      n = ++na;
    }
    else {
      if (*m)
        m++;
      if (*n)
        n++;
    }
  }
  return NO;
}

/* match_esc - The match() function with support for escaping characters such
 *  as '*', '?', '#' and '@'
 */
int match_esc(const char *mask, const char *name)
{
  const unsigned char *m = (const unsigned char *)mask;
  const unsigned char *n = (const unsigned char *)name;
  const unsigned char *ma = (const unsigned char *)mask;
  const unsigned char *na = (const unsigned char *)name;
  int wild = 0;
  int calls = 0;
  int quote = 0;
  s_assert(mask != NULL);
  s_assert(name != NULL);
  if (!mask || !name)
    return NO;
  while (calls++ < MATCH_MAX_CALLS) {
    if (quote)
      quote++;
    if (quote == 3)
      quote = 0;
    if (*m == '\\' && !quote) {
      m++;
      quote = 1;
      continue;
    }
    if (!quote && *m == '*') {
      while (*m == '*')
        m++;
      if (*m == '\\') {
        m++;
        /* This means it is an invalid mask -A1kmm */
        if (!*m)
          return NO;
        quote = 2;
      }
      wild = 1;
      ma = m;
      na = n;
    }
    if (!*m) {
      if (!*n)
        return YES;
      if (quote)
        return NO;
      for (m--; (m > (const unsigned char *)mask) &&
        (*m == '?' || *m == '#' || *m == '@'); m--);
      if (*m == '*' && (m > (const unsigned char *)mask))
        return YES;
      if (!wild)
        return NO;
      m = ma;
      n = ++na;
    }
    else if (!*n) {
      if (quote)
        return NO;
      while (*m == '*')
        m++;
      return (*m == 0);
    }
    if (ToLower(*m) != ToLower(*n) && !(!quote && *m == '?') &&
      !(!quote && *m == '@' && IsLetter(*n)) &&
      !(!quote && *m == '#' && IsDigit(*n))) {
      if (!wild)
        return NO;
      m = ma;
      n = ++na;
    }
    else {
      if (*m)
        m++;
      if (*n)
        n++;
    }
  }
  return NO;
}

int comp_with_mask(void *addr, void *dest, u_int mask)
{
  if (memcmp(addr, dest, mask / 8) == 0) {
    int n = mask / 8;
    int m = ((-1) << (8 - (mask % 8)));
    if (mask % 8 == 0 || (((u_char *) addr)[n] & m) == (((u_char *) dest)[n] & m))
      return YES;
  }
  return NO;
}

/* match_cidr
 * input  - mask, address
 * output - success, 1
 *        - failure, 0
 */
int match_cidr(const char *s1, const char *s2)
{
  struct irc_inaddr ipaddr, maskaddr;
  char mask[BUFSIZE];
  char address[NICKLEN + USERLEN + HOSTLEN + 6];
  char *ipmask;
  char *ip;
  char *len;
  int cidrlen, aftype;
  strcpy(mask, s1);
  strcpy(address, s2);
  ipmask = strrchr(mask, '@');
  if (ipmask == NULL)
    return NO;
  *ipmask++ = '\0';
  ip = strrchr(address, '@');
  if (ip == NULL)
    return NO;
  *ip++ = '\0';

  len = strrchr(ipmask, '/');
  if (len == NULL)
    return NO;
  *len++ = '\0';
  cidrlen = atoi(len);
  if (cidrlen == 0)
    return NO;
#ifdef IPV6
  if (strchr(ip, ':') && strchr(ipmask, ':'))
    aftype = AF_INET6;
  else
#endif
  if (!strchr(ip, ':') && !strchr(ipmask, ':'))
    aftype = AF_INET;
  else
    return NO;
  inetpton(aftype, ip, &ipaddr);
  inetpton(aftype, ipmask, &maskaddr);
  if (comp_with_mask(&IN_ADDR(ipaddr), &IN_ADDR(maskaddr), cidrlen) &&
      match(mask, address))
    return YES;
  else
    return NO;
}

/* collapse - collapses a string containing multiple *'s */
char * collapse(char *pattern)
{
  char *p = pattern, *po = pattern;
  char c;
  int f = 0;
  if (p == NULL)
    return NULL;
  while ((c = *p++)) {
    if (c == '*') {
      if (!(f & 1))
        *po++ = '*';
      f |= 1;
    }
    else {
      *po++ = c;
      f &= ~1;
    }
  }
  *po++ = 0;
  return pattern;
}

/* collapse_esc - collapse() with support for escaping characters */
char * collapse_esc(char *pattern)
{
  char *p = pattern, *po = pattern;
  char c;
  int f = 0;
  if (p == NULL)
    return NULL;
  while ((c = *p++)) {
    if (!(f & 2) && c == '*') {
      if (!(f & 1))
        *po++ = '*';
      f |= 1;
    }
    else if (!(f & 2) && c == '\\') {
      *po++ = '\\';
      f |= 2;
    }
    else {
      *po++ = c;
      f &= ~3;
    }
  }
  *po++ = 0;
  return pattern;
}

/* irccmp - case insensitive comparison of two 0 terminated strings.
 *
 *  <0, if s1 lexicographically less than s2
 *   0, if s1 equal to s2
 *  >0, if s1 lexicographically greater than s2
 */
int irccmp(const char *s1, const char *s2)
{
  const unsigned char *str1 = (const unsigned char *)s1;
  const unsigned char *str2 = (const unsigned char *)s2;
  int res;
  s_assert(s1 != NULL);
  s_assert(s2 != NULL);
  while ((res = ToUpper(*str1) - ToUpper(*str2)) == 0) {
    if (*str1 == '\0')
      return NO;
    str1++;
    str2++;
  }
  return res;
}

int ircncmp(const char *s1, const char *s2, int n)
{
  const unsigned char *str1 = (const unsigned char *)s1;
  const unsigned char *str2 = (const unsigned char *)s2;
  int res;
  s_assert(s1 != NULL);
  s_assert(s2 != NULL);
  while ((res = ToUpper(*str1) - ToUpper(*str2)) == 0) {
    str1++;
    str2++;
    n--;
    if (n == 0 || (*str1 == '\0' && *str2 == '\0'))
      return NO;
  }
  return res;
}

const unsigned char ToLowerTab[] = {
  0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
  0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c,
  0x1d, 0x1e, 0x1f, ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')', '*', '+',
  ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':',
  ';', '<', '=', '>', '?', '@', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
  'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
  'y', 'z', '{', '|', '}', '~', '_', '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g',
  'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
  'w', 'x', 'y', 'z', '{', '|', '}', '~', 0x7f, 0x80, 0x81, 0x82, 0x83, 0x84,
  0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91,
  0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e,
  0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab,
  0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8,
  0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5,
  0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2,
  0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
  0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec,
  0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
  0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
};

const unsigned char ToUpperTab[] = {
  0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
  0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c,
  0x1d, 0x1e, 0x1f, ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')', '*', '+',
  ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':',
  ';', '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
  'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
  'Y', 'Z', '[', '\\', ']', '^', 0x5f, '`', 'A', 'B', 'C', 'D', 'E', 'F', 'G',
  'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
  'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', 0x7f, 0x80, 0x81, 0x82, 0x83, 0x84,
  0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91,
  0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e,
  0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xab,
  0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8,
  0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5,
  0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2,
  0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
  0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec,
  0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
  0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
};

/* CharAttrs table - allow anything but ^G, comma or space in channels */
const unsigned int CharAttrs[] = {
  /* 0 */
  CNTRL_C,
  /* 1 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 2 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 3 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 4 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 5 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 6 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* BEL */
  CNTRL_C | NONEOS_C,
  /* 8 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 9 */
  CNTRL_C | SPACE_C | CHAN_C | NONEOS_C,
  /* 10 */
  CNTRL_C | SPACE_C | CHAN_C | NONEOS_C | EOL_C,
  /* 11 */
  CNTRL_C | SPACE_C | CHAN_C | NONEOS_C,
  /* 12 */
  CNTRL_C | SPACE_C | CHAN_C | NONEOS_C,
  /* 13 */
  CNTRL_C | SPACE_C | CHAN_C | NONEOS_C | EOL_C,
  /* 14 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 15 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 16 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 17 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 18 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 19 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 20 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 21 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 22 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 23 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 24 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 25 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 26 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 27 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 28 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 29 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 30 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* 31 */
  CNTRL_C | CHAN_C | NONEOS_C,
  /* SP */
  PRINT_C | SPACE_C,
  /* ! */
  PRINT_C | KWILD_C | CHAN_C | NONEOS_C,
  /* " */
  PRINT_C | CHAN_C | NONEOS_C,
  /* # */
  PRINT_C | CHANPFX_C | CHAN_C | NONEOS_C,
  /* $ */
  PRINT_C | CHAN_C | NONEOS_C | USER_C,
  /* % */
  PRINT_C | CHAN_C | NONEOS_C,
  /* & */
  PRINT_C | CHANPFX_C | CHAN_C | NONEOS_C,
  /* ' */
  PRINT_C | CHAN_C | NONEOS_C,
  /* ( */
  PRINT_C | CHAN_C | NONEOS_C,
  /* ) */
  PRINT_C | CHAN_C | NONEOS_C,
  /* * */
  PRINT_C | KWILD_C | MWILD_C | CHAN_C | NONEOS_C | SERV_C,
  /* + */
  PRINT_C | CHAN_C | NONEOS_C,
  /* , */
  PRINT_C | NONEOS_C,
  /* - */
  PRINT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* . */
  PRINT_C | KWILD_C | CHAN_C | NONEOS_C | USER_C | HOST_C | SERV_C,
  /* / */
  PRINT_C | CHAN_C | NONEOS_C,
  /* 0 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 1 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 2 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 3 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 4 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 5 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 6 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 7 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 8 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* 9 */
  PRINT_C | DIGIT_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* : */
  PRINT_C | CHAN_C | NONEOS_C | HOST_C,
  /* ; */
  PRINT_C | CHAN_C | NONEOS_C,
  /* < */
  PRINT_C | CHAN_C | NONEOS_C,
  /* = */
  PRINT_C | CHAN_C | NONEOS_C,
  /* > */
  PRINT_C | CHAN_C | NONEOS_C,
  /* ? */
  PRINT_C | KWILD_C | MWILD_C | CHAN_C | NONEOS_C,
  /* @ */
  PRINT_C | KWILD_C | CHAN_C | NONEOS_C,
  /* A */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* B */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* C */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* D */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* E */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* F */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* G */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* H */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* I */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* J */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* K */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* L */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* M */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* N */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* O */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* P */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* Q */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* R */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* S */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* T */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* U */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* V */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* W */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* X */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* Y */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* Z */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* [ */
  PRINT_C | ALPHA_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* \ */
  PRINT_C | ALPHA_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* ] */
  PRINT_C | ALPHA_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* ^ */
  PRINT_C | ALPHA_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* _ */
  PRINT_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* ` */
  PRINT_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* a */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* b */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* c */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* d */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* e */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* f */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* g */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* h */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* i */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* j */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* k */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* l */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* m */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* n */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* o */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* p */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* q */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* r */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* s */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* t */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* u */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* v */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* w */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* x */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* y */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* z */
  PRINT_C | ALPHA_C | LETTER_C | NICK_C | CHAN_C | NONEOS_C | USER_C | HOST_C,
  /* { */
  PRINT_C | ALPHA_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* | */
  PRINT_C | ALPHA_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* } */
  PRINT_C | ALPHA_C | NICK_C | CHAN_C | NONEOS_C | USER_C,
  /* ~ */
  PRINT_C | ALPHA_C | CHAN_C | NONEOS_C | USER_C,
  /* del */
  CHAN_C | NONEOS_C,
  CHAN_C | NONEOS_C, /* 0x80 */
  CHAN_C | NONEOS_C, /* 81 */
  CHAN_C | NONEOS_C, /* 82 */
  CHAN_C | NONEOS_C, /* 83 */
  CHAN_C | NONEOS_C, /* 84 */
  CHAN_C | NONEOS_C, /* 85 */
  CHAN_C | NONEOS_C, /* 86 */
  CHAN_C | NONEOS_C, /* 87 */
  CHAN_C | NONEOS_C, /* 88 */
  CHAN_C | NONEOS_C, /* 89 */
  CHAN_C | NONEOS_C, /* 8A */
  CHAN_C | NONEOS_C, /* 8B */
  CHAN_C | NONEOS_C, /* 8C */
  CHAN_C | NONEOS_C, /* 8D */
  CHAN_C | NONEOS_C, /* 8E */
  CHAN_C | NONEOS_C, /* 8F */
  CHAN_C | NONEOS_C, /* 90 */
  CHAN_C | NONEOS_C, /* 91 */
  CHAN_C | NONEOS_C, /* 92 */
  CHAN_C | NONEOS_C, /* 93 */
  CHAN_C | NONEOS_C, /* 94 */
  CHAN_C | NONEOS_C, /* 95 */
  CHAN_C | NONEOS_C, /* 96 */
  CHAN_C | NONEOS_C, /* 97 */
  CHAN_C | NONEOS_C, /* 98 */
  CHAN_C | NONEOS_C, /* 99 */
  CHAN_C | NONEOS_C, /* 9A */
  CHAN_C | NONEOS_C, /* 9B */
  CHAN_C | NONEOS_C, /* 9C */
  CHAN_C | NONEOS_C, /* 9D */
  CHAN_C | NONEOS_C, /* 9E */
  CHAN_C | NONEOS_C, /* 9F */
  CHAN_C | NONEOS_C, /* A0 */
  CHAN_C | NONEOS_C, /* A1 */
  CHAN_C | NONEOS_C, /* A2 */
  CHAN_C | NONEOS_C, /* A3 */
  CHAN_C | NONEOS_C, /* A4 */
  CHAN_C | NONEOS_C, /* A5 */
  CHAN_C | NONEOS_C, /* A6 */
  CHAN_C | NONEOS_C, /* A7 */
  CHAN_C | NONEOS_C, /* A8 */
  CHAN_C | NONEOS_C, /* A9 */
  CHAN_C | NONEOS_C, /* AA */
  CHAN_C | NONEOS_C, /* AB */
  CHAN_C | NONEOS_C, /* AC */
  CHAN_C | NONEOS_C, /* AD */
  CHAN_C | NONEOS_C, /* AE */
  CHAN_C | NONEOS_C, /* AF */
  CHAN_C | NONEOS_C, /* B0 */
  CHAN_C | NONEOS_C, /* B1 */
  CHAN_C | NONEOS_C, /* B2 */
  CHAN_C | NONEOS_C, /* B3 */
  CHAN_C | NONEOS_C, /* B4 */
  CHAN_C | NONEOS_C, /* B5 */
  CHAN_C | NONEOS_C, /* B6 */
  CHAN_C | NONEOS_C, /* B7 */
  CHAN_C | NONEOS_C, /* B8 */
  CHAN_C | NONEOS_C, /* B9 */
  CHAN_C | NONEOS_C, /* BA */
  CHAN_C | NONEOS_C, /* BB */
  CHAN_C | NONEOS_C, /* BC */
  CHAN_C | NONEOS_C, /* BD */
  CHAN_C | NONEOS_C, /* BE */
  CHAN_C | NONEOS_C, /* BF */
  CHAN_C | NONEOS_C, /* C0 */
  CHAN_C | NONEOS_C, /* C1 */
  CHAN_C | NONEOS_C, /* C2 */
  CHAN_C | NONEOS_C, /* C3 */
  CHAN_C | NONEOS_C, /* C4 */
  CHAN_C | NONEOS_C, /* C5 */
  CHAN_C | NONEOS_C, /* C6 */
  CHAN_C | NONEOS_C, /* C7 */
  CHAN_C | NONEOS_C, /* C8 */
  CHAN_C | NONEOS_C, /* C9 */
  CHAN_C | NONEOS_C, /* CA */
  CHAN_C | NONEOS_C, /* CB */
  CHAN_C | NONEOS_C, /* CC */
  CHAN_C | NONEOS_C, /* CD */
  CHAN_C | NONEOS_C, /* CE */
  CHAN_C | NONEOS_C, /* CF */
  CHAN_C | NONEOS_C, /* D0 */
  CHAN_C | NONEOS_C, /* D1 */
  CHAN_C | NONEOS_C, /* D2 */
  CHAN_C | NONEOS_C, /* D3 */
  CHAN_C | NONEOS_C, /* D4 */
  CHAN_C | NONEOS_C, /* D5 */
  CHAN_C | NONEOS_C, /* D6 */
  CHAN_C | NONEOS_C, /* D7 */
  CHAN_C | NONEOS_C, /* D8 */
  CHAN_C | NONEOS_C, /* D9 */
  CHAN_C | NONEOS_C, /* DA */
  CHAN_C | NONEOS_C, /* DB */
  CHAN_C | NONEOS_C, /* DC */
  CHAN_C | NONEOS_C, /* DD */
  CHAN_C | NONEOS_C, /* DE */
  CHAN_C | NONEOS_C, /* DF */
  CHAN_C | NONEOS_C, /* E0 */
  CHAN_C | NONEOS_C, /* E1 */
  CHAN_C | NONEOS_C, /* E2 */
  CHAN_C | NONEOS_C, /* E3 */
  CHAN_C | NONEOS_C, /* E4 */
  CHAN_C | NONEOS_C, /* E5 */
  CHAN_C | NONEOS_C, /* E6 */
  CHAN_C | NONEOS_C, /* E7 */
  CHAN_C | NONEOS_C, /* E8 */
  CHAN_C | NONEOS_C, /* E9 */
  CHAN_C | NONEOS_C, /* EA */
  CHAN_C | NONEOS_C, /* EB */
  CHAN_C | NONEOS_C, /* EC */
  CHAN_C | NONEOS_C, /* ED */
  CHAN_C | NONEOS_C, /* EE */
  CHAN_C | NONEOS_C, /* EF */
  CHAN_C | NONEOS_C, /* F0 */
  CHAN_C | NONEOS_C, /* F1 */
  CHAN_C | NONEOS_C, /* F2 */
  CHAN_C | NONEOS_C, /* F3 */
  CHAN_C | NONEOS_C, /* F4 */
  CHAN_C | NONEOS_C, /* F5 */
  CHAN_C | NONEOS_C, /* F6 */
  CHAN_C | NONEOS_C, /* F7 */
  CHAN_C | NONEOS_C, /* F8 */
  CHAN_C | NONEOS_C, /* F9 */
  CHAN_C | NONEOS_C, /* FA */
  CHAN_C | NONEOS_C, /* FB */
  CHAN_C | NONEOS_C, /* FC */
  CHAN_C | NONEOS_C, /* FD */
  CHAN_C | NONEOS_C, /* FE */
  CHAN_C | NONEOS_C  /* FF */
};