/* rsa.c: Functions for use with RSA public key cryptography.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 */

#include "stdinc.h"

#ifdef HAVE_LIBCRYPTO
#include <openssl/pem.h>
#include <openssl/rand.h>
#include <openssl/rsa.h>
#include <openssl/md5.h>
#include <openssl/bn.h>
#include <openssl/evp.h>
#include <openssl/err.h>

#include "memory.h"
#include "rsa.h"
#include "s_conf.h"
#include "s_log.h"
#include "client.h"

void report_crypto_errors(void);
int verify_private_key(void);
static void binary_to_hex(unsigned char *bin, char *hex, int length);
int generate_challenge(char **, char **, RSA *);
int get_randomness(unsigned char *buf, int length);

/* report_crypto_errors - Dump crypto error list to log */
void report_crypto_errors(void)
{
  unsigned long e = 0;
  unsigned long cnt = 0;
  ERR_load_crypto_strings();
  while ((cnt < 100) && (e = ERR_get_error())) {
    ilog(L_CRIT, "SSL error: %s", ERR_error_string(e, 0));
    cnt++;
  }
}

/* verify_private_key - reread private key and verify against inmem key */
int verify_private_key(void)
{
  BIO *file;
  RSA *key;
  RSA *mkey;
  if (ServerInfo.rsa_private_key == NULL) {
    ilog(L_NOTICE, "Directive rsa_private_key is missing!");
    return 0;
  }
  /* If rsa_private_key_file isn't available, error out. */
  if (ServerInfo.rsa_private_key_file == NULL) {
    ilog(L_NOTICE, "Directive rsa_private_key is missing!");
    return 0;
  }
  file = BIO_new_file(ServerInfo.rsa_private_key_file, "r");
  /* If BIO_new_file returned NULL (according to OpenSSL docs), then an error
   * occurred.
   */
  if (file == NULL) {
    ilog(L_NOTICE, "Directive rsa_private_key points to a file that cannot "
      "be opened");
    return 0;
  }
  /*
   * jdc -- Let's do this a little differently.  According to the OpenSSL
   * documentation, you need to METHOD_free(key) before assigning it.  Don't
   * believe me?  Check out the following URL:
   * http://www.openssl.org/docs/crypto/pem.html#BUGS P.S. -- I have no idea
   * why the key= assignment has to be typecasted. For some reason the system
   * thinks PEM_read_bio_RSAPrivateKey is returning an int, not a RSA *.
   * androsyn -- Thats because you didn't have a prototype and including pem.h
   * breaks things for some reason..
   */
  key = (RSA *) PEM_read_bio_RSAPrivateKey(file, NULL, 0, NULL);
  if (key == NULL) {
    ilog(L_NOTICE, "PEM_read_bio_RSAPrivateKey() failed; possibly not RSA?");
    report_crypto_errors();
    return 0;
  }
  BIO_set_close(file, BIO_CLOSE);
  BIO_free(file);
  mkey = ServerInfo.rsa_private_key;
  /* Compare the in-memory key to the key we just loaded above.  If any of the
   * portions don't match, then logically we have a different in-memory key vs.
   * the one we just loaded.  This is bad, mmmkay?
   *
   * Instead of printing out specific errors, we could just indicate that the
   * private key is corrupt...
   */
  if (mkey->pad != key->pad || mkey->version != key->version || BN_cmp(mkey->n,
      key->n) || BN_cmp(mkey->e, key->e) || BN_cmp(mkey->d, key->d) ||
      BN_cmp(mkey->p, key->p) || BN_cmp(mkey->q, key->q) || BN_cmp(mkey->dmp1,
      key->dmp1) || BN_cmp(mkey->dmq1, key->dmq1) || BN_cmp(mkey->iqmp,
      key->iqmp)) {
    ilog(L_CRIT, "RSA private key is corrupt!");
    return 0;
  }
  RSA_free(key);
  return 1;
}

static void binary_to_hex(unsigned char *bin, char *hex, int length)
{
  const char *trans = "0123456789ABCDEF";
  int i;
  for (i = 0; i < length; i++) {
    hex[i << 1] = trans[bin[i] >> 4];
    hex[(i << 1) + 1] = trans[bin[i] & 0xf];
  }
  hex[i << 1] = '\0';
}

int get_randomness(unsigned char *buf, int length)
{
  /* Seed OpenSSL PRNG with EGD enthropy pool -kre */
  if (ConfigFileEntry.use_egd && (ConfigFileEntry.egdpool_path != NULL)) {
    if (RAND_egd(ConfigFileEntry.egdpool_path) == -1)
      return -1;
  }
  if (RAND_status())
    return RAND_bytes(buf, length);
  else /* XXX - abort? */
    return RAND_pseudo_bytes(buf, length);
}

int generate_challenge(char **r_challenge, char **r_response, RSA * rsa)
{
  unsigned char secret[32], *tmp;
  unsigned long length, ret;
  if (!rsa)
    return -1;
  get_randomness(secret, 32);
  *r_response = MyMalloc(65);
  binary_to_hex(secret, *r_response, 32);
  length = RSA_size(rsa);
  tmp = MyMalloc(length);
  ret = RSA_public_encrypt(32, secret, tmp, rsa, RSA_PKCS1_PADDING);
  *r_challenge = MyMalloc((length << 1) + 1);
  binary_to_hex(tmp, *r_challenge, length);
  (*r_challenge)[length << 1] = 0;
  MyFree(tmp);
  if (ret < 0) {
    report_crypto_errors();
    return -1;
  }
  return 0;
}
#endif /* HAVE_LIBCRYPTO */
