/* s_auth.c: Functions for querying a users ident.
 *
 * Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 * Copyright (C) 2003-2004 ircd-immortal development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 *
 * Rewrote most of the code here. When a client connects to the server and
 * passes initial socket validation checks, it is owned by this module (auth)
 * which returns it to the rest of the server when dns and auth queries are
 * finished. Until the client is released, the server does not know it exists
 * and does not process any messages from it.
 * -Bleep (Thomas Helvey) <tomh@inxpress.net>, July 6, 1999
 */

#include "stdinc.h"
#include "tools.h"
#include "s_auth.h"
#include "s_conf.h"
#include "client.h"
#include "common.h"
#include "event.h"
#include "fdlist.h"
#include "irc_string.h"
#include "sprintf_irc.h"
#include "ircd.h"
#include "numeric.h"
#include "packet.h"
#include "res.h"
#include "s_bsd.h"
#include "s_log.h"
#include "s_stats.h"
#include "send.h"
#include "memory.h"
#include "config.h"

#define REPORT_LOGON        "Please wait a moment while we authenticate your " \
                            "connection...\r\n"
#define REPORT_FAIL_DNS     "*** Notice -- Your hostname could not be "        \
                            "resolved\r\n"
#define REPORT_FAIL_ID      "*** Notice -- No identd response, username "      \
                            "prefixed with a tilde (~)\r\n"
#define REPORT_IP_MISMATCH  "*** Notice -- Your hostname does not match IP "   \
                            "address\r\n"
#define REPORT_HOST_INVALID "*** Notice -- Your hostname is invalid\r\n"

static dlink_list auth_poll_list;
static BlockHeap *auth_heap;
static EVH timeout_auth_queries_event;
static PF read_auth_reply;
static CNCB auth_connect_callback;

/* init_auth - Initialise the auth code */
void init_auth(void)
{
  eventAddIsh("timeout_auth_queries_event", timeout_auth_queries_event, NULL,
    1);
  auth_heap = BlockHeapCreate(sizeof(struct AuthRequest), LCLIENT_HEAP_SIZE);
}

/* sendheader - Send report with server mask as source */
void sendheader(const int descriptor, const char *report)
{
  char text[BUFSIZE];
  ircsprintf(text, ":%s NOTICE * :%s", me.host, report);
  send(descriptor, text, strlen(text), 0);
  return;
}

/* make_auth_request - allocate a new auth request */
static struct AuthRequest * make_auth_request(struct Client * client)
{
  struct AuthRequest *request = BlockHeapAlloc(auth_heap);
  memset(request, 0, sizeof(struct AuthRequest));
  client->localClient->auth_request = request;
  request->fd = -1;
  request->client = client;
  request->timeout = CurrentTime + ConfigFileEntry.connect_timeout;
  return request;
}

/* free_auth_request - cleanup auth request allocations */
static void free_auth_request(struct AuthRequest * request)
{
  BlockHeapFree(auth_heap, request);
}

/* release_auth_client - release auth client from auth system
 * this adds the client into the local client lists so it can be read by
 * the main io processing loop
 */
static void release_auth_client(struct AuthRequest * auth)
{
  struct Client *client = auth->client;
  if (IsDNSPending(auth) || IsDoingAuth(auth))
    return;
  client->localClient->auth_request = NULL;
  dlinkDelete(&auth->node, &auth_poll_list);
  free_auth_request(auth);
  if (client->localClient->fd > highest_fd)
    highest_fd = client->localClient->fd;
  /* When a client has auth'ed, we want to start reading what it sends us. This
   * is what read_packet() does. -- adrian
   */
  client->localClient->allow_read = MAX_FLOOD;
  comm_setflush(client->localClient->fd, 1000, flood_recalc, client);
  add_client_to_list(client);
  read_packet(client->localClient->fd, client);
}

/* auth_dns_callback - called when resolver query finishes
 * if the query resulted in a successful search, hp will contain
 * a non-null pointer, otherwise hp will be null.
 * set the client on it's way to a connection completion, regardless
 * of success of failure
 */
static void auth_dns_callback(void *vptr, adns_answer * reply)
{
  struct AuthRequest *auth = (struct AuthRequest *) vptr;
  ClearDNSPending(auth);
  if (reply && (reply->status == adns_s_ok)) {
    if (strlen(*reply->rrs.str) <= HOSTLEN)
      strlcpy(auth->client->host, *reply->rrs.str, sizeof(auth->client->host));
    else
      sendheader(auth->client->localClient->fd, REPORT_HOST_INVALID);
  }
  else {
#ifdef IPV6
    if (auth->client->localClient->aftype == AF_INET6
      && ConfigFileEntry.fallback_to_ip6_int == 1 && auth->ip6_int == 0) {
      struct Client *client = auth->client;
      auth->ip6_int = 1;
      MyFree(reply);
      if (adns_getaddr(&client->localClient->ip, client->localClient->aftype,
          &auth->dns_query, 1))
        ClearDNSPending(auth);
      else {
        SetDNSPending(auth);
        return;
      }
    }
#endif
    sendheader(auth->client->localClient->fd, REPORT_FAIL_DNS);
  }
  MyFree(reply);
  if (!IsDoingAuth(auth) && !IsDNSPending(auth))
    release_auth_client(auth);
}

/* authsenderr - handle auth send errors */
static void auth_error(struct AuthRequest * auth)
{
  ++ServerStats->is_abad;
  fd_close(auth->fd);
  auth->fd = -1;
  ClearAuth(auth);
  sendheader(auth->client->localClient->fd, REPORT_FAIL_ID);
  if (!IsDNSPending(auth))
    release_auth_client(auth);
}

/* start_auth_query - Flag the client to show that an attempt to contact the
 * ident server on the client's host. The connect and subsequently the socket
 * are all put into 'non-blocking' mode. Should the connect or any later phase
 * of the identifing process fail, it is aborted and the user is given a
 * username of "unknown".
 */
static int start_auth_query(struct AuthRequest * auth)
{
  struct irc_sockaddr localaddr;
  socklen_t locallen = sizeof(struct irc_sockaddr);
  int fd;
  if ((fd = comm_open(DEF_FAM, SOCK_STREAM, 0, "ident")) == -1) {
    report_error(L_ALL, "creating auth stream socket %s:%s",
      get_client_name(auth->client, SHOW_IP, YES), errno);
    ilog(L_ERROR, "Unable to create auth socket for %s:%m",
      log_client_name(auth->client));
    ++ServerStats->is_abad;
    return NO;
  }
  if ((MAXCONNECTIONS - 10) < fd) {
    sendto_realops_flags(UMODE_ALL, L_ALL, "Can't allocate fd for auth on %s",
      get_client_name(auth->client, SHOW_IP, YES));
    fd_close(fd);
    return NO;
  }
  if (!set_non_blocking(fd)) {
    report_error(L_ALL, NONB_ERROR_MSG, get_client_name(auth->client, SHOW_IP,
      YES), errno);
    fd_close(fd);
    return NO;
  }
  /*
   * get the local address of the client and bind to that to make the auth
   * request.  This used to be done only for ifdef VIRTUAL_HOST, but needs to
   * be done for all clients since the ident request must originate from that
   * same address-- and machines with multiple IP addresses are common now
   */
  memset(&localaddr, 0, locallen);
  getsockname(auth->client->localClient->fd, (struct sockaddr *) &
    SOCKADDR(localaddr), &locallen);
  S_PORT(localaddr) = htons(0);
  auth->fd = fd;
  SetAuthConnect(auth);
  comm_connect_tcp(fd, auth->client->localClient->sockhost, 113,
    (struct sockaddr *) & SOCKADDR(localaddr), locallen, auth_connect_callback,
    auth, DEF_FAM, GlobalSetOptions.ident_timeout);
  return 1; /* We suceed here for now */
}

/* GetValidIdent - parse ident query reply from identd server
 * input  - pointer to ident buf
 * output - NULL if no valid ident found, otherwise pointer to name
 */
static char * GetValidIdent(char *buf)
{
  int remp = 0;
  int locp = 0;
  char *colon1Ptr;
  char *colon2Ptr;
  char *colon3Ptr;
  char *commaPtr;
  char *remotePortString;
  /* All this to get rid of a sscanf() fun. */
  remotePortString = buf;
  colon1Ptr = strchr(remotePortString, ':');
  if (!colon1Ptr)
    return NO;
  *colon1Ptr = '\0';
  colon1Ptr++;
  colon2Ptr = strchr(colon1Ptr, ':');
  if (!colon2Ptr)
    return NO;
  *colon2Ptr = '\0';
  colon2Ptr++;
  commaPtr = strchr(remotePortString, ',');
  if (!commaPtr)
    return NO;
  *commaPtr = '\0';
  commaPtr++;
  remp = atoi(remotePortString);
  if (!remp)
    return NO;
  locp = atoi(commaPtr);
  if (!locp)
    return NO;
  /* look for USERID bordered by first pair of colons */
  if (!strstr(colon1Ptr, "USERID"))
    return NO;
  colon3Ptr = strchr(colon2Ptr, ':');
  if (!colon3Ptr)
    return NO;
  *colon3Ptr = '\0';
  colon3Ptr++;
  return colon3Ptr;
}

/* start_auth - starts auth (identd) and dns queries for a client */
void start_auth(struct Client * client)
{
  struct AuthRequest *auth = 0;
  s_assert(0 != client);
  if (client == NULL)
    return;
  auth = make_auth_request(client);
  auth->dns_query.ptr = auth;
  auth->dns_query.callback = auth_dns_callback;
  sendheader(client->localClient->fd, REPORT_LOGON);
  if (adns_getaddr(&client->localClient->ip, client->localClient->aftype,
      &auth->dns_query, 0)) {
#ifdef IPV6
    /* Try ip6.int if configured to */
    if (client->localClient->aftype == AF_INET6
      && ConfigFileEntry.fallback_to_ip6_int == 1 && auth->ip6_int == 0) {
      auth->ip6_int = 1;
      if (!adns_getaddr(&client->localClient->ip, client->localClient->aftype,
          &auth->dns_query, 1))
        SetDNSPending(auth);
      else
        sendheader(client->localClient->fd, REPORT_FAIL_DNS);
    }
    else
#endif
      sendheader(client->localClient->fd, REPORT_FAIL_DNS);
  }
  else
    SetDNSPending(auth);
  if (ConfigFileEntry.disable_auth == 0)
    start_auth_query(auth);
  /* Not checking DNS or AUTH..let them free */
  if (!IsDoingAuth(auth) && !IsDNSPending(auth)) {
    release_auth_client(auth);
    return;
  }
  dlinkAdd(auth, &auth->node, &auth_poll_list);
}
/* timeout_auth_queries - timeout resolver and identd requests
 * allow clients through if requests failed
 */
static void timeout_auth_queries_event(void *notused)
{
  dlink_node *ptr;
  dlink_node *next_ptr;
  struct AuthRequest *auth;
  DLINK_FOREACH_SAFE(ptr, next_ptr, auth_poll_list.head) {
    auth = ptr->data;
    if (auth->timeout < CurrentTime) {
      if (auth->fd >= 0)
        fd_close(auth->fd);
      if (IsDoingAuth(auth)) {
        ClearAuth(auth);
        sendheader(auth->client->localClient->fd, REPORT_FAIL_ID);
      }
      if (IsDNSPending(auth)) {
        ClearDNSPending(auth);
        delete_adns_queries(&auth->dns_query);
        sendheader(auth->client->localClient->fd, REPORT_FAIL_DNS);
      }
      ilog(L_INFO, "DNS/AUTH timeout %s", log_client_name(auth->client));
      auth->client->since = CurrentTime;
      release_auth_client(auth);
    }
  }
}

/* auth_connect_callback - deal with the result of comm_connect_tcp()
 *
 * If the connection failed, we simply close the auth fd and report a failure.
 * If the connection suceeded send the ident server a query giving "theirport ,
 * ourport". The write is only attempted once so it is deemed to be a fail if
 * the entire write doesn't write all the data given. This shouldn't be a
 * problem since the socket should have a write buffer far greater than this
 * message to store it in should problems arise -avalon
 */
static void auth_connect_callback(int fd, int error, void *data)
{
  struct AuthRequest *auth = data;
  struct sockaddr_in us;
  struct sockaddr_in them;
  char authbuf[32];
  socklen_t ulen = sizeof(struct sockaddr_in);
  socklen_t tlen = sizeof(struct sockaddr_in);
  /* Check the error */
  if (error != COMM_OK) {
    /* We had an error during connection :( */
    auth_error(auth);
    return;
  }
  if (getsockname(auth->client->localClient->fd, (struct sockaddr *) & us,
      (socklen_t *) & ulen) || getpeername(auth->client->localClient->fd,
      (struct sockaddr *) & them, (socklen_t *) & tlen)) {
    ilog(L_INFO, "auth get{sock,peer}name error for %s:%m",
      log_client_name(auth->client));
    auth_error(auth);
    return;
  }
  ircsprintf(authbuf, "%u , %u\r\n", (unsigned int)ntohs(them.sin_port),
    (unsigned int)ntohs(us.sin_port));
  if (send(auth->fd, authbuf, strlen(authbuf), 0) == -1) {
    auth_error(auth);
    return;
  }
  ClearAuthConnect(auth);
  SetAuthPending(auth);
  read_auth_reply(auth->fd, auth);
}

/* read_auth_reply - read the reply from the ident server we connected to
 * We only give it one shot, if the reply isn't good the first time
 * fail the authentication entirely. -Bleep
 */
#define AUTH_BUFSIZ 128
static void read_auth_reply(int fd, void *data)
{
  struct AuthRequest *auth = data;
  char *s = (char *)NULL;
  char *t = (char *)NULL;
  int len;
  int count;
  char buf[AUTH_BUFSIZ + 1]; /* buffer to read auth reply into */
  len = recv(auth->fd, buf, AUTH_BUFSIZ, 0);
  if (len < 0 && ignoreErrno(errno)) {
    comm_setselect(fd, FDLIST_IDLECLIENT, COMM_SELECT_READ, read_auth_reply,
      auth, 0);
    return;
  }
  if (len > 0) {
    buf[len] = '\0';
    if ((s = GetValidIdent(buf))) {
      t = auth->client->username;
      while (*s == '~' || *s == '^')
        s++;
      for (count = USERLEN; *s && count; s++) {
        if (*s == '@')
          break;
        if (!IsSpace(*s) && *s != ':' && *s != '[') {
          *t++ = *s;
          count--;
        }
      }
      *t = '\0';
    }
  }
  fd_close(auth->fd);
  auth->fd = -1;
  ClearAuth(auth);
  if (s == NULL) {
    ++ServerStats->is_abad;
    strcpy(auth->client->username, "unknown");
    sendheader(auth->client->localClient->fd, REPORT_FAIL_ID);
  }
  else {
    ++ServerStats->is_asuc;
    SetGotId(auth->client);
  }
  if (!IsDNSPending(auth))
    release_auth_client(auth);
}

/* delete_auth_queries - cancel pending DNS/AUTH requests */
void delete_auth_queries(struct Client * target_p)
{
  struct AuthRequest *auth;
  if (target_p == NULL || target_p->localClient == NULL ||
      target_p->localClient->auth_request == NULL)
    return;
  auth = target_p->localClient->auth_request;
  target_p->localClient->auth_request = NULL;
  if (IsDNSPending(auth))
    delete_adns_queries(&auth->dns_query);
  if (auth->fd >= 0)
    fd_close(auth->fd);
  dlinkDelete(&auth->node, &auth_poll_list);
  free_auth_request(auth);
}
