/* epoll.c: Linux epoll compatible network routines.
 *
 *  Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 *  Copyright (C) 1996-2002 Hybrid Development Team
 *  Copyright (C) 2001 Adrian Chadd <adrian@creative.net.au>
 *  Copyright (C) 2002 ircd-ratbox development team
 *  Copyright (C) 2002 Aaron Sethman <androsyn@ratbox.org>
 *  Copyright (C) 2003-2004 ircd-immortal development team
 *
 *  This file is licensed under the GNU General Public License, which is
 *  distributed as LICENSE.
 */

#include "config.h"
#include "stdinc.h"
#ifdef USE_EPOLL
#include <sys/epoll.h>
#include "s_bsd.h"
#include "class.h"
#include "client.h"
#include "common.h"
#include "irc_string.h"
#include "ircd.h"
#include "listener.h"
#include "numeric.h"
#include "packet.h"
#include "res.h"
#include "restart.h"
#include "s_auth.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "s_stats.h"
#include "send.h"
#include "memory.h"

static int ep;

#ifndef HAVE_SYS_EPOLL_H
#error "You must have a working sys/epoll.h"
#else
#include <sys/epoll.h>
#include <sys/syscall.h>

/* use our own for now */
_syscall1(int, epoll_create, int, maxfds);
_syscall4(int, epoll_ctl, int, epfd, int, op, int, fd, struct epoll_event *,
  events);
_syscall4(int, epoll_wait, int, epfd, struct epoll_event *, pevents,
  int, maxevents, int, timeout);
#endif /* HAVE_EPOLL_CTL */

/* init_netio - called to initialize network loop code */
void init_netio(void)
{
  ep = epoll_create(HARD_FDLIMIT);
  if (ep < 0) {
    ilog(L_CRIT, "init_netio: Couldn't open epoll fd!\n");
    exit(115);           /* Whee! */
  }
  fd_note(ep, "epoll file descriptor");
}
/* comm_setselect - called to [de]register interest in a pending IO state */
void comm_setselect(int fd, fdlist_t list, unsigned int type, PF * handler,
  void *client_data, time_t timeout)
{
  struct epoll_event ep_event;
  fde_t *F = &fd_table[fd];
  int old_flags = F->pflags;
  int op = -1;
  s_assert(fd >= 0);
  s_assert(F->flags.open);
  /* Update the list, even though we're not using it .. */
  F->list = list;
  if (type & COMM_SELECT_READ) {
    if (handler != NULL)
      F->pflags |= EPOLLIN;
    else
      F->pflags &= ~EPOLLIN;
    F->read_handler = handler;
    F->read_data = client_data;
  }
  if (type & COMM_SELECT_WRITE) {
    if (handler != NULL)
      F->pflags |= EPOLLOUT;
    else
      F->pflags &= ~EPOLLOUT;
    F->write_handler = handler;
    F->write_data = client_data;
  }
  if (timeout)
    F->timeout = CurrentTime + (timeout / 1000);
  if (old_flags == 0 && F->pflags == 0)
    return;
  else if (F->pflags <= 0)
    op = EPOLL_CTL_DEL;
  else if (old_flags == 0 && F->pflags > 0)
    op = EPOLL_CTL_ADD;
  else if (F->pflags != old_flags)
    op = EPOLL_CTL_MOD;
  if (op == -1)
    return;

  ep_event.events = F->pflags;
  ep_event.data.ptr = F;
  if (epoll_ctl(ep, op, fd, &ep_event) != 0)
    ilog(L_CRIT, "comm_setselect(): epoll_ctl failed: %s", strerror(errno));
}
/* comm_select - do IO events
 * input  - maximum time delay
 * output - success, 0
 *        - failure, -1
 * side effects - deregisters future interest in IO and calls the handlers if
 *                an event occurs
 *
 * check all connections for new connections and input data that is to be
 * processed. also check for connections with data queued and if we can send it.
 *
 * called to do the new-style IO (courtesy of squid), this routine handles the
 * stuff we've hidden in comm_setselect and fd_table[] and calls callbacks for
 * IO ready events
 */
int comm_select(unsigned long delay)
{
  int num, i, flags, old_flags, op;
  struct epoll_event ep_event;
  static struct epoll_event pfd[HARD_FDLIMIT];
  void *data;
  num = epoll_wait(ep, pfd, HARD_FDLIMIT, delay);
  set_time();
  if (num < 0 && !ignoreErrno(errno))
    return COMM_ERROR;
  if (num == 0)
    return COMM_OK;
  for (i = 0; i < num; i++) {
    PF *hdl;
    fde_t *F = pfd[i].data.ptr;
    old_flags = F->pflags;
    if (pfd[i].events & (EPOLLIN | EPOLLHUP | EPOLLERR)) {
      hdl = F->read_handler;
      data = F->read_data;
      F->read_handler = NULL;
      F->read_data = NULL;
      if (hdl)
        hdl(F->fd, data);
      else
        ilog(L_WARN, "epoll.c: NULL read handler called");
    }
    if (F->flags.open == 0)
      continue;
    if (pfd[i].events & (EPOLLOUT | EPOLLHUP | EPOLLERR)) {
      hdl = F->write_handler;
      data = F->write_data;
      F->write_handler = NULL;
      F->write_data = NULL;
      if (hdl)
        hdl(F->fd, data);
      else
        ilog(L_WARN, "epoll.c: NULL write handler called");
    }
    if (F->flags.open == 0)
      continue;
    flags = 0;
    if (F->read_handler != NULL)
      flags |= EPOLLIN;
    if (F->write_handler != NULL)
      flags |= EPOLLOUT;
    if (old_flags != flags) {
      if (flags == 0)
        op = EPOLL_CTL_DEL;
      else
        op = EPOLL_CTL_MOD;
      F->pflags = ep_event.events = flags;
      ep_event.data.ptr = F;
      if (epoll_ctl(ep, op, F->fd, &ep_event) != 0) {
        ilog(L_CRIT, "comm_setselect(): epoll_ctl failed: %s", strerror(errno));
      }
    }
  }
  return COMM_OK;
}
#endif
