/* s_bsd_poll.c: POSIX poll() compatible network routines.
 *
 *  Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 *  Copyright (C) 1996-2002 Hybrid Development Team
 *  Copyright (C) 2001 Adrian Chadd <adrian@creative.net.au>
 *  Copyright (C) 2002      ircd-ratbox development team
 *  Copyright (C) 2003-2004 ircd-immortal development team
 *
 *  This file is licensed under the GNU General Public License, which is
 *  distributed as LICENSE.
 */

#include "config.h"
#ifdef USE_POLL
#include "stdinc.h"
#include <sys/poll.h>
#include "fdlist.h"
#include "s_bsd.h"
#include "class.h"
#include "client.h"
#include "common.h"
#include "irc_string.h"
#include "ircd.h"
#include "listener.h"
#include "numeric.h"
#include "packet.h"
#include "res.h"
#include "restart.h"
#include "s_auth.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "s_stats.h"
#include "send.h"
#include "s_debug.h"
#include "s_bsd.h"
#include "memory.h"

#ifndef POLLRDNORM
#define POLLRDNORM POLLIN
#endif
#ifndef POLLWRNORM
#define POLLWRNORM POLLOUT
#endif

struct _pollfd_list {
  struct pollfd pollfds[MAXCONNECTIONS];
  int maxindex; /* highest FD number */
};
typedef struct _pollfd_list pollfd_list_t;
pollfd_list_t pollfd_list;
static void poll_update_pollfds(int, short, PF *);
static unsigned long last_count = 0; 
static unsigned long empty_count = 0;

/* Private functions */
/* find a spare slot in the fd list */
static inline int poll_findslot(void)
{
  int i;
  for (i = 0; i < MAXCONNECTIONS; i++) {
    if (pollfd_list.pollfds[i].fd == -1)
      return i; /* MATCH */
  }
  s_assert(1 == 0);
  /* NOTREACHED */
  return -1;
}

/* set and clear entries in the pollfds[] array. */
static void poll_update_pollfds(int fd, short event, PF * handler)
{
  fde_t *F = &fd_table[fd];
  int comm_index;
  if (F->comm_index < 0) {
    F->comm_index = poll_findslot();
  }
  comm_index = F->comm_index;
  /* Update the events */
  if (handler) {
    F->list = FDLIST_IDLECLIENT;
    pollfd_list.pollfds[comm_index].events |= event;
    pollfd_list.pollfds[comm_index].fd = fd;
    /* update maxindex here */
    if (comm_index > pollfd_list.maxindex)
      pollfd_list.maxindex = comm_index;
  }
  else {
    if (comm_index >= 0) {
      pollfd_list.pollfds[comm_index].events &= ~event;
      if (pollfd_list.pollfds[comm_index].events == 0) {
        pollfd_list.pollfds[comm_index].fd = -1;
        pollfd_list.pollfds[comm_index].revents = 0;
        F->comm_index = -1;
        F->list = FDLIST_NONE;
        /* update pollfd_list.maxindex here */
        if (comm_index == pollfd_list.maxindex)
          while (pollfd_list.maxindex >= 0 &&
            pollfd_list.pollfds[pollfd_list.maxindex].fd == -1)
            pollfd_list.maxindex--;
      }
    }
  }
}

/* Public functions */
/* init_netio - called to initialize network loop code */
void init_netio(void)
{
  int fd;
  for (fd = 0; fd < MAXCONNECTIONS; fd++)
    pollfd_list.pollfds[fd].fd = -1;
  pollfd_list.maxindex = 0;
}

/* comm_setselect - called to [de]register interest in a pending IO state */
void comm_setselect(int fd, fdlist_t list, unsigned int type, PF * handler,
  void *client_data, time_t timeout)
{
  fde_t *F = &fd_table[fd];
  s_assert(fd >= 0);
  s_assert(F->flags.open);
  if (type & COMM_SELECT_READ) {
    F->read_handler = handler;
    F->read_data = client_data;
    poll_update_pollfds(fd, POLLRDNORM, handler);
  }
  if (type & COMM_SELECT_WRITE) {
    F->write_handler = handler;
    F->write_data = client_data;
    poll_update_pollfds(fd, POLLWRNORM, handler);
  }
  if (timeout)
    F->timeout = CurrentTime + (timeout / 1000);
}

static void irc_sleep(unsigned long useconds) 	 
{
#ifdef HAVE_NANOSLEEP 	 
  struct timespec t; 	 
  t.tv_sec = useconds / (unsigned long) 1000000; 	 
  t.tv_nsec = (useconds % (unsigned long) 1000000) * 1000; 	 
  nanosleep(&t, (struct timespec *) NULL); 	 
#else 	 
  struct timeval t; 	 
  t.tv_sec = 0; 	 
  t.tv_usec = useconds; 	 
  select(0, NULL, NULL, NULL, &t); 	 
#endif
  return; 	 
}

/* comm_select - do IO events
 * input  - maximum time delay
 * output - success, 0
 *        - failure, -1
 * side effects - deregisters future interest in IO and calls the handlers if
 *                an event occurs
 *
 * check all connections for new connections and input data that is to be
 * processed. also check for connections with data queued and if we can send it.
 *
 * called to do the new-style IO (courtesy of squid), this routine handles the
 * stuff we've hidden in comm_setselect and fd_table[] and calls callbacks for
 * IO ready events
 */
int comm_select(unsigned long delay)
{
  int num;
  int fd;
  int ci;
  unsigned long ndelay;
  PF *hdl;

  if (last_count > 0) {
    empty_count = 0;
    ndelay = 0;
  }
  else {
    ndelay = ++empty_count * 15000;
    if (ndelay > delay * 1000)
      ndelay = delay * 1000;	
  }
	
  for (;;) {
    /* XXX kill that +1 later ! -- adrian */
    if (ndelay > 0)
      irc_sleep(ndelay); 
    last_count = num = poll(pollfd_list.pollfds, pollfd_list.maxindex + 1, 0);
    if (num >= 0)
      break;
    if (ignoreErrno(errno))
      continue;
    /* error! */
    set_time();
    return -1;
    /* NOTREACHED */
  }
  /* update current time again, eww */
  set_time();
  if (num == 0)
    return 0;
  for (ci = 0; ci < pollfd_list.maxindex + 1; ci++) {
    fde_t *F;
    int revents;
    if (((revents = pollfd_list.pollfds[ci].revents) == 0) ||
      (pollfd_list.pollfds[ci].fd) == -1)
      continue;
    fd = pollfd_list.pollfds[ci].fd;
    F = &fd_table[fd];
    if (revents & (POLLRDNORM | POLLIN | POLLHUP | POLLERR)) {
      hdl = F->read_handler;
      F->read_handler = NULL;
      poll_update_pollfds(fd, POLLRDNORM, NULL);
      if (hdl)
        hdl(fd, F->read_data);
    }
    if (revents & (POLLWRNORM | POLLOUT | POLLHUP | POLLERR)) {
      hdl = F->write_handler;
      F->write_handler = NULL;
      poll_update_pollfds(fd, POLLWRNORM, NULL);
      if (hdl)
        hdl(fd, F->write_data);
    }
  }
  return 0;
}
#endif
