/* s_bsd_select.c: select() compatible network routines.
 *
 *  Copyright (C) 1990 Jarkko Oikarinen and University of Oulu, Co Center
 *  Copyright (C) 1996-2002 Hybrid Development Team
 *  Copyright (C) 2001      Adrian Chadd <adrian@creative.net.au>
 *  Copyright (C) 2002      ircd-ratbox development team
 *  Copyright (C) 2003-2004 ircd-immortal development team
 *
 *  This file is licensed under the GNU General Public License, which is
 *  distributed as LICENSE.
 */

#include "config.h"
#ifdef USE_SELECT
#include "stdinc.h"
#include "fdlist.h"
#include "s_bsd.h"
#include "class.h"
#include "client.h"
#include "common.h"
#include "irc_string.h"
#include "ircd.h"
#include "listener.h"
#include "numeric.h"
#include "packet.h"
#include "res.h"
#include "restart.h"
#include "s_auth.h"
#include "s_conf.h"
#include "s_log.h"
#include "s_serv.h"
#include "s_stats.h"
#include "send.h"
#include "s_debug.h"
#include "s_bsd.h"
#include "memory.h"

#if HARD_FDLIMIT_ >= FD_SETSIZE
#error "HARD_FDLIMIT_ must be <FD_SETSIZE (try poll instead of select)"
#endif

/* Note that this is only a single list - multiple lists is kinda pointless
 * under select because the list size is a function of the highest FD -adrian
 */
fd_set select_readfds;
fd_set select_writefds;
fd_set tmpreadfds;
fd_set tmpwritefds;
static void select_update_selectfds(int fd, short event, PF * handler);

/* Private functions */
/* set and clear entries in the select array */
static void select_update_selectfds(int fd, short event, PF * handler)
{
  /* Update the read/write set */
  if (event & COMM_SELECT_READ) {
    if (handler)
      FD_SET(fd, &select_readfds);
    else
      FD_CLR(fd, &select_readfds);
  }
  if (event & COMM_SELECT_WRITE) {
    if (handler)
      FD_SET(fd, &select_writefds);
    else
      FD_CLR(fd, &select_writefds);
  }
}

/* Public functions */
/* init_netio - called to initialize network loop code */
void init_netio(void)
{
  FD_ZERO(&select_readfds);
  FD_ZERO(&select_writefds);
}

/* comm_setselect - called to [de]register interest in a pending IO state */
void comm_setselect(int fd, fdlist_t list, unsigned int type, PF * handler,
  void *client_data, time_t timeout)
{
  fde_t *F = &fd_table[fd];
  s_assert(fd >= 0);
  s_assert(F->flags.open);
#ifdef NOTYET
  debug(5, 5) ("commSetSelect: FD %d type %d, %s\n", fd, type, handler ? "SET" :
    "CLEAR");
#endif
  if (type & COMM_SELECT_READ) {
    F->read_handler = handler;
    F->read_data = client_data;
    select_update_selectfds(fd, COMM_SELECT_READ, handler);
  }
  if (type & COMM_SELECT_WRITE) {
    F->write_handler = handler;
    F->write_data = client_data;
    select_update_selectfds(fd, COMM_SELECT_WRITE, handler);
  }
  if (timeout)
    F->timeout = CurrentTime + (timeout / 1000);
}

/* comm_select - do IO events
 * input  - maximum time delay
 * output - success, 0
 *        - failure, -1
 * side effects - deregisters future interest in IO and calls the handlers if
 *                an event occurs
 *
 * check all connections for new connections and input data that is to be
 * processed. also check for connections with data queued and if we can send it.
 *
 * called to do the new-style IO (courtesy of squid), this routine handles the
 * stuff we've hidden in comm_setselect and fd_table[] and calls callbacks for
 * IO ready events
 */

int comm_select(unsigned long delay)
{
  int num;
  int fd;
  PF *hdl;
  fde_t *F;
  struct timeval to;
  /* Copy over the read/write sets so we don't have to rebuild em */
  memcpy(&tmpreadfds, &select_readfds, sizeof(fd_set));
  memcpy(&tmpwritefds, &select_writefds, sizeof(fd_set));
  for (;;) {
    to.tv_sec = 0;
    to.tv_usec = delay * 1000;
    num = select(highest_fd + 1, &tmpreadfds, &tmpwritefds, NULL, &to);
    if (num >= 0)
      break;
    if (ignoreErrno(errno))
      continue;
    set_time();
    return -1; /* error! */
    /* NOTREACHED */
  }
  set_time();
  if (num == 0)
    return 0;
  for (fd = 0; fd < highest_fd + 1; fd++) {
    F = &fd_table[fd];
    if (FD_ISSET(fd, &tmpreadfds)) {
      hdl = F->read_handler;
      F->read_handler = NULL;
      select_update_selectfds(fd, COMM_SELECT_READ, NULL);
      if (hdl)
        hdl(fd, F->read_data);
    }
    if (FD_ISSET(fd, &tmpwritefds)) {
      hdl = F->write_handler;
      F->write_handler = NULL;
      select_update_selectfds(fd, COMM_SELECT_WRITE, NULL);
      if (hdl)
        hdl(fd, F->write_data);
    }
  }
  return 0;
}
#endif /* USE_SELECT */
