/* tools.c: Various functions needed here and there.
 *
 * Copyright (C) 1996-2002 Hybrid Development Team
 * Copyright (C) 2002      ircd-ratbox development team
 *
 * This file is licensed under the GNU General Public License, which is
 * distributed as LICENSE.
 *
 * Here is the original header:
 *
 * Useful stuff, ripped from places ..
 * adrian chadd <adrian@creative.net.au>
 *
 * When you update these functions make sure you update the ones in tools.h
 * as well!!!
 */

#include "stdinc.h"
#include "tools.h"
#include "balloc.h"
#ifndef NDEBUG
/* frob some memory. debugging time. -adrian */
void mem_frob(void *data, int len)
{
  /* correct for Intel only! little endian */
  unsigned char b[4] = {0xef, 0xbe, 0xad, 0xde};
  int i;
  char *cdata = data;
  for (i = 0; i < len; i++) {
    *cdata = b[i % 4];
    cdata++;
  }
}
#endif

/* dlink_ routines from squid, except dlinkAddBefore, which is mine -adrian */
void dlinkAdd(void *data, dlink_node * m, dlink_list * list)
{
  m->data = data;
  m->next = list->head;
  /* Assumption: If list->tail != NULL, list->head != NULL */
  if (list->head != NULL)
    list->head->prev = m;
  else if (list->tail == NULL)
    list->tail = m;
  list->head = m;
  list->length++;
}

void dlinkAddBefore(dlink_node * b, void *data, dlink_node * m,
  dlink_list * list)
{
  /* Shortcut - if its the first one, call dlinkAdd only */
  if (b == list->head)
    dlinkAdd(data, m, list);
  else {
    m->data = data;
    b->prev->next = m;
    m->prev = b->prev;
    b->prev = m;
    m->next = b;
    list->length++;
  }
}

void dlinkMoveTail(dlink_node *m, dlink_list *list)
{
 	if(list->tail == m)
 	        return;
 	/* From here assume that m->next != NULL as that can only be at the tail and
 	 * assume that the node is on the list
 	 */
 	m->next->prev = m->prev;
  if(m->prev != NULL)
    m->prev->next = m->next;
  else
    list->head = m->next;
  list->tail->next = m;
  m->prev = list->tail;
  m->next = NULL;
  list->tail = m;
}

void dlinkAddTail(void *data, dlink_node * m, dlink_list * list)
{
  m->data = data;
  m->next = NULL;
  m->prev = list->tail;
  /* Assumption: If list->tail != NULL, list->head != NULL */
  if (list->tail != NULL)
    list->tail->next = m;
  else if (list->head == NULL)
    list->head = m;
  list->tail = m;
  list->length++;
}

/* Profiles say this function is called the most often of all non-spontaneous
 * functions. Make sure it's efficient!
 */
void dlinkDelete(dlink_node * m, dlink_list * list)
{
  /* Assumption: If m->next == NULL, then list->tail == m and, If m->prev ==
   * NULL, then list->head == m
   */
  if (m->next)
    m->next->prev = m->prev;
  else
    list->tail = m->prev;
  if (m->prev)
    m->prev->next = m->next;
  else
    list->head = m->next;
  /* Set this to NULL does matter */
  m->next = m->prev = NULL;
  list->length--;
}

/* dlinkFind - look for ptr in the llist pointed to by link
 * inputs       - list to search
 *              - data
 * output       - pointer to link or NULL if not found
 */
dlink_node * dlinkFind(dlink_list * list, void *data)
{
  dlink_node *ptr;
  DLINK_FOREACH(ptr, list->head) {
    if (ptr->data == data)
      return ptr;
  }
  return NULL;
}

void dlinkMoveList(dlink_list * from, dlink_list * to)
{
  /* case one, nothing in from list */
  if (from->head == NULL)
    return;
  /* case two, nothing in to list */
  if (to->head == NULL) {
    to->head = from->head;
    to->tail = from->tail;
    from->head = from->tail = NULL;
    to->length = from->length;
    from->length = 0;
    return;
  }
  /* last case, play with the links */
  from->tail->next = to->head;
  to->head->prev = from->tail;
  to->head = from->head;
  from->head = from->tail = NULL;
  to->length += from->length;
  from->length = 0;
}
dlink_node * dlinkFindDelete(dlink_list * list, void *data)
{
  dlink_node *m;
  DLINK_FOREACH(m, list->head) {
    if (m->data != data)
      continue;
    if (m->next)
      m->next->prev = m->prev;
    else
      list->tail = m->prev;
    if (m->prev)
      m->prev->next = m->next;
    else
      list->head = m->next;
    m->next = m->prev = NULL;
    list->length--;
    return m;
  }
  return NULL;
}
int dlinkFindDestroy(dlink_list * list, void *data)
{
  void *ptr = dlinkFindDelete(list, data);
  if (ptr != NULL) {
    free_dlink_node(ptr);
    return 1;
  }
  return 0;
}
void dlinkMoveNode(dlink_node * m, dlink_list * oldlist, dlink_list * newlist)
{
  /* Assumption: If m->next == NULL, then list->tail == m and, If m->prev ==
   * NULL, then list->head == m
   */
  if (m->next)
    m->next->prev = m->prev;
  else
    oldlist->tail = m->prev;
  if (m->prev)
    m->prev->next = m->next;
  else
    oldlist->head = m->next;
  m->prev = NULL;
  m->next = newlist->head;
  if (newlist->head != NULL)
    newlist->head->prev = m;
  else if (newlist->tail == NULL)
    newlist->tail = m;
  newlist->head = m;
  oldlist->length--;
  newlist->length++;
}

/* init_dlink_nodes */
static BlockHeap *dnode_heap;
void init_dlink_nodes(void)
{
  dnode_heap = BlockHeapCreate(sizeof(dlink_node), DNODE_HEAP_SIZE);
  if (dnode_heap == NULL)
    outofmemory();
}

/* make_dlink_node
 * output - pointer to new dlink_node
 */
dlink_node * make_dlink_node(void)
{
  dlink_node *lp;
  lp = (dlink_node *) BlockHeapAlloc(dnode_heap);
  lp->next = NULL;
  lp->prev = NULL;
  return lp;
}

/* free_dlink_node
 * inputs       - pointer to dlink_node
 * side effects - free given dlink_node
 */
void free_dlink_node(dlink_node * ptr)
{
  BlockHeapFree(dnode_heap, ptr);
}
