/* simple password generator by Nelson Minar (minar@reed.edu)
 * copyright 1991, all rights reserved.
 * You can use this code as long as my name stays with it.
 *
 * md5 and getopt support by W. Campbell (wcampbel@botbay.net)
 * /dev/random for salt generation by Aaron Sethman (androsyn@ratbox.org)
 * VMS support by Edward Brocklesby
 * crypt.c implementation by phk@login.dknet.dk
 * Updated messages by Rachel Llorenna (frequency@frequency.dk)
 */

#ifdef VMS
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include descrip
#include psldef
#include iodef
#include ssdef
#include starlet
#include uaidef
#include stsdef

#define FLAG_MD5      0x00000001
#define FLAG_DES      0x00000002
#define FLAG_SALT     0x00000004
#define FLAG_PASS     0x00000008
#define FLAG_LENGTH   0x00000010
#define FLAG_BLOWFISH 0x00000020
#define FLAG_ROUNDS   0x00000040
#define FLAG_EXT      0x00000080
#define FLAG_PURDY    0x00000100
#define FLAG_USER     0x00000200

static char *getpass();
extern char *crypt();
char *make_des_salt();
char *make_ext_salt(int);
char *make_ext_salt_para(int, char *);
char *make_md5_salt(int);
char *make_md5_salt_para(char *);
char *make_bf_salt(int, int);
char *make_bf_salt_para(int, char *);
char *int_to_base64(int);
void full_usage();
void brief_usage();
static char saltChars[] = /* 0 .. 63, ascii - 64 */
  "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

int main(int argc, char *argv[])
{
  char *plaintext = NULL;
  extern char *optarg;
  int c;
  char *saltpara = NULL;
  char *salt;
  char *username;
  int flag = 0;
  int length = 0;
  srandom(time(NULL));
  while ((c = getopt(argc, argv, "mvu:h?l:s:p:")) != -1) {
    switch (c) {
     case 'm':
      flag |= FLAG_MD5;
      break;
     case 'l':
      flag |= FLAG_LENGTH;
      length = atoi(optarg);
      break;
     case 's':
      flag |= FLAG_SALT;
      saltpara = optarg;
      break;
     case 'p':
      flag |= FLAG_PASS;
      plaintext = optarg;
      break;
     case 'v':
      flag |= FLAG_PURDY;
      break;
     case 'u':
      flag |= FLAG_USER;
      username = optarg;
      break;
     case 'h':
      full_usage();
      /* NOTREACHED */
      break;
     case '?':
      brief_usage();
      /* NOTREACHED */
      break;
     default:
      printf("Invalid Option: -%c\n", c);
      break;
    }
  }
  /* default to md5 */
  if (!(flag & FLAG_MD5) && !(flag & FLAG_PURDY))
    flag |= FLAG_MD5;
  if (flag & FLAG_MD5) {
    if (length == 0)
      length = 8;
    if (flag & FLAG_SALT)
      salt = make_md5_salt_para(saltpara);
    else
      salt = make_md5_salt(length);
  }
  if (flag & FLAG_PASS) {
    if (!plaintext)
      printf("Please enter a valid password.\n");
  }
  else
    plaintext = getpass("plaintext: ");
  if ((flag & FLAG_PURDY) && !username)
    username = getpass("username: ");
  if (flag & FLAG_PURDY) {
    $DESCRIPTOR(plain_desc, plaintext);
    $DESCRIPTOR(user_desc, username);
    unsigned long r;
    long hash[2];
    r = sys$hash_password(&plain_desc, UAI$C_PURDY_S, 0, &user_desc, &hash);
    if (!$VMS_STATUS_SUCCESS(r))
      exit(r);
    printf("%x%x\n", hash[1], hash[0]);
  }
  else
    printf("%s\n", crypt(plaintext, salt));
  return 0;
}

char *int_to_base64(int value)
{
  static char buf[5];
  int i;
  for (i = 0; i < 4; i++) {
    buf[i] = saltChars[value & 63];
    value >>= 6; /* Right shifting 6 places is the same as dividing by 64 */
  }
  buf[i] = '\0'; /* not needed as it's static, and thus initialized to \0 */
  return buf;
}

char *make_md5_salt_para(char *saltpara)
{
  static char salt[21];
  if (saltpara && (strlen(saltpara) <= 16)) {
    /* sprintf used for portability, length is checked above */
    sprintf(salt, "$1$%s$", saltpara);
    return salt;
  }
  printf("Invalid salt, please use up to 16 random alphanumeric characters\n");
  exit(1);
  /* NOTREACHED */
  return NULL;
}

char *make_md5_salt(int length)
{
  static char salt[21];
  int i;
  if (length > 16) {
    printf("MD5 salt length too long\n");
    exit(0);
  }
  salt[0] = '$';
  salt[1] = '1';
  salt[2] = '$';
  for (i = 3; i < (length + 3); i++)
    salt[i] = saltChars[random() % 64];
  salt[length + 3] = '$';
  salt[length + 4] = '\0';
  return salt;
}

void full_usage()
{
  printf("mkpasswd [-m|-v] [-l saltlen] [-r rounds] [-s salt] [-p plaintext] [-u username]\n");
  printf("-m Generate a MD5 password\n");
  printf("-v Generate a VMS Purdy password\n");
  printf("-l Specify a length for a random MD5 salt\n");
  printf("-s Specify a salt, up to 16 alphanumeric characters for MD5\n");
  printf("-p Specify a plaintext password to use\n");
  printf("-u Specify the username for Purdy hash\n");
  printf("-v Generate a VMS Purdy password\n");
  printf("Example: mkpasswd -m -s 3dr -p test\n");
  exit(0);
}
void brief_usage()
{
  printf("mkpasswd - password hash generator\n");
  printf(" MD5: mkpasswd -m [-l saltlen] [-s salt] [-p plaintext]\n");
  printf(" VMS Purdy: mkpasswd -v [-p plaintext] [-u username]\n");
  printf("Use -h for full usage\n");
  exit(0);
}

/* getpass replacement for VMS */
static char * getpass(char *prompt)
{
  static char password[2][64];
  static int i = 0;
  int result;
  int chan;
  int promptlen;
  unsigned short r;
  $DESCRIPTOR(devnam, "SYS$INPUT");
  struct {
    short result;
    short count;
    int info;
  } iosb;
  promptlen = strlen(prompt);
  r = sys$assign(&devnam, &chan, PSL$C_USER, 0, 0);
  if (!$VMS_STATUS_SUCCESS(r))
    exit(r);
  r = sys$qiow(0, chan, IO$_READPROMPT | IO$M_PURGE | IO$M_NOECHO, &iosb, 0, 0,
    password[i], 255, 0, 0, prompt, promptlen);
  if ($VMS_STATUS_SUCCESS(r))
    r = iosb.result;
  if (!$VMS_STATUS_SUCCESS(r))
    exit(r);
  password[i][iosb.count] = '\0';
  r = sys$dassgn(chan);
  printf("\n");
  if (!$VMS_STATUS_SUCCESS(r))
    exit(r);
  return password[i++];
}
#endif /* VMS */
